
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * alpha2cut.c - 2-rank cutting-plane algorithm
 *
 * NOTE: We noticed that sometimes CPLEX does not add a cut.
 * This might be caused by the deletion of variables in the presolved problem.
 * Setting CPX_PARAM_PREIND to CPX_OFF might solve this problem.
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOinitalpha2callback(struct ecointernal *ei)
{
  int vertices, status;

  vertices = ei->grp->vertices;
  ei->alpha2_cut_set = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->alpha2clique_cut_set = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->alpha2_cut_temp1 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->alpha2_cut_temp2 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->alpha2_cut_temp3 = ECOallocvector(&status, vertices);
  if (status) return status;

  ei->alpha2_cuts_applied = 0;

  return ECOERR_OK;
}

int ECOalpha2separation(struct ecointernal *ei, int j)
{
  int status, v, i, vertices, flag, bestv1, bestv2, cutadded, tries;
  int alpha2_size, alpha2clique_size, v1, v2, k, degree, degree2, cand2count;
  ECOVECptr descending, forbidden, alpha2_set, alpha2clique_set;
  ECOVECptr candidates1, candidates2, noneighbor, noneighborv1, clqv1set;
  double bestval, val;

  /* set variables */
  vertices = ei->grp->vertices;
  descending = ei->cut_descending;
  forbidden = ei->cut_forbidden;
  alpha2_set = ei->alpha2_cut_set;
  alpha2clique_set = ei->alpha2clique_cut_set;
  candidates1 = ei->alpha2_cut_temp1;
  candidates2 = ei->alpha2_cut_temp2;
  noneighbor = ei->alpha2_cut_temp3;
  noneighborv1 = ei->cut_temp1;
  clqv1set = ei->cut_temp2;
  tries = 0;

  if (ei->alpha2_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;

  /* we scan up to ceil(n/2)-1 */
  if (j >= ECOceil(vertices, 2)) return ECOERR_OK;

  /* check if all vertices are forbidden (here, we do not know if 'descending' is already defined) */
  k = 0;
  for (v = 1; v <= vertices; v++) if (ECOvget(forbidden, v) == 0) k++;

  /* we ask for at least 5 non-forbidden vertices */
  flag = (k >= 5) ? 0 : 1;

  while (flag == 0) {
    /* get non-edge with highest value, and vertices not forbidden */
    bestval = -999.0;
    bestv1 = 0;
    bestv2 = 0;
    for (v1 = 1; v1 < vertices; v1++) {
      if (ECOvget(forbidden, v1) == 1) continue;
      degree = ECOgetdegree(ei, v1);
      ECOvfill(noneighbor, 1);
      ECOvset(noneighbor, v1, 0);
      for (k = 1; k <= degree; k++) ECOvset(noneighbor, ECOgetneighbor(ei, v1, k), 0);
      /* now, noneighbor = V - N[v] */
      for (v2 = v1 + 1; v2 <= vertices; v2++) {
        if (ECOvget(forbidden, v2) == 0 && ECOvget(noneighbor, v2) == 1) {
          val = ECOxval(v1, j, ei) + ECOxval(v2, j, ei);
          if (bestval < val) {
            /* update non-edge */
            bestval = val;
            bestv1 = v1;
            bestv2 = v2;
          }
        }
      }
    }

    /* there exist bestv1 and bestv2? */
	if (bestv1 == 0) return ECOERR_OK;

	if (bestval < ECO_WJTOL) return ECOERR_OK;
	cutadded = 0;

    /* start with alpha2 = {v1, v2}, clique = empty */
    ECOvmaxsize(alpha2_set);
    ECOvset(alpha2_set, 1, bestv1);
    ECOvset(alpha2_set, 2, bestv2);
    alpha2_size = 2;
    ECOvmaxsize(alpha2clique_set);
    alpha2clique_size = 0;

    /* compute candidates:
     * candidates1 (for alpha2clique): intersection of N(v1) and N(v2)
     * candidates2 (for alpha2): union of N(v1) and N(v2) */
    ECOvfill(candidates1, 0);
    ECOvfill(candidates2, 0);
	cand2count = 0;

    degree = ECOgetdegree(ei, bestv1);
    for (k = 1; k <= degree; k++) {
      v = ECOgetneighbor(ei, bestv1, k);
      ECOvset(candidates1, v, 1);
      ECOvset(candidates2, v, 1);
	  cand2count++;
    }

    ECOvfill(noneighbor, 1);
    ECOvset(noneighbor, bestv2, 0);
    degree = ECOgetdegree(ei, bestv2);
    for (k = 1; k <= degree; k++) {
      v = ECOgetneighbor(ei, bestv2, k);
      ECOvset(noneighbor, v, 0);
      ECOvset(candidates2, v, 1);
	  cand2count++;
    }
    for (v = 1; v <= vertices; v++) if (ECOvget(noneighbor, v) == 1) ECOvset(candidates1, v, 0);

	/* we need at least 3 candidates in 'candidates2' */
	if (cand2count >= 3) {
	  /* add candidates to the sets in descending order */
      for (i = 1; i <= vertices; i++) {
        v = ECOvget(descending, i);
        if (ECOvget(candidates1, v) == 1) {
          /* v is adyacent to alpha2_set and alpha2clique_set ==> we add it to alpha2clique */
          ECOvset(candidates1, v, 0);
          alpha2clique_size++;
          ECOvset(alpha2clique_set, alpha2clique_size, v);
          /* reduce candidates */
          degree = ECOgetdegree(ei, v);
          ECOvfill(noneighbor, 1);
          ECOvset(noneighbor, v, 0);
          for (k = 1; k <= degree; k++) ECOvset(noneighbor, ECOgetneighbor(ei, v, k), 0);
          for (v1 = 1; v1 <= vertices; v1++) {
            if (ECOvget(noneighbor, v1) == 1) {
              if (ECOvget(candidates1, v1) == 1) ECOvset(candidates1, v1, 0);
              if (ECOvget(candidates2, v1) == 1) ECOvset(candidates2, v1, 0);
            }
          }
        }
        else {
          if (ECOvget(candidates2, v) == 1) {
            /* v is adyacent to alpha2clique_set and alpha({v} cup alpha2_set) = 2 */
            ECOvset(candidates2, v, 0);
            alpha2_size++;
            ECOvset(alpha2_set, alpha2_size, v);
            /* reduce candidates */
            degree = ECOgetdegree(ei, v);
            ECOvfill(noneighbor, 1);
            ECOvset(noneighbor, v, 0);
            for (k = 1; k <= degree; k++) ECOvset(noneighbor, ECOgetneighbor(ei, v, k), 0);
            for (v1 = 1; v1 <= vertices; v1++) {
              if (ECOvget(noneighbor, v1) == 1) {
                if (ECOvget(candidates1, v1) == 1) ECOvset(candidates1, v1, 0);
                if (ECOvget(candidates2, v1) == 1) {
                  /* we are supposing that clqv1set (intersection between alpha2_set and N(v1))
                   * is a clique, so:
                   *   v adyacent to clqv1set ==> v1 is still a valid candidate */
                  degree2 = ECOgetdegree(ei, v1);
                  ECOvfill(noneighborv1, 1);
                  ECOvset(noneighborv1, v1, 0);
                  for (k = 1; k <= degree2; k++) ECOvset(noneighborv1, ECOgetneighbor(ei, v1, k), 0);
                  ECOvfill(clqv1set, 0);
                  for (k = 1; k <= alpha2_size - 1; k++) { /* <-- don't consider vertex v (last appended) */
                    v2 = ECOvget(alpha2_set, k);
                    if (ECOvget(noneighborv1, v2) == 1) ECOvset(clqv1set, v2, 1);
                  }

                  /* the idea is to turn off vertices of clqv1set if v is adyacent to it */
                  degree2 = ECOgetdegree(ei, v);
                  for (k = 1; k <= degree2; k++) {
                    v2 = ECOgetneighbor(ei, v, k);
                    if (ECOvget(clqv1set, v2) == 1) ECOvset(clqv1set, v2, 0);
                  }

                  flag = 0;
                  for (v2 = 1; v2 <= vertices; v2++) flag |= ECOvget(clqv1set, v2);
                  if (flag == 1) ECOvset(candidates2, v1, 0);
                }
              }
            }
          }
        }
      }
      ECOvredim(alpha2_set, alpha2_size);
      ECOvredim(alpha2clique_set, alpha2clique_size);

      /* try to add as cut */
#ifdef ECODEBUGCUTS
      printf("Alpha2_set:");
      ECOshowvector(alpha2_set);
      printf(" and Alpha2_clique_set:");
      ECOshowvector(alpha2clique_set);
#endif
      status = ECOaddalpha2cut(ei, alpha2_set, alpha2clique_set, j, &cutadded);
      if (status) return status;

#ifdef ECODEF_ALPHA2AGGRESIVE
      /* always forbid the first 3 vertices of alpha2_set */
	  ECOvset(forbidden, ECOvget(alpha2_set, 1), 1);
	  ECOvset(forbidden, ECOvget(alpha2_set, 2), 1);
	  if (alpha2_size >= 3) ECOvset(forbidden, ECOvget(alpha2_set, 3), 1);
#else
	  /* always forbid all vertices of alpha2_set */
      for (v = 1; v <= alpha2_size; v++) ECOvset(forbidden, ECOvget(alpha2_set, v), 1);
#endif

	  /* we also test how many times we tried to form a2-rank inequality */
	  if (cutadded) {
		/* reset number of trials */
		tries = 0;
	  }
	  else {
		tries++;
		/* If 5 attempts are reached, not worth more time consuming :( */
		if (tries == 5) return ECOERR_OK;
	  }
    }
    else {
	  /* |candidates2| < 3, forbid bestv1 and bestv2 */
	  ECOvset(forbidden, bestv1, 1);
	  ECOvset(forbidden, bestv2, 1);
	}

    /* can we generate more inequalities? */
    flag = 1;
    if (ei->alpha2_cuts_applied_round < ECO_MAXFAMILYCUTROUND) {
      /* does exists some allowed vertices yet? */
      k = 0;
      for (v = 1; v <= vertices; v++) if (ECOvget(forbidden, v) == 0) k++;
      /* we ask for at least 5 non-forbidden vertices */
      flag = (k >= 5) ? 0 : 1;
    }
    else return ECOERR_OK;
  }

  return ECOERR_OK;
}

int ECOaddalpha2cut(struct ecointernal *ei, ECOVECptr alpha2_set,
                    ECOVECptr alpha2clique_set, int j, int *cutadded)
{
  int status, v, alpha2_size, alpha2clique_size;
  struct ecocut *cu;

  if (ei->alpha2_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK; 

  alpha2_size = ECOvsize(alpha2_set);
  alpha2clique_size = ECOvsize(alpha2clique_set);

  /* we want at least 5 vertices in the set */
  if (alpha2_size < 5) return ECOERR_OK;
  cu = &ei->cu;

  ECObegincut(cu, 'L', 0);
#ifdef ECODEBUGCUTS
  printf("Alpha2 cut:");
#endif
  for (v = 1; v <= alpha2_size; v++) {
    ECOaddxcut(cu, ECOvget(alpha2_set, v), j, 1);
#ifdef ECODEBUGCUTS
    printf(" + x%d_%d", ECOvget(alpha2_set, v), j);
#endif
  }
  for (v = 1; v <= alpha2clique_size; v++) {
    ECOaddxcut(cu, ECOvget(alpha2clique_set, v), j, 2);
#ifdef ECODEBUGCUTS
    printf(" + 2 x%d_%d", ECOvget(alpha2clique_set, v), j);
#endif
  }
  ECOaddwcut(cu, j, -2);
#ifdef ECODEBUGCUTS
  printf(" - 2 w%d <= 0\n", j);
#endif
  status = ECOendcut(cu, cutadded, 2.0 * ECO_VIOLATION, ECO_OFF);
  if (status) return status;

  if (*cutadded) {
/* this line could be useful: printf("VAL = %.4f\n", (ECOxval(ECOvget(alpha2_set, 1), j, ei) + ECOxval(ECOvget(alpha2_set, 2), j, ei))/ECOwval(j,ei)); */
    ei->alpha2_cuts_applied++;
    ei->alpha2_cuts_applied_round++;
  }
  return ECOERR_OK;
}
