
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * branch.c - user-defined branch strategy and implicit enumeration algorithm
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOinitbranchcallback(struct ecointernal *ei)
{
  int status;

#ifdef ECODEBUGBRANCH
  printf("Branching strategy enabled.\n");
#endif

  /* note: also used by the enumeration routine */
  ei->intfeas = malloc(sizeof(int)*ei->variables);
  if (ei->intfeas == NULL) return ECOERR_MEMORY;
  ei->colority = ECOallocvector(&status, ei->max_color);
  if (status) return status;
  ei->varind = malloc(sizeof(int)*ei->variables);
  if (ei->varind == NULL) return ECOERR_MEMORY;
  ei->varlu = malloc(sizeof(int)*ei->variables);
  if (ei->varlu == NULL) return ECOERR_MEMORY;
  ei->varbd = malloc(sizeof(int)*ei->variables);
  if (ei->varbd == NULL) return ECOERR_MEMORY;

  return ECOERR_OK;
}

int ECObranchcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei,
                               int *useraction_p)
{
  int status, v, j, j2, vertices, max_color, index, degree, n, u, mybd, ub, lb;
  int bestimpact, bestsubimpact, bestv, bestj, impact, subimpact, index2, maxj;
  int exists_upper_bound, c;
  double objval;
  ECOVECptr colority;
  int *intfeas, *varind, *varbd;
  char *varlu;

  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  intfeas = ei->intfeas;
  varind = ei->varind;
  varlu = ei->varlu;
  varbd = ei->varbd;
  colority = ei->colority;

  /* check if optimality is reached */
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_REMAINING, &ei->LB_current);
  if (status) return status | ECOERR_CPLEX;
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_INTEGER, &ei->UB_current);
  if (status) return status | ECOERR_CPLEX;
  if (ei->LB_current + ECO_WJTOL > ei->UB_current) return ECOERR_OK;

  /* obtain current fractional solution and infeasible variables */
  status = CPXgetcallbacknodex(cenv, cbdata, wherefrom, ei->x, 0, ei->variables - 1);
  if (status) return status | ECOERR_CPLEX;
  status = CPXgetcallbacknodeintfeas(cenv, cbdata, wherefrom, intfeas, 0, ei->variables - 1);
  if (status) return status | ECOERR_CPLEX;
  status = CPXgetcallbacknodeobjval (cenv, cbdata, wherefrom, &objval);
  if (status) return status | ECOERR_CPLEX;

  exists_upper_bound = 1;
  ub = max_color;
  if ((double)max_color < ei->UB_current - ECO_WJTOL) exists_upper_bound = 0;
  if (exists_upper_bound == 1) ub = (int)(ei->UB_current + ECO_WJTOL);
  lb = (int)(ei->LB_current + 1.0 - ECO_WJTOL);

#ifdef ECODEBUGBRANCH
  printf("Fractional variables:");
#endif
  bestimpact = -1;
  bestsubimpact = -1;
  bestv = -1;
  bestj = -1;
  /* we only check variables between j=1..ub-1 */
  for (j = 1; j <= ub-1; j++) {
    for (v = 1; v <= vertices; v++) {
      index = ECOmget(ei->xvar_state, v, j);
	  if (index >= 2) {
        if (intfeas[index - 2] == CPX_INTEGER_INFEASIBLE) {
          /* variable x_vj is not integer, so compute impact */
#ifdef ECODEF_ABRULE
          /* impact = A, subimpact = B */
       	  impact = 0;
		  degree = ECOgetdegree(ei, v);
		  for (n = 1; n <= degree; n++) {
			u = ECOgetneighbor(ei, v, n);
            index2 = ECOmget(ei->xvar_state, u, j);
			if (index2 >= 2) {
			  if (intfeas[index2 - 2] == CPX_INTEGER_INFEASIBLE) impact++;
			}
		  }
		  if (impact >= bestimpact) {
		    ECOvfill(colority, 0);
		    subimpact = 0;
		    for (n = 1; n <= degree; n++) {
			  u = ECOgetneighbor(ei, v, n);
			  c = 0;
              for (j2 = 1; j2 <= max_color; j2++) {
			    if (ECOxval(u, j2, ei) > 1.0 - ECO_EPSILON) {
				  c = j2;
				  break;
			    }
			  }
			  if (c > 0) {
			    if (ECOvget(colority, c) == 0) {
				  ECOvset(colority, c, 1);
				  subimpact++;
			    }
			  }
		    }
			if (impact > bestimpact || subimpact > bestsubimpact) {
              bestimpact = impact;
			  bestsubimpact = subimpact;
			  bestv = v;
			  bestj = j;
			}
		  }
#else
          /* impact = B, subimpact = A */
		  degree = ECOgetdegree(ei, v);
		  ECOvfill(colority, 0);
		  impact = 0;
		  for (n = 1; n <= degree; n++) {
			u = ECOgetneighbor(ei, v, n);
			c = 0;
            for (j2 = 1; j2 <= max_color; j2++) {
			  if (ECOxval(u, j2, ei) > 1.0 - ECO_EPSILON) {
				c = j2;
				break;
			  }
			}
			if (c > 0) {
			  if (ECOvget(colority, c) == 0) {
				ECOvset(colority, c, 1);
				impact++;
			  }
			}
		  }
		  if (impact >= bestimpact) {
         	subimpact = 0;
		    for (n = 1; n <= degree; n++) {
			  u = ECOgetneighbor(ei, v, n);
              index2 = ECOmget(ei->xvar_state, u, j);
			  if (index2 >= 2) {
			    if (intfeas[index2 - 2] == CPX_INTEGER_INFEASIBLE) subimpact++;
			  }
		    }
			if (impact > bestimpact || subimpact > bestsubimpact) {
              bestimpact = impact;
			  bestsubimpact = subimpact;
			  bestv = v;
			  bestj = j;
			}
		  }
#endif
        }
      }
    }
  }
  if (bestimpact == -1) {
#ifdef ECODEBUGBRANCH
    printf("Fractional variable not found!\n");
#endif
	return ECOERR_OK;
  }

  /* Up node */
  index = ECOmget(ei->xvar_state, bestv, bestj);
  varind[0] = index - 2;
  varlu[0] = 'L';
  varbd[0] = 1;
  mybd = 1;

  degree = ECOgetdegree(ei, bestv);
  for (n = 1; n <= degree; n++) { /* neighbors are set to zero */
    index2 = ECOmget(ei->xvar_state, ECOgetneighbor(ei, bestv, n), bestj);
	if (index2 >= 2) {
	  varind[mybd] = index2 - 2;
      varlu[mybd] = 'U';
      varbd[mybd++] = 0;
	}
  }

  maxj = (bestj > lb ? bestj : lb);
  for (j = 1; j <= maxj; j++) { /* wj with j <= max(lb,bestj) are set to one */
    index2 = ECOvget(ei->wvar_state, j);
    if (index2 >= 2) {
      varind[mybd] = index2 - 2;
      varlu[mybd] = 'L';
      varbd[mybd++] = 1;
	}
  }

  if (exists_upper_bound == 1) {
	for (j = ub; j <= max_color; j++) { /* variables with j >= ub are set to zero */
	  for (v = 1; v <= vertices; v++) {
        index2 = ECOmget(ei->xvar_state, v, j);
	    if (index2 >= 2) {
	      varind[mybd] = index2 - 2;
          varlu[mybd] = 'U';
          varbd[mybd++] = 0;
		}
	  }
      index2 = ECOvget(ei->wvar_state, j);
      if (index2 >= 2) {
	    varind[mybd] = index2 - 2;
        varlu[mybd] = 'U';
        varbd[mybd++] = 0;
	  }
	}
  }

  status = CPXbranchcallbackbranchbds(cenv, cbdata, wherefrom, objval, mybd,
	  varind, varlu, varbd, NULL, &j);
  if (status) return status | ECOERR_CPLEX;
#ifdef ECODEBUGBRANCH
  printf("\nBest: x%d_%d=%.2f impact=%d subimpact=%d\n", bestv, bestj,
	     ECOxval(bestv, bestj, ei), bestimpact, bestsubimpact);
  printf("BDCNT UP: Bounds changed = %d\n", mybd);
#endif
 
  /* Down node */
  varind[0] = index - 2;
  varlu[0] = 'U';
  varbd[0] = 0;
  mybd = 1;

  for (j = 1; j <= lb; j++) { /* wj with j <= lb are set to one */
    index2 = ECOvget(ei->wvar_state, j);
    if (index2 >= 2) {
      varind[mybd] = index2 - 2;
      varlu[mybd] = 'L';
      varbd[mybd++] = 1;
	}
  }

  if (exists_upper_bound == 1) {
	for (j = ub; j <= max_color; j++) { /* variables with j >= ub are set to zero */
	  for (v = 1; v <= vertices; v++) {
        index2 = ECOmget(ei->xvar_state, v, j);
	    if (index2 >= 2) {
	      varind[mybd] = index2 - 2;
          varlu[mybd] = 'U';
          varbd[mybd++] = 0;
		}
	  }
      index2 = ECOvget(ei->wvar_state, j);
      if (index2 >= 2) {
	    varind[mybd] = index2 - 2;
        varlu[mybd] = 'U';
        varbd[mybd++] = 0;
	  }
	}
  }

  status = CPXbranchcallbackbranchbds(cenv, cbdata, wherefrom, objval, mybd,
      varind, varlu, varbd, NULL, &j);
  if (status) return status | ECOERR_CPLEX;
#ifdef ECODEBUGBRANCH
  printf("BDCNT DOWN: Bounds changed = %d\n", mybd);
#endif

  *useraction_p = CPX_CALLBACK_SET;
  return ECOERR_OK;
}

int ECOenumcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei,
                               int *useraction_p)
{
  ECOMTXptr xvjstate;
  ECOVECptr bestcoloring, curcoloring, size_classes;
  double timeconsumed; 

  int status, ub, lb, cur_numcols, surplus, vars;
  int ptr1, ptr2, vertices, max_color, namesize, v, j, k;
  int xvj, vert, degree, uncolored, mybd, index;
  char name[32], *unused[1], number[16];
  char longnumber[32];
  int *intfeas, *varind, *varbd;
  char *varlu;
  CPXLPptr lp;
  double objval, value;

  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  xvjstate = ei->tabu_tenure;
  bestcoloring = ei->tabu_bestcoloring;
  curcoloring = ei->tabu_curcoloring;
  size_classes = ei->tabu_bigclasses;
  varind = ei->varind;
  varlu = ei->varlu;
  varbd = ei->varbd;

  /* check if optimality is reached */
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_REMAINING, &ei->LB_current);
  if (status) return status | ECOERR_CPLEX;
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_INTEGER, &ei->UB_current);
  if (status) return status | ECOERR_CPLEX;
  if (ei->LB_current + ECO_WJTOL > ei->UB_current) return ECOERR_OK;

  /* now, get upper and lower bounds
     if upper bound does not exists, do not execute the enumeration
     otherwise, UB = best_coloring and LB = ceil(objective_value) */
  status = CPXgetcallbacknodeobjval(cenv, cbdata, wherefrom, &objval);
  if (status) return status | ECOERR_CPLEX;
  if ((double)max_color < ei->UB_current - ECO_WJTOL) return ECOERR_OK; 
  ub = (int)(ei->UB_current + ECO_WJTOL);
  lb = (int)(objval + 1.0 - ECO_WJTOL);

  ei->last_t = ei->start_t;
  ei->nodeprocessed = 0;

  /* fill curcoloring and state of variables xvj:
   * if xvj = 0, vertex "v" can not use color "j"
   * if xvj = 1, vertex "v" is already painted with color "j"
   * if xvj = 2, it is not known yet whether vertex "v" uses color "j" or not */
  ECOvfill(curcoloring, 0);
  ECOvfill(size_classes, 0);
  ei->curnumberofcolors = 0;
  ei->isdynamic = ECO_ISDYNAMICVAL;
  if (ei->isdynamic == 0) {
    /* STATIC COLOR SELECTION: we consider how we fix variables in ECOordersymmetry (xvar_state) */
    for (v = 1; v <= vertices; v++) {
      for (j = 1; j <= max_color; j++) {
	    xvj = ECOmget(ei->xvar_state, v, j);
		if (xvj >= 2) xvj = 2;
        ECOmset(xvjstate, v, j, xvj);
	    if (xvj == 1) {
          ECOvset(curcoloring, v, j);
		  if (j > ei->curnumberofcolors) ei->curnumberofcolors = j;
		  ECOvincr(size_classes, j);
	    }
	  }
    }
  }
  else {
    /* DYNAMIC COLOR SELECTION: we only fix the colors for the initial clique */
    for (v = 1; v <= vertices; v++) for (j = 1; j <= max_color; j++) ECOmset(xvjstate, v, j, 2);
    for (v = 1; v <= ECOvsize(ei->clique_set); v++) {
      vert = ECOvget(ei->ordering_symmetry, v);
      /* fix x_vv = 1 for v in the clique: variable "vert" is the vertex, variable "v" is the color */
	  for (j = 1; j <= max_color; j++) ECOmset(xvjstate, vert, j, 0);
      ECOmset(xvjstate, vert, v, 1);
      ECOvset(curcoloring, vert, v);
	  ei->curnumberofcolors++;
	  ECOvincr(size_classes, v);
      /* fix x_uv = 0 for u adjacent to v */
      degree = ECOgetdegree(ei, vert);
      for (k = 1; k <= degree; k++) ECOmset(xvjstate, ECOgetneighbor(ei, vert, k), v, 0);
	}
  }

  /* get bounds of variables from LP node: since this node is presolved, we are forced to see the
   "name" of the variable in order to know which variable is talking about */
  status = CPXgetcallbacknodelp(cenv, cbdata, wherefrom, &lp);
  if (status) return status | ECOERR_CPLEX;
  cur_numcols = CPXgetnumcols (cenv, lp);
  for (vars = 0; vars < cur_numcols; vars++) {
    CPXgetcolname(cenv, lp, unused, name, 0, &namesize, vars, vars);
	namesize = -namesize;
	if (namesize >= 32) ECOabort(ECOERR_INTERNAL);
    CPXgetcolname(cenv, lp, unused, name, 32, &surplus, vars, vars);
	name[namesize] = 0;
    if (name[0] == 'x') {
	  /* obtain "v" and "j" from name */
	  ptr1 = 1;
	  ptr2 = 0;
	  while (name[ptr1] != '_') {
		number[ptr2] = name[ptr1];
	    ptr1++; ptr2++;
		if (ptr2 == 16) ECOabort(ECOERR_INTERNAL);
	  }
	  number[ptr2] = 0;
	  v = atoi(number);
#ifdef ECODEBUGENUM
	  if (v < 1 || v > vertices) ECOabort(ECOERR_INTERNAL);
#endif
	  ptr1++;
	  ptr2 = 0;
	  while (name[ptr1] != 0) {
		number[ptr2] = name[ptr1];
	    ptr1++; ptr2++;
		if (ptr2 == 16) ECOabort(ECOERR_INTERNAL);
	  }
	  number[ptr2] = 0;
	  j = atoi(number);
#ifdef ECODEBUGENUM
	  if (j < 1 || j > max_color) ECOabort(ECOERR_INTERNAL);
#endif

      CPXgetub(cenv, lp, &value, vars, vars);
	  if (value < 0.5) {
        /* upper bound is zero! */
        if (ei->isdynamic == 0) {
		  /* we do not consider color j for v in STATIC SELECTION */
#ifdef ECODEBUGENUM
		  if (ECOmget(xvjstate, v, j) == 1) ECOabort(ECOERR_INTERNAL);
#endif
          ECOmset(xvjstate, v, j, 0);
		}
	  }
      CPXgetlb(cenv, lp, &value, vars, vars);
	  if (value > 0.5) {
	    /* lower bound is one! paint v with j */
#ifdef ECODEBUGENUM
        if (ECOmget(xvjstate, v, j) == 0) ECOabort(ECOERR_INTERNAL);
#endif
		for (k = 1; k <= max_color; k++) ECOmset(xvjstate, v, k, 0);
        ECOmset(xvjstate, v, j, 1);
        ECOvset(curcoloring, v, j);
        degree = ECOgetdegree(ei, v);
        for (k = 1; k <= degree; k++) {
		  vert = ECOgetneighbor(ei, v, k);
#ifdef ECODEBUGENUM
		  if (ECOmget(xvjstate, vert, j) == 1) ECOabort(ECOERR_INTERNAL);
#endif
		  ECOmset(xvjstate, vert, j, 0);
		}
        if (j > ei->curnumberofcolors) ei->curnumberofcolors = j;
        ECOvincr(size_classes, j);
	  }
	}
  }

#ifdef ECODEBUGENUM
  printf("Initial coloring with %d colors:", ei->curnumberofcolors);
  ECOshowvector(curcoloring);
  printf("Size of classes:");
  ECOshowvector(size_classes);
/*  for (v = 1; v <= vertices; v++) {
    printf(" vertex %d are { ", v);
    for (j = 1; j <= max_color; j++) {
      xvj = ECOmget(xvjstate, v, j);
	  if (xvj == 1) printf("%d(unique) ", j);
	  if (xvj == 2) printf("%d ", j);
	}
	printf("}\n");
  }*/
#endif

  /* count number of uncolored vertices */
  uncolored = 0;
  for (v = 1; v <= vertices; v++) if (ECOvget(curcoloring, v) == 0) uncolored++;
#ifdef ECODEBUGENUM
  printf("uncolored = %d,  ub = %d,  product = %d\n", uncolored, ub, uncolored * ub);
#endif

  /* apply criterion */
  if (ei->env->par[ECO_PARAM_IMPVERT] == 0) {
    if (uncolored * ub > ECO_ENUMFACTOR) return ECOERR_OK;
  }
  else {
    if (uncolored > ei->env->par[ECO_PARAM_IMPVERT]) return ECOERR_OK;
  }

  /* fill bestcoloring with garbage and start branching procedure */
#ifdef ECODEBUGENUM
  if (ei->curnumberofcolors > lb) ECOabort(ECOERR_INTERNAL);
#endif
  ei->curnumberofcolors = lb;
  ECOvcopy(ei->grp->coloring, bestcoloring);
  ei->bestnumberofcolors = ub;

  CPXmsg(ei->env->cpxlog, "Starting enumeration with UB=%d, LB=%d (uncolored vertices=%d, time remaining=%.2f).\n", ub, lb, uncolored, ei->enumavailabletime);
  ei->enuminittime = ECOclock();
  status = ECOenumbranch(ei, lb);
  timeconsumed = ECOclock() - ei->enuminittime;
  ei->enumavailabletime -= timeconsumed;
  sprintf(longnumber, "%lld", ei->nodeprocessed);
  CPXmsg(ei->env->cpxlog, "Enumeration completed (%s nodes processed in %.2f seconds): ", longnumber, timeconsumed);
  switch (status) {
    case ECOENUM_TIMEEXHAUSTED: /* time limit reached */
      CPXmsg(ei->env->cpxlog, "time exhausted!\n");
      if (ei->bestnumberofcolors < ub) {
        ECOvcopy(bestcoloring, ei->primal_coloring);
        ei->primal_colors_used = ei->bestnumberofcolors;
        ei->enum_found_sol = 1;
      }
      ei->enumactivated = 0;
      break;      
    case ECOENUM_OK: /* a solution has found */
    case ECOENUM_OPTIMAL: /* a solution of min_colors has found */
      if (ei->bestnumberofcolors < ub) {
        CPXmsg(ei->env->cpxlog, "a %d-eqcol was found.\n", ei->bestnumberofcolors);
        ECOvcopy(bestcoloring, ei->primal_coloring);
        ei->primal_colors_used = ei->bestnumberofcolors;
        ei->enum_found_sol = 1;
      }
      else return ECOERR_INTERNAL;
      *useraction_p = CPX_CALLBACK_SET;
      break;
    case ECOENUM_INFEASIBLE: /* no solutions found */
      CPXmsg(ei->env->cpxlog, "infeasible.\n");
      *useraction_p = CPX_CALLBACK_SET;
      break;
    default:
      return status;
  }
  return ECOERR_OK;
}
