
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * callbacks.c - CPLEX callback routines
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

/************************** BEGIN CPLEX CALLBACKS **************************/

int CPXPUBLIC CPXincumcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, void *cbhandle,
                              double objval, double *x, int *isfeas_p, int *useraction_p)
{
  struct ecointernal *ei;
  int status;
  double start_t;

  /* prologue */
#ifdef ECOPARALLEL
  ei = ECOgetei((struct ecointernal*)cbhandle);
#else
  ei = (struct ecointernal*)cbhandle;
#endif
  *isfeas_p = 1;
  *useraction_p = CPX_CALLBACK_SET;
  start_t = ECOclock();

  /* execute callback */
  status = ECOincumcallback(cenv, cbdata, wherefrom, ei, x, objval);
  if (status) {
    *useraction_p = CPX_CALLBACK_FAIL;
    return status;
  }

  /* epilogue */
  ei->heurcallback_time += ECOclock() - start_t;
  ei->callbacks_made++;
  return ECOERR_OK;
}

int CPXPUBLIC CPXnodecallback(CPXCENVptr cenv, void *cbdata, int wherefrom, void *cbhandle,
                              int *nodeindex_p, int *useraction_p)
{
  struct ecointernal *ei;
  int status;

  /* prologue */
#ifdef ECOPARALLEL
  ei = ECOgetei((struct ecointernal*)cbhandle);
#else
  ei = (struct ecointernal*)cbhandle;
#endif
  *useraction_p = CPX_CALLBACK_DEFAULT;

  /* execute callback */
  status = ECOnodecallback(cenv, cbdata, wherefrom, ei);
  if (status) {
    *useraction_p = CPX_CALLBACK_FAIL;
    return status;
  }

  /* epilogue */
  ei->callbacks_made++;
  return ECOERR_OK;
}

int CPXPUBLIC CPXbranchcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, void *cbhandle, int type, int sos,
        int nodecnt, int bdcnt, const double *nodeest,
        const int *nodebeg, const int *indices,
        const char *lu, const int *bd, int *useraction_p)
{
  struct ecointernal *ei;
  int status;
  double start_t;

  /* prologue */
#ifdef ECOPARALLEL
  ei = ECOgetei((struct ecointernal*)cbhandle);
#else
  ei = (struct ecointernal*)cbhandle;
#endif
  *useraction_p = CPX_CALLBACK_DEFAULT;
  start_t = ECOclock();

  /* execute callback only if CPLEX branchs */
  if (nodecnt == 2) {
    /* First, try to execute the implicit enumeration */
    if (ei->enumactivated == 1) {
      status = ECOenumcallback(cenv, cbdata, wherefrom, ei, useraction_p);
      if (status) {
        *useraction_p = CPX_CALLBACK_FAIL;
        return status;
      }
	}
	/* If enumeration does not occur, then execute branching strategy */
	if (*useraction_p == CPX_CALLBACK_DEFAULT && ei->env->par[ECO_PARAM_BRANCH] == ECO_BRANCH_CALLBACK) {
      status = ECObranchcallback(cenv, cbdata, wherefrom, ei, useraction_p);
      if (status) {
        *useraction_p = CPX_CALLBACK_FAIL;
        return status;
      }
    }
  }

  /* epilogue */
  ei->branchcallback_time += ECOclock() - start_t;
  ei->callbacks_made++;
  return ECOERR_OK;
}

int CPXPUBLIC CPXheurcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, void *cbhandle,
                              double *objval_p, double *x, int *checkfeas_p, int *useraction_p)
{
  struct ecointernal *ei;
  int status;
  double start_t;

  /* prologue */
#ifdef ECOPARALLEL
  ei = ECOgetei((struct ecointernal*)cbhandle);
#else
  ei = (struct ecointernal*)cbhandle;
#endif
  *useraction_p = CPX_CALLBACK_DEFAULT;
#ifdef ECODEF_PRIMAL_NOCHECKFEAS
  *checkfeas_p = 0;
#else
  *checkfeas_p = 1;
#endif
  start_t = ECOclock();

  /* execute callback */
  status = ECOheurcallback(cenv, cbdata, wherefrom, ei, objval_p, x, useraction_p);
  if (status) {
    *useraction_p = CPX_CALLBACK_FAIL;
#ifndef ECODEF_EXITAFTERROOT
    return status;
#endif
  }

  /* epilogue */
  ei->heurcallback_time += ECOclock() - start_t;
  if (ei->heurcallback_time >= (double)ei->env->par[ECO_PARAM_MAXTIMEHEUR] && ei->primal_enabled != 0) {
    CPXmsg(ei->env->cpxlog, "Time for primal heuristic exhausted.\n");
    ei->primal_enabled = 0;
  }
  ei->callbacks_made++;
  return ECOERR_OK;
}

int CPXPUBLIC CPXcutcallback(CPXCENVptr cenv, void *cbdata, int wherefrom,
                             void *cbhandle, int *useraction_p)
{
  struct ecointernal *ei;
  int status;
  double start_t;

  /* prologue */
#ifdef ECOPARALLEL
  ei = ECOgetei((struct ecointernal*)cbhandle);
#else
  ei = (struct ecointernal*)cbhandle;
#endif
  *useraction_p = CPX_CALLBACK_DEFAULT;
  start_t = ECOclock();

  /* execute callback */
  status = ECOcutcallback(cenv, cbdata, wherefrom, ei, useraction_p);
  if (status) {
    *useraction_p = CPX_CALLBACK_FAIL;
    return status;
  }

  /* epilogue */
  ei->cutcallback_time += ECOclock() - start_t;
  ei->callbacks_made++;
  return ECOERR_OK;
}

/*************************** END CPLEX CALLBACKS ***************************/

int ECOnodecallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei)
{
  int status, node;

#ifdef ECODEF_WRITENODEFILE
  char lpfile[ECO_MAXFILENAMESIZE];
  CPXLPptr lp;
#endif

  /* obtain node count */
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_NODE_COUNT, &node);
  if (status) return status | ECOERR_CPLEX;
#ifdef ECODEBUGCUTS
  printf("ECOnodecallback: current node = %d\n", node);
#endif

#ifdef ECODEF_WRITENODEFILE
  status = CPXgetcallbacknodelp(cenv, cbdata, wherefrom, &lp);
  if (status) return status | ECOERR_CPLEX;
  sprintf(lpfile, "%s_%d.lp", ei->env->filename, node);
  CPXwriteprob(cenv, lp, lpfile, NULL);
  printf("Formulation written in %s.\n", lpfile);
#endif

  if (node > 0 && ei->isroot == 1) {
    /* we start the branch & bound process */
    CPXmsg(ei->env->cpxlog, "Root node stage finished.\n");
    ei->isroot = 0;
#ifdef ECODEF_EXITAFTERROOT
    return 1;
#endif
	if (ei->env->par[ECO_PARAM_MAXROUNDSATNODE] == 0) {
	  /* A Cut-and-Branch is requested. */
	  ei->cut_enabled = 0;
	}
    ECOshowcutlog(ei);
  }

  if (ei->isroot == 0) {
    /* enables cuts sometimes */
	if (node % ei->env->par[ECO_PARAM_SKIPFACTORCUT] == 0) {ei->cut_enabled_node = 1;}
    else ei->cut_enabled_node = 0;
    /* enables heuristic sometimes */
    if (node % ECO_SKIPFACTORHEUR == 0 && !(ei->env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_ONLYROOT)) ei->primal_enabled_node = 1;
    else ei->primal_enabled_node = 0;
    /* log number of cuts */
    if (ei->cut_enabled == 1 && node % 1000 == 0) ECOshowcutlog(ei);
  }

  return ECOERR_OK;
}

int ECOcallbackinit(struct ecointernal *ei) {
  int status;
  CPXENVptr cplexenv;
#ifdef ECODEF_WRITEROOTFILE
  char rootname[ECO_MAXFILENAMESIZE];
#endif

  cplexenv = ei->env->cplexenv;
  CPXsetintparam(cplexenv, CPX_PARAM_PRELINEAR, CPX_OFF);
  CPXsetintparam(cplexenv, CPX_PARAM_REDUCE, CPX_PREREDUCE_PRIMALONLY);
  CPXsetintparam(cplexenv, CPX_PARAM_MIPCBREDLP, CPX_OFF);
  CPXsetnodecallbackfunc(cplexenv, CPXnodecallback, ei);
  ei->isroot = 1;
  ei->callbacks_made = 0;
  ei->branchcallback_time = 0;
  ei->cutcallback_time = 0;
  ei->heurcallback_time = 0;

  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_ON) {
    CPXsetcutcallbackfunc(cplexenv, CPXcutcallback, ei);
    status = ECOinitcutcallback(ei);
    if (status) return status;
  }

  if (ei->env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_ON) {
    ei->enum_found_sol = 0;
    CPXsetincumbentcallbackfunc(cplexenv, CPXincumcallback, ei);
    CPXsetheuristiccallbackfunc(cplexenv, CPXheurcallback, ei);
    status = ECOinitheurcallback(ei);
    if (status) return status;
  }

  if (ei->env->par[ECO_PARAM_BRANCH] == ECO_BRANCH_CALLBACK
	  || ei->env->par[ECO_PARAM_IMPENUM] == ECO_ON) {
    CPXsetbranchcallbackfunc(cplexenv, CPXbranchcallback, ei);
    status = ECOinitbranchcallback(ei);
    if (status) return status;
  }

#ifdef ECODEF_WRITEROOTFILE
  strcpy(rootname, ei->env->filename);
  strcat(rootname, ".roo");
  ei->rootstream = fopen(rootname, "wt");
  if (ei->rootstream == NULL) return ECOERR_CREAT;
  fprintf(ei->rootstream, "Rounds  Previous obj  Cuts\n");
#endif

#ifdef ECOPARALLEL
  status = ECOcreateeis(ei);
  if (status) return status;
#endif

  return ECOERR_OK;
}
