
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * cliquecut.c - clique cutting-plane algorithm
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOinitcliquecallback(struct ecointernal *ei)
{
  int vertices, status;

  vertices = ei->grp->vertices;
  ei->clique_cut_set = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->clique_ujks = ECOallocvector(&status, vertices);
  if (status) return status;

  ei->clique_cuts_applied = 0;
  ei->ujks_cuts_applied = 0;
  return ECOERR_OK;
}

int ECOcliqueseparation(struct ecointernal *ei, int j)
{
  int status, v, i, vertices, counter;
  int flag, clique_size, cutadded, firstv, max_color;
  ECOVECptr descending, forbidden;
  ECOVECptr clique_set, forbidden1, forbidden2;
  double xvj;
  int u, lowerk, upperk, ceilk, min_color, degree, k, ujks_active;
  ECOVECptr outside_clique;

  /* set variables */
  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  min_color = (int)(ei->objective + 1.0 - ECO_WJTOL);
  descending = ei->cut_descending;
  forbidden = ei->cut_forbidden;
  clique_set = ei->clique_cut_set;
  forbidden1 = ei->cut_temp1;
  forbidden2 = ei->cut_temp2;
  outside_clique = ei->clique_ujks;

  if (ei->clique_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;

  ujks_active = 0;
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_UJKS) ujks_active = 1;

  /* check if all vertices are forbidden (here, we do not know if 'descending' is already defined) */
  flag = 1;
  for (v = 1; v <= vertices; v++) {
    if (ECOvget(forbidden, v) == 0) {
	  flag = 0;
	  break;
	}
  }
  if (flag) return ECOERR_OK;

  ECOvcopy(forbidden, forbidden1);

  for (;;) {
    /* find the first non-forbidden vertex, according to 'descending' */
    firstv = -1;
    for (i = 1; i <= vertices; i++) {
      v = ECOvget(descending, i);
      if (ECOvget(forbidden1, v) == 0) {
        firstv = v;
        break;
      }
    }

    /* if all vertices are forbidden, or xvj == 0, exit */
    if (firstv == -1) return ECOERR_OK;
    if (ECOxval(firstv, j, ei) < ECO_EPSILON) return ECOERR_OK;

    ECOvcopy(forbidden1, forbidden2);
    flag = 0;
    counter = 0;

    do {
      ECOvredim(clique_set, 1);
      ECOvset(clique_set, 1, firstv);

      /* find a clique with at least 3 vertices */
      status = ECOgetclique(ei, descending, forbidden1, clique_set, 0);
      if (status) return status;
      clique_size = ECOvsize(clique_set);
#ifdef ECODEBUGCUTS
      printf("Clique candidate found:");
      ECOshowvector(clique_set);
#endif

      if (clique_size == 1) {
        /* the clique has only the first vertex, that means all its neighborhood are marked as forbidden, so exit */
        flag = 1;
      }
      else {
        /* is it maximal? or it does need to add forbidden vertices to be maximal? */
        status = ECOgetclique(ei, descending, forbidden1, clique_set, 1);
        if (status) return status;
        if (ECOvsize(clique_set) == clique_size) {
          /* if it is maximal and has at least 3 vertices (in the case that adjacency constraints are used instead of neighborhood constraints), try to add as cut */
          if (clique_size >= (ei->adyac_substitutes==1 ? 2 : 3)) {
#ifdef ECODEBUGCUTS
            printf("Clique proposed:");
            ECOshowvector(clique_set);
#endif
            status = ECOaddcliquecut(ei, clique_set, j, &cutadded);
            if (status) return status;

            if (cutadded) {
			  /* forbid last vertex if the cut is added */
              ECOvset(forbidden1, ECOvget(clique_set, clique_size), 1);
#ifdef ECODEF_ALPHA2NOFORBIDCLIQUE
			  /* also, forbid the first and the third vertex of the clique in 'forbidden' (has impact over 2-rank inequalities) */
			  ECOvset(forbidden, ECOvget(clique_set, 1), 1);
			  if (clique_size >= 3) ECOvset(forbidden, ECOvget(clique_set, 3), 1);
#else
			  /* also, forbid all vertices of the clique in 'forbidden' (has impact over 2-rank inequalities) */
			  for (i = 1; i <= clique_size; i++) ECOvset(forbidden, ECOvget(clique_set, i), 1);
#endif
              if (ei->clique_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;
            }
		  }
/*********************** begin (u,j,k,S)-clique-neighborhood cut separation ***********************/
          if (ujks_active) {
            /* if the clique cut is not violated, we try to add a (u,j,k,Q)-clique-neighborhood cut */
            if (ei->ujks_cuts_applied_round < ECO_MAXFAMILYCUTROUND) {
              ECOvfill(outside_clique, 1);
              for (i = 1; i <= clique_size; i++) {
                v = ECOvget(clique_set, i);
                ECOvset(outside_clique, v, 0);
                degree = ECOgetdegree(ei, v);
                for (k = 1; k <= degree; k++) ECOvset(outside_clique, ECOgetneighbor(ei, v, k), 0);
              }
              /* now, "outside_clique" has the vertices that are not adjacent to the clique */
#ifdef ECODEBUGCUTS
              printf("Vertices not adjacent to the clique:");
              for (v = 1; v <= vertices; v++) if (ECOvget(outside_clique, v) == 1) printf(" %d", v);
		      printf("\n");
#endif
	          /* we attempt to add cuts: we search over "u" in outside_clique such that alpha(N(u)) >= 2 and 0 < xuj < 1
                   * we also ask for u such that |N(u)| >= 3 in order to have a support different from the clique ineq. */
		      for (u = 1; u <= vertices; u++) {
                if (ECOvget(outside_clique, u) == 1 && ECOvget(ei->neighbormaxindep1_size, u) >= 2 && ECOgetdegree(ei, u) >= 3) {
                  xvj = ECOxval(u, j, ei);
                  if (xvj > ECO_EPSILON && xvj < 1.0 - ECO_EPSILON) {
                    /* we scan values of k such that max{ceil(n/UB),3} <= ceil(n/k) <= min{ceil(n/LB),ceil(n/j),alpha(N(u))+1} */
                    lowerk = ECOceil(vertices, max_color);
			        if (lowerk < 3) lowerk = 3;
			        upperk = ECOceil(vertices, min_color);
			        if (upperk > ECOceil(vertices, j)) upperk = ECOceil(vertices, j);
			        if (upperk > ECOvget(ei->neighbormaxindep2_size, u) + 1) upperk = ECOvget(ei->neighbormaxindep2_size, u) + 1;
                    for (ceilk = lowerk; ceilk <= upperk; ceilk++) {
                      k = ECOceil(vertices, ceilk - 1) - 1;
		      if (ECOceil(vertices, k) == ceilk) {
                        status = ECOaddujkscut(ei, u, j, k, clique_set, &cutadded);
                        if (status) return status;
                      }
                    }
                  }
                }
              }
            }
          }
/*********************** end (u,j,k,S)-clique-neighborhood cut separation ***********************/
        }
#ifdef ECODEBUGCUTS
        else {
          printf("oops... may be maximalized:");
          ECOshowvector(clique_set);
        }
#endif
        /* at this point, the clique has at least 2 vertices, so forbid the second one
         (let us notice that this vertex is a neighbor of the first vertex) */
        ECOvset(forbidden1, ECOvget(clique_set, 2), 1);
      }
#ifdef ECODEBUGCUTS
      else {
        printf("oops... too small to be used :(\n");
      }
#endif

      counter++;
    } while (flag == 0 && counter < 3); /* <-- exit if the neighborhood of "firstv" is forbidden, or we checked 3 cliques including "firstv" */

    /* once we generated cliques with the first vertex, we forbid it */
    ECOvcopy(forbidden2, forbidden1);
    ECOvset(forbidden1, firstv, 1);
  }

  return ECOERR_OK;
}

int ECOaddcliquecut(struct ecointernal *ei, ECOVECptr clique_set, int j, int *cutadded)
{
  int status, v, clique_size;
  struct ecocut *cu;

  if (ei->clique_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK; 

  clique_size = ECOvsize(clique_set);
  cu = &ei->cu;

  ECObegincut(cu, 'L', 0);
#ifdef ECODEBUGCUTS
  printf("Clique cut:");
#endif
  for (v = 1; v <= clique_size; v++) {
    ECOaddxcut(cu, ECOvget(clique_set, v), j, 1);
#ifdef ECODEBUGCUTS
    printf(" + x%d_%d", ECOvget(clique_set, v), j);
#endif
  }
  ECOaddwcut(cu, j, -1);
#ifdef ECODEBUGCUTS
  printf(" - w%d <= 0\n", j);
#endif
  status = ECOendcut(cu, cutadded, ECO_VIOLATION, ECO_OFF);
  if (status) return status;

  if (*cutadded) {
    ei->clique_cuts_applied++;
    ei->clique_cuts_applied_round++;
  }
  return ECOERR_OK;
}
int ECObuklval(struct ecointernal *ei, int u, int k, int l)
{
  int b, alpha;

  if (l >= k) return ECOceil(ei->grp->vertices, k);

  b = ECOceil(ei->grp->vertices, l);
  alpha = ECOvget(ei->neighbormaxindep2_size, u);
  if (b > alpha + 1) b = alpha + 1;
  return b;
}

int ECOaddujkscut(struct ecointernal *ei, int u, int j, int k, ECOVECptr clique_set, int *cutadded)
{
  int status, v, l, clique_size, degree, ceilk, vertices;
  int max_color, coef;
  struct ecocut *cu;

  if (ei->ujks_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK; 

  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  ceilk = ECOceil(vertices, k);
  clique_size = ECOvsize(clique_set);
  cu = &ei->cu;

  ECObegincut(cu, 'L', 0);
  ECOaddxcut(cu, u, j, ceilk - 1);
#ifdef ECODEBUGCUTS
  printf("(%d,%d,%d,Q)-clique-neighborhood cut: %d x%d_%d", u, j, k, ceilk - 1, u, j);
#endif
  degree = ECOgetdegree(ei, u);
  for (v = 1; v <= degree; v++) {
    ECOaddxcut(cu, ECOgetneighbor(ei, u, v), j, 1);
#ifdef ECODEBUGCUTS
    printf(" + x%d_%d", ECOgetneighbor(ei, u, v), j);
#endif
  }

  for (v = 1; v <= clique_size; v++) {
    ECOaddxcut(cu, ECOvget(clique_set, v), j, 1);
#ifdef ECODEBUGCUTS
    printf(" + x%d_%d", ECOvget(clique_set, v), j);
#endif
  }

  for (l = k + 1; l <= max_color; l++) {
    coef = ceilk - ECOceil(vertices, l);
    if (coef != 0) {
      ECOaddxcut(cu, u, l, coef);
#ifdef ECODEBUGCUTS
      printf(" + %d x%d_%d", coef, u, l);
#endif
    }
  }

  for (l = j; l <= max_color; l++) {
    coef = ECObuklval(ei, u, k, l);
    if (l > j) coef -= ECObuklval(ei, u, k, l-1);
    if (coef != 0) {
      ECOaddwcut(cu, l, -coef);
#ifdef ECODEBUGCUTS
      printf(" + %d w%d", -coef, l);
#endif
    }
  }
#ifdef ECODEBUGCUTS
  printf(" <= 0\n");
#endif
  status = ECOendcut(cu, cutadded, ECO_VIOLATION / 10.0, ECO_ON); /* ten times less than clique inequalities */
  if (status) return status;

  if (*cutadded) {
    ei->ujks_cuts_applied++;
	ei->ujks_cuts_applied_round++;
  }
  return ECOERR_OK;
}
