
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * cscut.c - S-color ineq cutting-plane algorithm
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>

int ECOinitcscallback(struct ecointernal *ei)
{
  int vertices, status;

  vertices = ei->grp->vertices;
  ei->cs_cut_index = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cs_cut_value = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cs_cut_descending = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cs_cut_c_set = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cs_cut_s_set = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cs_cut_forbidden = ECOallocvector(&status, vertices);
  if (status) return status;

  ei->cs_cuts_applied = 0;

  return ECOERR_OK;
}

int beta(int n, int k, ECOVECptr c_set)
{
  int j, dk, r, c_size;

  /* D_k = C cap {1, ..., k}, dk = |D_k| */
  c_size = ECOvsize(c_set);
  dk = 0;
  for (j = 1; j <= c_size; j++) if (ECOvget(c_set, j) <= k) dk++;
  if (dk == 0) return 0;

  /* Compute beta = floor(n/k)|D_k| + min(|D_k|, n%k) */
  r = n % k;
  if (r > dk) r = dk;
  return ECOfloor(n, k) * dk + r;
}

int ECOcsseparation(struct ecointernal *ei)
{
  int status, vertices, first_j_frac, last_j_frac, flag;
  int j, k, minimal_size, fracvars, v, s_size, coef, cutadded, color;
  ECOVECptr index, value, descending, s_set, best_s_set, forbidden;
  double xvj, sum, infact, best_infact;

  /* set variables */
  vertices = ei->grp->vertices;
  first_j_frac= ei->first_j_frac;
  last_j_frac= ei->last_j_frac;
  index = ei->cs_cut_index;
  value = ei->cs_cut_value;
  descending = ei->cs_cut_descending;
  s_set = ei->cs_cut_c_set;
  best_s_set = ei->cs_cut_s_set;
  forbidden = ei->cs_cut_forbidden;

  ei->cs_cuts_applied_round = 0;

  /* in order to continue, we need at least one fractional variable wj */
  if (ei->is_wj_integer) return ECOERR_OK;

  /* then, we compute the minimal size of S, i.e. 1 + min_{k / w_k frac.} {n mod k} */
//  minimal_size = vertices % first_j_frac;
//  for (k = first_j_frac + 1; k <= last_j_frac; k++) if (minimal_size > vertices % k) minimal_size = vertices % k;
  minimal_size = last_j_frac - 3;
  for (k = 1; k <= last_j_frac - 1; k++) if (vertices % k >= 1 && minimal_size > vertices % k) minimal_size = vertices % k;
  minimal_size++;
  if (minimal_size <= 1) minimal_size = 2; /* <-- we need at least 2 color classes */
#ifdef ECODEBUGCUTS
  printf("minimal_size = %d, maximal_size = %d\n", minimal_size, last_j_frac-2);
#endif
  if (minimal_size > last_j_frac - 2) return ECOERR_OK;

  /* generates an array of "fractional"-ness of each class */
  flag = 0;
  ECOvredim(index, last_j_frac);
  ECOvredim(value, last_j_frac);
  ECOvredim(descending, last_j_frac);
  for (j = 1; j <= last_j_frac; j++) {
    fracvars = 0;
    for (v = 1; v <= vertices; v++) {
      xvj = ECOxval(v, j, ei);
      if (xvj > ECO_EPSILON && xvj < 1.0 - ECO_EPSILON) fracvars++;
    }
#ifdef ECODEBUGCUTS
    printf("Class: %d  -  FracVars: %d\n", j, fracvars);
#endif
    ECOvset(index, j, j);
    ECOvset(value, j, fracvars);
  }

  /* sort the arrays */
  status = ECOsort(index, value, descending, ECO_DESCENDING);
  if (status) return status;
#ifdef ECODEBUGCUTS
  printf("Color order:"); ECOshowvector(descending);
#endif

  ECOvredim(forbidden, last_j_frac);
  ECOvfill(forbidden, 0);

  do {
    /* compute a color set of at least "minimal_size" elements */
    flag = 0;
    s_size = 0;
    j = 1;
    sum = 0.0;
    do {
      color = ECOvget(descending, j);
      if (ECOvget(forbidden, color) == 0) {
        s_size++;
        ECOvredim(best_s_set, s_size);
        ECOvset(best_s_set, s_size, color);
        for (v = 1; v <= vertices; v++) sum += ECOxval(v, color, ei);
      }
      j++;
    } while (s_size < minimal_size && j <= last_j_frac);

    if (s_size == minimal_size) {
      /* we compute violation of first inequality */
      best_infact = sum;
      for (k = 1; k <= last_j_frac; k++) {
        if (k == 1) coef = beta(vertices, k, best_s_set);
        else coef = beta(vertices, k, best_s_set) - beta(vertices, k - 1, best_s_set);
        best_infact -= (double)coef * ECOwval(k, ei);
      }
#ifdef ECODEBUGCUTS
      printf("Initial color set (sum = %.4f, infact=%.4f):", sum, best_infact);
      ECOshowvector(best_s_set);
#endif
      if (j <= last_j_frac) {
        /* we scan for better S sets */
        ECOvredim(s_set, s_size);
        ECOvcopy(best_s_set, s_set);
        do {
          color = ECOvget(descending, j);
          if (ECOvget(forbidden, color) == 0) {
            s_size++;
            ECOvredim(s_set, s_size);
            ECOvset(s_set, s_size, color);
            for (v = 1; v <= vertices; v++) sum += ECOxval(v, color, ei);
            infact = sum;
            for (k = 1; k <= last_j_frac; k++) {
              if (k == 1) coef = beta(vertices, k, s_set);
              else coef = beta(vertices, k, s_set) - beta(vertices, k - 1, s_set);
              infact -= (double)coef * ECOwval(k, ei);
            }
#ifdef ECODEBUGCUTS
            printf("Other color set (sum = %.4f, infact=%.4f):", sum, infact);
            ECOshowvector(s_set);
#endif
            if (infact > best_infact + ECO_EPSILON) {
              /* update S set */
              ECOvredim(best_s_set, s_size);
              ECOvcopy(s_set, best_s_set);
              best_infact = infact;
            }
          }
          j++;
        } while (s_size < last_j_frac - 2 && j <= last_j_frac);
      }
      
      /* at this point, we have the best S set, so we attemp to add it as cut */
      if (best_infact > ECO_EPSILON) {
#ifdef ECODEBUGCUTS
        printf("Color set selected (infact=%.4f):", best_infact);
        ECOshowvector(best_s_set);
#endif
        status = ECOaddcscut(ei, best_s_set, &cutadded);
        if (status) return status;
        if (ei->cs_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;
      }

      /* we also forbid the last color */
      ECOvset(forbidden, ECOvget(best_s_set, ECOvsize(best_s_set)), 1);
    }
    else flag = 1; /* <-- we couldn't reach a color set of "minimal_size" elements */
  } while (flag == 0);

  return ECOERR_OK;
}

int ECOaddcscut(struct ecointernal *ei, ECOVECptr s_set, int *cutadded)
{
  int status, v, j, k, max_color, coef, vertices, s_size;
  struct ecocut *cu;

  if (ei->cs_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;

  cu = &ei->cu;
  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  s_size = ECOvsize(s_set);

  ECObegincut(cu, 'L', 0);
#ifdef ECODEBUGCUTS
  printf("S-color cut:");
#endif
  for (v = 1; v <= vertices; v++) {
    for (j = 1; j <= s_size; j++) {
      ECOaddxcut(cu, v, ECOvget(s_set, j), 1);
#ifdef ECODEBUGCUTS
      printf(" + x%d_%d", v, ECOvget(s_set, j));
#endif
    }
  }
  for (k = 1; k <= max_color; k++) {
    if (k == 1) coef = beta(vertices, k, s_set);
    else coef = beta(vertices, k, s_set) - beta(vertices, k - 1, s_set);
    if (coef != 0) {
      ECOaddwcut(cu, k, -coef);
#ifdef ECODEBUGCUTS
      printf(" + %d w%d", -coef, k);
#endif
    }
  }
#ifdef ECODEBUGCUTS
  printf(" <= 0\n");
#endif
  status = ECOendcut(cu, cutadded, ECO_VIOLATION / 10.0, ECO_OFF); /* ten times less than clique inequalities */
  if (status) return status;

  if (*cutadded) {
    ei->cs_cuts_applied++;
    ei->cs_cuts_applied_round++;
  }
  return ECOERR_OK;
}
