
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * cuts.c - cuts management routines
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOinitcutcallback(struct ecointernal *ei)
{
  int status, vertices;

  vertices = ei->grp->vertices;

  /* ask for memory */
  ei->cut_temp1 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cut_temp2 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cut_descending = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cut_forbidden = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->cu.cutind = malloc(sizeof(int)*ei->variables);
  if (ei->cu.cutind == NULL) return ECOERR_MEMORY;
  ei->cu.cutval = malloc(sizeof(double)*ei->variables);
  if (ei->cu.cutval == NULL) return ECOERR_MEMORY;

  ei->cu.ei = ei;
  ei->cut_enabled = 1;
  ei->cut_enabled_node = 1;
  ei->round = 0;
  ei->cuts_applied = 0;
  ei->cut_executed = 0;

  CPXmsg(ei->env->cpxlog, "Cutting-plane routines enabled: ");
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_CLIQUE) {
    CPXmsg(ei->env->cpxlog, "clique ");
    status = ECOinitcliquecallback(ei);
    if (status) return status;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_ALPHA2) {
    CPXmsg(ei->env->cpxlog, "2-rank ");
    status = ECOinitalpha2callback(ei);
    if (status) return status;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_BLOCK) {
    CPXmsg(ei->env->cpxlog, "block ");
    ei->block_cuts_applied = 0;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_NEIGHBOR) {
    CPXmsg(ei->env->cpxlog, "neighborhood ");
    ei->neighborhood_cuts_applied = 0;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_OUTSIDENEIGHBOR) {
    CPXmsg(ei->env->cpxlog, "outside-neighborhood ");
    ei->outsideneighbor_cuts_applied = 0;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_CS) {
    CPXmsg(ei->env->cpxlog, "S-color ");
    status = ECOinitcscallback(ei);
    if (status) return status;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_UJKS) CPXmsg(ei->env->cpxlog, "clique-neighborhood ");
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_EXTRA) {
    CPXmsg(ei->env->cpxlog, "extrasym ");
    ei->extra_cuts_applied = 0;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_CPLEX) CPXmsg(ei->env->cpxlog, "cplex ");
  CPXmsg(ei->env->cpxlog, "\n");

  return ECOERR_OK;
}

/* these routine fills 'forbidden' and 'descending' vectors:
   warning: if all vertices are forbidden for a given j, 'descending' is not defined! */
int ECOcutordering(struct ecointernal *ei, int j)
{
  int v, i, vertices, flag, status;
  double xvj;
  ECOVECptr index, value, descending, forbidden;

  vertices = ei->grp->vertices;
  index = ei->cut_temp1;
  value = ei->cut_temp2;
  descending = ei->cut_descending;
  forbidden = ei->cut_forbidden;

  /* fill the arrays 'index' and 'value' */
  flag = 0;
  for (v = 1; v <= vertices; v++) {
    xvj = ECOxval(v, j, ei);
    if (xvj > ECO_EPSILON && xvj < 1.0 - ECO_EPSILON) flag = 1;
    ECOvset(index, v, v);
    ECOvset(value, v, (int)(xvj / ECO_EPSILON));
  }

  /* fill 'forbidden' (every vertex is forbidden) */
  ECOvfill(forbidden, 1);

  /* check if the solution is integral: in that case, all vertices remains forbiden */
  if (flag == 0) return ECOERR_OK;

  /* sort the arrays */
  status = ECOsort(index, value, descending, ECO_DESCENDING);
  if (status) return status;
#ifdef ECODEBUGCUTS
    printf("Vertex ordering:");
#endif
  for (i = 1; i <= vertices; i++) {
    v = ECOvget(descending, i);
    xvj = ECOxval(v, j, ei);
    /* we left forbidden v such that x_vj = 1 or d(v) <= 1:
     * --> if x_vj = 1, the clique/2-rank cut does not cut off the solution
     * --> if d(v) <= 1, "v" is an isolated vertex, or is the end of an edge
     */
    if (xvj < 1.0 - ECO_EPSILON && ECOgetdegree(ei, v) >= 2) {
      ECOvset(forbidden, v, 0);
#ifdef ECODEBUGCUTS
      printf(" %d", v);
#endif
    }
  }
#ifdef ECODEBUGCUTS
  printf("\n");
#endif

  return ECOERR_OK;
}

void ECOshowcutlog(struct ecointernal *ei)
{
  /* log number of cuts */
  CPXmsg(ei->env->cpxlog, "Total cuts applied: %d\n", ei->cuts_applied);
  CPXmsg(ei->env->cpxlog, "Time consumed by callback routines: %.2f seg. (branch=%.2f, cut=%.2f, heur=%.2f)\n",
		     ei->branchcallback_time + ei->cutcallback_time + ei->heurcallback_time,
			 ei->branchcallback_time, ei->cutcallback_time, ei->heurcallback_time);
  if (ei->cuts_applied == 0) return;
  CPXmsg(ei->env->cpxlog, "Details of the cuts applied:\n");
  if (ei->clique_cuts_applied > 0)
        CPXmsg(ei->env->cpxlog, " *) Clique cuts applied: %d\n", ei->clique_cuts_applied);
  if (ei->alpha2_cuts_applied > 0)
        CPXmsg(ei->env->cpxlog, " *) 2-rank cuts applied: %d\n", ei->alpha2_cuts_applied);
  if (ei->block_cuts_applied > 0)
        CPXmsg(ei->env->cpxlog, " *) Block color cuts applied: %d\n", ei->block_cuts_applied);
  if (ei->neighborhood_cuts_applied > 0)
        CPXmsg(ei->env->cpxlog, " *) Neighborhood cuts applied: %d\n", ei->neighborhood_cuts_applied);
  if (ei->outsideneighbor_cuts_applied > 0)
        CPXmsg(ei->env->cpxlog, " *) Outside-neighborhood cuts applied: %d\n", ei->outsideneighbor_cuts_applied);
  if (ei->ujks_cuts_applied > 0)
        CPXmsg(ei->env->cpxlog, " *) ujks-clique-neighborhood cuts applied: %d\n", ei->ujks_cuts_applied);
  if (ei->cs_cuts_applied > 0)
        CPXmsg(ei->env->cpxlog, " *) S-color cuts applied: %d\n", ei->cs_cuts_applied);
  if (ei->extra_cuts_applied > 0)
        CPXmsg(ei->env->cpxlog, " *) Extra symmetry breakers applied: %d\n", ei->extra_cuts_applied);
}

int ECOcutcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei,
                   int *useraction_p)
{
  int status, j, depth, max_color, first_j_frac, last_j_frac;
  double wj;
#ifdef ECODEBUGCUTS
  int vertices, v;
#endif
#ifdef ECODEF_WRITEROOTFILE
  double objval;
#endif
#ifdef ECODEF_EXITAFTERROOT
  double objvalroot;
  int newobjvalroot;
#endif

  ei->cuts_applied_round = 0;
  ei->cu.cenv = cenv;
  ei->cu.cbdata = cbdata;
  ei->cu.wherefrom = wherefrom;
  ei->cu.useraction_p = useraction_p;

  /* obtain current fractional solution and bounds of chi_eq:
           1) UB_current = colors of best integer solution found
           2) LB_current = best lower bound of the tree
           3) objective = sum of w*_j */
  status = CPXgetcallbacknodex(cenv, cbdata, wherefrom, ei->x, 0, ei->variables - 1);
  if (status) return status | ECOERR_CPLEX;
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_REMAINING,  &ei->LB_current);
  if (status) return status | ECOERR_CPLEX;
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_INTEGER, &ei->UB_current);
  if (status) return status | ECOERR_CPLEX;

  /* computes objetive function */
  max_color = ei->max_color;
  ei->objective = 0.0;
  for (j = 1; j <= max_color; j++) ei->objective += ECOwval(j, ei);

#ifdef ECODEBUGCUTS
  vertices = ei->grp->vertices;

  printf("Fractional solution:\n");
  for (j = 1; j <= max_color; j++) {
    printf("[ ");
    for (v = 1; v <= vertices; v++) printf("x%d_%d=%.4f  ", v, j, ECOxval(v, j, ei));
    printf("w%d=%.4f ]\n", j, ECOwval(j, ei));
  }
#endif

  first_j_frac = -1;
  last_j_frac = max_color + 1;
  for (j = 1; j <= max_color; j++) {
    wj = ECOwval(j, ei);
    if (wj < 1.0 - ECO_WJTOL) {
      if (first_j_frac == -1) first_j_frac = j;
      if (wj < ECO_WJTOL && last_j_frac == max_color + 1) last_j_frac = j;
    }
  }
  if (first_j_frac == -1) return ECOERR_OK;
  last_j_frac--;

  /* now, first_j_frac has the first index where wj < 1, and last_j_frac has the last index where wj > 0 */
  ei->is_wj_integer = (first_j_frac > last_j_frac ? 1 : 0);
  ei->first_j_frac = first_j_frac;
  ei->last_j_frac = last_j_frac;
#ifdef ECODEBUGCUTS
  printf("first_j_frac = %d, last_j_frac = %d, is_wj_integer = %d\n", first_j_frac, last_j_frac, ei->is_wj_integer);
#endif

#ifdef ECODEF_EXITAFTERROOT
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_REMAINING, &objvalroot);
  if (status) return status | ECOERR_CPLEX;
  /* if we are executing only root node, then we take some statistics. */
  newobjvalroot = (int)(objvalroot + 1.0 - ECO_WJTOL);
  if (ei->round == 0) {
    ei->firstobjvalroot = newobjvalroot;
    ei->firsttimeroot = ECOclock();
  }
  if (newobjvalroot > ei->bestobjvalroot) {
    ei->bestobjvalroot = newobjvalroot;
    ei->bestroundroot = ei->round;
    ei->besttimeroot = ECOclock() - ei->firsttimeroot;
    CPXmsg(ei->env->cpxlog, "New lower bound: %d     Round: %d     Time: %.2f\n", newobjvalroot, ei->round, ei->besttimeroot);
	printf("Summary: ");
        printf("Clique: %d ", ei->clique_cuts_applied);
        printf("2-rank: %d ", ei->alpha2_cuts_applied);
        printf("Block: %d ", ei->block_cuts_applied);
        printf("Nhood: %d ", ei->neighborhood_cuts_applied);
        printf("Outside: %d ", ei->outsideneighbor_cuts_applied);
        printf("ujks: %d ", ei->ujks_cuts_applied);
        printf("S-color: %d ", ei->cs_cuts_applied);
        printf("Extra: %d\n", ei->extra_cuts_applied);
  }
#endif

  /* call separation routines:
   *  1) Extra symmetry breaker
   *  2) Clique (and ujks-clique-neighborhood)
   *  3) 2-rank
   *  4) Block color
   *  5) Neighborhood
   *  6) Outside-neighborhood
   *  7) S-color ineq */

  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_EXTRA) {
    status = ECOextraseparation(ei);
    if (status) return status;
  }

  /* Except extra sym. breakers, all the cuts are regulated according to the "cut parameters". */
  if (ei->cut_enabled == 0 || ei->cut_enabled_node == 0) return ECOERR_OK;
  status = CPXgetcallbacknodeinfo(cenv, cbdata, wherefrom, 0, CPX_CALLBACK_INFO_NODE_DEPTH, &depth);
  if (status) return status | ECOERR_CPLEX;
  if (depth > ei->env->par[ECO_PARAM_MAXDEPTHTREE]) return ECOERR_OK;

  ei->cut_executed++;
  ei->round++;
#ifdef ECODEBUGCUTS
  printf("Executing cuts with: exec = %d, round = %d, depth = %d\n", ei->cut_executed, ei->round, depth);
#endif
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_CLIQUE) ei->clique_cuts_applied_round = 0;
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_UJKS) ei->ujks_cuts_applied_round = 0;
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_ALPHA2) ei->alpha2_cuts_applied_round = 0;

  for (j = 1; j <= last_j_frac; j++) {
    status = ECOcutordering(ei, j);
    if (status) return status;
    if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_CLIQUE) {
      status = ECOcliqueseparation(ei, j);
      if (status) return status;
	}
    if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_ALPHA2) {
      status = ECOalpha2separation(ei, j);
      if (status) return status;
	}
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_BLOCK) {
    status = ECOblockseparation(ei);
    if (status) return status;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_NEIGHBOR) {
    status = ECOneighborhoodseparation(ei);
    if (status) return status;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_OUTSIDENEIGHBOR) {
    status = ECOoutsideneighborseparation(ei);
    if (status) return status;
  }
  if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_CS) {
    status = ECOcsseparation(ei);
    if (status) return status;
  }
#ifdef ECODEF_REPORTDEPTH
  if (ei->cuts_applied_round >= 10 && depth >= 5) {
    CPXmsg(ei->env->cpxlog, "Depth: %d   Number of cuts applied: %d\n", depth, ei->cuts_applied_round);
  }
#endif

#ifdef ECODEBUG
  /* report how many cuts were added during the callback */
  if (ei->cuts_applied_round > 0) {
    printf("CUTS: ");
    if (ei->clique_cuts_applied_round > 0) printf("Clique: %d ", ei->clique_cuts_applied_round);
    if (ei->alpha2_cuts_applied_round > 0) printf("2-rank: %d ", ei->alpha2_cuts_applied_round);
    if (ei->block_cuts_applied_round > 0) printf("Block: %d ", ei->block_cuts_applied_round);
    if (ei->neighborhood_cuts_applied_round > 0) printf("Nhood: %d ", ei->neighborhood_cuts_applied_round);
    if (ei->outsideneighbor_cuts_applied_round > 0) printf("Outside: %d ", ei->outsideneighbor_cuts_applied_round);
    if (ei->ujks_cuts_applied_round > 0) printf("ujks: %d ", ei->ujks_cuts_applied_round);
    if (ei->cs_cuts_applied_round > 0) printf("S-color: %d ", ei->cs_cuts_applied_round);
    if (ei->extra_cuts_applied_round > 0) printf("Extra: %d ", ei->extra_cuts_applied_round);
    printf("\n");
  }
#endif

#ifdef ECODEF_WRITEROOTFILE
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_REMAINING, &objval);
  if (status) return status | ECOERR_CPLEX;
  fprintf(ei->rootstream, "  %2d    %3.6lf    %4d\n", ei->round, objval, ei->cuts_applied_round);
#endif

  if (ei->isroot == 1) {
    /* root node: disable after limit rounds is achieved */
    if (ei->round >= ei->env->par[ECO_PARAM_MAXROUNDSATROOT]) {
      CPXmsg(ei->env->cpxlog, "Max rounds for root node achieved.\n");
      ei->cut_enabled_node = 0;
      ei->round = 0;
    }
  }
  else {
    /* subproblem */
    if (ei->round >= ei->env->par[ECO_PARAM_MAXROUNDSATNODE]) {
      ei->cut_enabled_node = 0;
      ei->round = 0;
    }
  }

  return ECOERR_OK;
}

void ECObegincut(struct ecocut *cu, char sense, int rhs)
{
  cu->rhs = (double)rhs;
  cu->sense = sense;
  cu->nzcnt = 0;
}

static void addvarcut(struct ecocut *cu, int index, int coef)
{
  if (coef == 0) return; /* <-- var has a null coef. */
  if (index == 0) return; /* <-- var is fixed to zero. */
  if (index == 1) {
    /* if var is fixed to one, we pass it as a constant to the righthand side */
    cu->rhs -= (double)coef;
    return;
  }

  /* var must be added */
  cu->cutind[cu->nzcnt] = index - 2;
  cu->cutval[cu->nzcnt] = (double)coef;
  cu->nzcnt++;
}

void ECOaddxcut(struct ecocut *cu, int v, int j, int coef)
{
  addvarcut(cu, ECOmget(cu->ei->xvar_state, v, j), coef);
}

void ECOaddwcut(struct ecocut *cu, int j, int coef)
{
  addvarcut(cu, ECOvget(cu->ei->wvar_state, j), coef);
}

int ECOendcut(struct ecocut *cu, int *cutadded, double threshold, int islocal)
{
  struct ecointernal *ei;
  int status, nzcnt, n;
  char sense;
  double cte, infact;

  *cutadded = 0;
  ei = cu->ei;

  nzcnt = cu->nzcnt;
  if (nzcnt == 0) {
#ifdef ECODEBUGCUTS
    printf("(constraint empty)\n");
    switch (cu->sense) {
      case 'E':
        if (cu->rhs < -ECO_EPSILON || cu->rhs > ECO_EPSILON) { /* rhs != 0 */
          printf("E) rhs = %lf\n", cu->rhs);
          ECOabort(ECOERR_INTERNAL);
        }
        break;
      case 'L':
        if (cu->rhs < -ECO_EPSILON) { /* rhs < 0 */
          printf("L) rhs = %lf\n", cu->rhs);
          ECOabort(ECOERR_INTERNAL);
        }
        break;
      case 'G':
        if (cu->rhs > ECO_EPSILON) { /* rhs > 0 */
          printf("G) rhs = %lf\n", cu->rhs);
          ECOabort(ECOERR_INTERNAL);
        }
        break;
    }
#endif
    return ECOERR_OK;
  }
  if (nzcnt == 1) {
#ifdef ECODEBUGCUTS
    printf("(just one variable) --> ");
#endif
    cte = cu->rhs / cu->cutval[0];
    sense = cu->sense;
    if (cu->cutval[0] < -ECO_EPSILON) {
      /* we need to invert the sense if coef < 0 */
      switch (sense) {
        case 'E': break;
        case 'L': sense = 'G'; break;
        case 'G': sense = 'L'; break;
      }
    }
    switch (sense) {
      case 'E': /* var = cte --- normalize it */
        cu->cutval[0] = 1.0;
        cu->rhs = cte;
        cu->sense = 'E';
        break;
      case 'L': /* var <= cte --- if cte < 1 then var = 0 */
        if (cte >= 1.0) {
#ifdef ECODEBUGCUTS
          printf("discarded!\n");
#endif
          return ECOERR_OK;
        }
        cu->cutval[0] = 1.0;
        cu->rhs = 0.0;
        cu->sense = 'E';
        break;
      case 'G': /* var >= cte --- if cte > 0 then var = 1 */
        if (cte <= 1.0) {
#ifdef ECODEBUGCUTS
          printf("discarded!\n");
#endif
          return ECOERR_OK;
        }
        cu->cutval[0] = 1.0;
        cu->rhs = 1.0;
        cu->sense = 'E';
        break;
    }
#ifdef ECODEBUGCUTS
    printf("added!\n");
#endif
  }
  else {
    /* the cut has several variables, check violation */
    infact = cu->rhs;
    for (n = 0; n < nzcnt; n++) infact -= cu->cutval[n] * ei->x[cu->cutind[n]];
    switch (cu->sense) {
      case 'E': if (infact < 0.0) infact = -infact; break;
      case 'L': infact = -infact; break;
      case 'G': break;
    }
#ifdef ECODEBUGCUTS
    printf("(infact = %.4f) --> ", infact);
#endif
    if (infact < threshold) {
#ifdef ECODEBUGCUTS
      printf("discarded!\n");
#endif
      return ECOERR_OK;
    }
#ifdef ECODEBUGCUTS
    printf("added!\n");
#endif
  }

  /* add the cut */
  if (islocal == ECO_ON) {
    status = CPXcutcallbackaddlocal(cu->cenv, cu->cbdata, cu->wherefrom,
                            nzcnt, cu->rhs, cu->sense, cu->cutind, cu->cutval);
  }
  else {
#if CPX_VERSION >= 1200
#ifdef ECODEF_PURGEABLE
    status = CPXcutcallbackadd(cu->cenv, cu->cbdata, cu->wherefrom,
                            nzcnt, cu->rhs, cu->sense, cu->cutind, cu->cutval, 1);
#else
    status = CPXcutcallbackadd(cu->cenv, cu->cbdata, cu->wherefrom,
                            nzcnt, cu->rhs, cu->sense, cu->cutind, cu->cutval, 0);
#endif
#else
  status = CPXcutcallbackadd(cu->cenv, cu->cbdata, cu->wherefrom,
                        nzcnt, cu->rhs, cu->sense, cu->cutind, cu->cutval);
#endif
  }

  if (status) return status | ECOERR_CPLEX;

  ei->cuts_applied++;
  ei->cuts_applied_round++;
  *(cu->useraction_p) = CPX_CALLBACK_SET;
  *cutadded = 1;
  return ECOERR_OK;
}
