
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * degree.c - order vertices by degree
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOinitdegrees(struct ecointernal *ei)
{
  int vertices, v, d, status, deg;
  ECOGRPptr grp;
  struct ecolist *hv, *hvnext;

  grp = ei->grp;
  vertices = grp->vertices;

  /* alloc memory */
  ei->degree = malloc(sizeof(ECOVECptr)*vertices);
  if (ei->degree == NULL) return ECOERR_MEMORY;

  /* fill "degree" structure and free "host_vertices" at the same time */
  for (v = 1; v <= vertices; v++) {
    hv = &(grp->hash_vertices[v-1]);
	deg = hv->value; 
    ei->degree[v-1] = ECOallocvector(&status, deg);
    if (status) ECOabort(ECOERR_MEMORY);
	d = 1;
	hv = hv->next;
	if (hv != NULL) {
      do {
	    ECOvset(ei->degree[v-1], d++, hv->value);
        hvnext = hv->next;
		free(hv);
		hv = hvnext;
	  } while (hv != NULL);
	}
  }
  free(grp->hash_vertices);
  grp->hash_vertices = NULL;

#ifdef ECODEBUG
  printf("Degree of vertices:\n");
  for (v = 1; v <= vertices; v++) {
    printf(" d(%d) = %d, N(%d) =", v, ECOgetdegree(ei, v), v);
    for (d = 1; d <= ECOgetdegree(ei, v); d++) printf(" %d", ECOgetneighbor(ei, v, d));
    printf(".\n");
  }
#endif
  return ECOERR_OK;
}

int ECOgetdegree(struct ecointernal *ei, int v)
{
#ifdef ECODEBUG
  if (v < 1 || v > ei->grp->vertices) ECOabort(ECOERR_BOUNDS);
#endif

  return ECOvsize(ei->degree[v - 1]);
}

int ECOgetneighbor(struct ecointernal *ei, int v, int i)
{
#ifdef ECODEBUG
  if (v < 1 || v > ei->grp->vertices) ECOabort(ECOERR_BOUNDS);
#endif

  return ECOvget(ei->degree[v - 1], i);
}

int ECOordervertices(struct ecointernal *ei)
{
  ECOVECptr index, value, subgraph;
  int v, i, n, n2, vertices, mindegree, minvert, degree, vert;
  int count_isolated, status;

  /* ask for memory */
  vertices = ei->grp->vertices;
  ei->ascending_by_degree = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->descending_by_degree = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->sl_ordering = ECOallocvector(&status, vertices);
  if (status) return status;

  /* ask for internal memory */
  index = ECOallocvector(&status, vertices);
  if (status) return status;
  value = ECOallocvector(&status, vertices);
  if (status) {
    ECOfreevector(&index);
    return status;
  }
  subgraph = ECOallocvector(&status, vertices);
  if (status) {
    ECOfreevector(&value);
    ECOfreevector(&index);
    return status;
  }

  /* load degree of vertices and sort */
  for (v = 1; v <= vertices; v++) {
    ECOvset(index, v, v);
    ECOvset(value, v, ECOgetdegree(ei, v));
  }
  status = ECOsort(index, value, ei->ascending_by_degree, ECO_ASCENDING);
  if (status) {
    ECOfreevector(&subgraph);
    ECOfreevector(&value);
    ECOfreevector(&index);
    return status;
  }

  /* count isolated vertices */
  count_isolated = 0;
  while (ECOgetdegree(ei, ECOvget(ei->ascending_by_degree, count_isolated + 1)) == 0)
    count_isolated++;
  ei->count_isolated = count_isolated;
#ifdef ECODEBUGINIT
  printf("Order of vertices (ascending by degree):");
  ECOshowvector(ei->ascending_by_degree);
#endif
  CPXmsg(ei->env->cpxlog, "There are %d isolated vertices.\n", count_isolated);

  /* order vertices in descending order by degree */
  for (v = 1; v <= vertices; v++) {
    ECOvfill(subgraph, 0);
	/* computes N^2[v] */
	for (n = 1; n <= ECOgetdegree(ei, v); n++) {
      vert = ECOgetneighbor(ei, v, n);
	  ECOvset(subgraph, vert, 1);
	  for (n2 = 1; n2 <= ECOgetdegree(ei, vert); n2++)
	    ECOvset(subgraph, ECOgetneighbor(ei, vert, n2), 1);
	}
	n = 0;
	/* n = |N^2[v]| */
	for (i = 1; i <= vertices; i++)
	  if (ECOvget(subgraph, i) == 1) n++;
    ECOvset(index, v, v);
    ECOvset(value, v, (vertices + 1) * ECOgetdegree(ei, v) + n);
  }
  status = ECOsort(index, value, ei->descending_by_degree, ECO_DESCENDING);
  if (status) {
    ECOfreevector(&subgraph);
    ECOfreevector(&value);
    ECOfreevector(&index);
    return status;
  }

#ifdef ECODEBUGINIT
  printf("Order of vertices (descending by degree):");
  ECOshowvector(ei->descending_by_degree);
#endif

  /* order vertices with SL method (Matula et al.) */
  ECOvfill(subgraph, 1);
  for (v = 1; v <= vertices; v++) {
    minvert = -1;
    mindegree = 1000000;
    /* find vertex with the smallest degree in the subgraph */
    for (i = 1; i <= vertices; i++) {
      vert = ECOvget(ei->ascending_by_degree, i);
      if (ECOvget(subgraph, vert)) {
        degree = 0;
        for (n = 1; n <= ECOgetdegree(ei, vert); n++)
          degree += ECOvget(subgraph, ECOgetneighbor(ei, vert, n));
        if (degree < mindegree) {
          minvert = vert;
          mindegree = degree;
        }
      }
    }
    ECOvset(ei->sl_ordering, vertices + 1 - v, minvert);
    ECOvset(subgraph, minvert, 0);
  }

#ifdef ECODEBUGINIT
  printf("SL ordering:");
  ECOshowvector(ei->sl_ordering);
#endif

  ECOfreevector(&subgraph);
  ECOfreevector(&value);
  ECOfreevector(&index);
  return ECOERR_OK;
}

int ECOmaxindepneighborhood(struct ecointernal *ei)
{
  int v, vertices, status, theta, beta;
  ECOVECptr indepset, neighbor;

  CPXmsg(ei->env->cpxlog, "Searching for bounds of alpha(N(v)).\n");

  /* request for memory */
  vertices = ei->grp->vertices;
  ei->neighbormaxindep1_size = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->neighbormaxindep2_size = ECOallocvector(&status, vertices);
  if (status) return status;
  indepset = ECOallocvector(&status, vertices);
  if (status) return status;

  for (v = 1; v <= vertices; v++) {
    neighbor = ei->degree[v - 1];
    /* beta = max. greedy indep. set in N(v) */
    status = ECOgreedyindepset(ei, neighbor, indepset, &beta);
    if (status) {
      ECOfreevector(&indepset);
      return status;
    }
    ECOvset(ei->neighbormaxindep1_size, v, beta);
    /* theta = max. greedy partition of cliques in N(v) */
    status = ECOgreedypartition(ei, neighbor, &theta);
    if (status) {
      ECOfreevector(&indepset);
      return status;
    }
    ECOvset(ei->neighbormaxindep2_size, v, theta);
#ifdef ECODEBUGINIT
    printf(" %d <= alpha(N(%d)) <= %d\n", beta, v, theta);
#endif
  }

  ECOfreevector(&indepset);
  return ECOERR_OK;
}

/* This function gives an upper bound of sum_{u in N(v)} x_uj */
int ECOrvalue(struct ecointernal *ei, int v, int j, int min_color)
{
  int r, s;

  if (j < min_color) j = min_color;
  r = ECOvget(ei->neighbormaxindep2_size, v);
  s = ECOceil(ei->grp->vertices, j);

  if (r > s) return s;
  return r;
}

int ECOfloor(int n, int j)
{
  return n / j;
}

int ECOceil(int n, int j)
{
  int d;

  d = n / j;
  if (n % j > 0) d++;
  return d;
}

int ECOtkj(int k, int j, int n)
{
  int d;

  if (j > k) return 0;
  d = n / k;
  if (j <= n % k) d++;
  return d;
}
