
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * enum.c - Enumeration algorithm (uses ideas from DSATUR)
 * We borrow some structures from tabu search, so ECOtabuinit must be executed before ECOenum!
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOenum(struct ecointernal *ei)
{
  ECOMTXptr xvjstate;
  ECOVECptr bestcoloring, curcoloring, size_classes;
  int xvj, vertices, max_color, v, j, vert, degree, status;
  char longnumber[32];

  /* initialize variables */
  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  xvjstate = ei->tabu_tenure;
  bestcoloring = ei->tabu_bestcoloring;
  curcoloring = ei->tabu_curcoloring;
  size_classes = ei->tabu_bigclasses;

  /* we must know a (max_color)-coloring in order to proceed! */
  if (max_color != ei->grp->colors_used) {
    CPXmsg(ei->env->cpxlog, "A %d-eqcol must be known in order to continue. Sorry!\n", max_color);
        CPXmsg(ei->env->cpxlog, "(try executing ecopt with options: -u3 -x5=%d)\n", max_color);
        return ECOERR_INTERNAL;
  }

  CPXmsg(ei->env->cpxlog, "Starting enumeration.\n");
  ei->last_t = ei->start_t;
  ei->nodeprocessed = 0;

  /* fill curcoloring and state of variables xvj:
   * if xvj = 0, vertex "v" can not use color "j"
   * if xvj = 1, vertex "v" is already painted with color "j"
   * if xvj = 2, it is not known yet whether vertex "v" uses color "j" or not */
  ECOvfill(curcoloring, 0);
  ECOvfill(size_classes, 0);
  ei->curnumberofcolors = 0;
#ifdef ECODEBUGENUM
  printf("Initial admissible colors for:\n");
#endif
  switch (ei->env->par[ECO_PARAM_STRATEGY]) {
    case ECO_STRATEGY_ENUM:
          ei->isdynamic = 0;
          break;
        case ECO_STRATEGY_ENUMDYN1:
          ei->isdynamic = 1;
          break;
        case ECO_STRATEGY_ENUMDYN2:
          ei->isdynamic = 2;
          break;
        default:
          return ECOERR_INTERNAL;
  }
  if (ei->isdynamic == 0) {
    /* STATIC COLOR SELECTION: we consider how we fix variables in ECOordersymmetry (xvar_state) */
    for (v = 1; v <= vertices; v++) {
#ifdef ECODEBUGENUM
      printf(" vertex %d are { ", v);
#endif
      for (j = 1; j <= max_color; j++) {
            xvj = ECOmget(ei->xvar_state, v, j);
                if (xvj >= 2) xvj = 2;
        ECOmset(xvjstate, v, j, xvj);
            if (xvj == 1) {
          ECOvset(curcoloring, v, j);
                  if (j > ei->curnumberofcolors) ei->curnumberofcolors = j;
                  ECOvincr(size_classes, j);
            }
#ifdef ECODEBUGENUM
            if (xvj == 1) printf("%d(unique) ", j);
            if (xvj == 2) printf("%d ", j);
#endif
          }
#ifdef ECODEBUGENUM
          printf("}\n");
#endif
    }
  }
  else {
    /* DYNAMIC COLOR SELECTION: we only fix the colors for the initial clique */
    for (v = 1; v <= vertices; v++) for (j = 1; j <= max_color; j++) ECOmset(xvjstate, v, j, 2);
    for (v = 1; v <= ECOvsize(ei->clique_set); v++) {
      vert = ECOvget(ei->ordering_symmetry, v);
      /* fix x_vv = 1 for v in the clique: variable "vert" is the vertex, variable "v" is the color */
          for (j = 1; j <= max_color; j++) ECOmset(xvjstate, vert, j, 0);
      ECOmset(xvjstate, vert, v, 1);
      ECOvset(curcoloring, vert, v);
          ei->curnumberofcolors++;
          ECOvincr(size_classes, v);
      /* fix x_uv = 0 for u adjacent to v */
      degree = ECOgetdegree(ei, vert);
      for (j = 1; j <= degree; j++) ECOmset(xvjstate, ECOgetneighbor(ei, vert, j), v, 0);
        }
#ifdef ECODEBUGENUM
    for (v = 1; v <= vertices; v++) {
      printf(" vertex %d are { ", v);
      for (j = 1; j <= max_color; j++) {
            xvj = ECOmget(xvjstate, v, j);
            if (xvj == 1) printf("%d(unique) ", j);
            if (xvj == 2) printf("%d ", j);
          }
          printf("}\n");
        }
#endif
  }
#ifdef ECODEBUGENUM
  printf("Initial coloring with %d colors:", ei->curnumberofcolors);
  ECOshowvector(curcoloring);
  printf("Size of classes:");
  ECOshowvector(size_classes);
#endif

  /* fill bestcoloring and start branching procedure */
  ECOvcopy(ei->grp->coloring, bestcoloring);
  ei->bestnumberofcolors = ei->grp->colors_used;
#ifdef ECODEBUGENUM
  printf("Best coloring with %d colors:", ei->bestnumberofcolors);
  ECOshowvector(bestcoloring);
#endif

  status = ECOenumbranch(ei, ei->min_color);
  switch (status) {
    case ECOENUM_OK: /* a solution has found */
        case ECOENUM_OPTIMAL: /* a solution of min_colors has found */
      CPXmsg(ei->env->cpxlog, "Optimal solution found.\n");
      ECOvcopy(bestcoloring, ei->grp->coloring);
      ei->grp->colors_used = ei->bestnumberofcolors;
          ei->grp->is_optimal = 1;
          break;
        case ECOENUM_TIMEEXHAUSTED: /* time limit reached */
      CPXmsg(ei->env->cpxlog, "Time exhausted. Giving current solution.\n");
      ECOvcopy(bestcoloring, ei->grp->coloring);
      ei->grp->colors_used = ei->bestnumberofcolors;
          ei->grp->is_optimal = 0;
          break;
        case ECOENUM_INFEASIBLE: /* no solutions found */
      CPXmsg(ei->env->cpxlog, "Infeasible. Initial solution is optimal.\n");
          ei->grp->is_optimal = 1;
      break;
        default:
      return status;
  }

  sprintf(longnumber, "%lld", ei->nodeprocessed);
  CPXmsg(ei->env->cpxlog, "Enumeration completed (%s nodes processed)\n", longnumber);
  return ECOERR_OK;
}

/* before calling this function, structures bestcoloring, curcoloring and size_classes must be adequately filled */
int ECOenumbranch(struct ecointernal *ei, int lb)
{
  ECOMTXptr xvjstate;
  ECOVECptr bestcoloring, curcoloring, size_classes, colority, vstate, nstate;
  ECOVECptr index, value, ordering;
  int i, v, bestv, bestimpact, bestsubimpact, neighbor, feasible, flag, status;
  int vertices, sum, max_color, n, c, j, degree, impact, subimpact, k, counter;
  int oldnumberofcolors, max_color_in_curcoloring;
  char longnumber[32];
  double clk;

  ei->nodeprocessed++;
  ei->totalnodeprocessed++;

  /* make some checks:
     warning!! we substract 10 seconds in order to give time to the optimizer.
     also, if we are running inside the B&C, we are restricted to a certain available time. */
  clk = ECOclock();
  if (clk - ei->start_t + 10.0 >= ei->maxtim_t) return ECOENUM_TIMEEXHAUSTED;
  if (ei->enumactivated == 1) {
    if (clk - ei->enuminittime >= ei->enumavailabletime) return ECOENUM_TIMEEXHAUSTED;
  }
  if (ei->curnumberofcolors >= ei->bestnumberofcolors) return ECOENUM_INFEASIBLE;

  /* log each 10 second */
  if (clk - ei->last_t >= 10.0) {
	sprintf(longnumber, "%lld", ei->nodeprocessed);
    CPXmsg(ei->env->cpxlog, " node: %s, best: %d, time: %.2f seg.\n", longnumber, ei->bestnumberofcolors, clk - ei->start_t);
    ei->last_t = clk;
  }

  /* initialize variables */
  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  xvjstate = ei->tabu_tenure;
  bestcoloring = ei->tabu_bestcoloring;
  curcoloring = ei->tabu_curcoloring;
  size_classes = ei->tabu_bigclasses;
  colority = ei->tabu_coloredvert;

  /* check if there exist enough vertices to paint */
  sum = 0;
  for (j = 1; j <= ei->bestnumberofcolors-1; j++) {
    i = ECOfloor(vertices, ei->bestnumberofcolors-1) - ECOvget(size_classes, j);
    if (i < 0) i = 0;
    sum += i;
  }
  for (v = 1; v <= vertices; v++) if (ECOvget(curcoloring, v) == 0) sum--;
  if (sum > 0) return ECOENUM_INFEASIBLE;

  /* find next vertex to paint */
  bestv = -1;
  bestimpact = -1;
  bestsubimpact = -1;
  for (v = 1; v <= vertices; v++) {
        if (ECOvget(curcoloring, v) == 0) {
          degree = ECOgetdegree(ei, v);
          /* compute impact of "v": we use Brelaz branching rule, based on colority of neighborhood of "v" */
      ECOvfill(colority, 0);
          impact = 0;
          for (n = 1; n <= degree; n++) {
                c = ECOvget(curcoloring, ECOgetneighbor(ei, v, n));
                if (c > 0) {
                  if (ECOvget(colority, c) == 0) {
                    ECOvset(colority, c, 1);
                        impact++;
                  }
                }
          }
          if (impact >= bestimpact) {
        /* compute subimpact: we use tie-breaking rule proposed by Sewell */
        subimpact = 0;
                for (j = 1; j < ei->bestnumberofcolors; j++) {
                  if (ECOmget(xvjstate, v, j) == 2) {
                        /* now, "j" is a candidate color for "v"
                         * we count how many neighbors could use this color */
                    for (n = 1; n <= degree; n++) {
                          if (ECOmget(xvjstate, ECOgetneighbor(ei, v, n), j) == 2) subimpact++;
                        }
                  }
                }
                if (impact > bestimpact || subimpact > bestsubimpact) {
          bestimpact = impact;
                  bestsubimpact = subimpact;
                  bestv = v;
                }
          }
        }
  }

  if (bestv == -1) {
#ifdef ECODEBUGENUM
//    printf("Leaf coloring:");
//    ECOshowvector(curcoloring);
//    printf("Size of classes:");
//    ECOshowvector(size_classes);
#endif
    /* the coloring is complete, check if it is equitable. */
        flag = 0;
    k = ei->curnumberofcolors;
    for (j = 1; j <= k; j++) {
      c = ECOvget(size_classes, j);
          if (c < ECOfloor(vertices, k) || c > ECOceil(vertices, k)) {
            flag = 1;
                break;
          }
        }
        if (flag == 1) return ECOENUM_INFEASIBLE;

        /* update bestcoloring */
    ECOvcopy(curcoloring, bestcoloring);
    ei->bestnumberofcolors = k;
        CPXmsg(ei->env->cpxlog, "A %d-eqcol has been found.\n", k);
        if (k <= lb) return ECOENUM_OPTIMAL;
        return ECOENUM_OK;
  }

#ifdef ECODEBUGENUM
//  printf("Vertex chosen: %d (impact=%d,subimpact=%d)\n", bestv, bestimpact, bestsubimpact);
#endif
  /* we first ask for memory */
  vstate = ECOallocvector(&status, max_color);
  if (status) return status;
  nstate = ECOallocvector(&status, vertices);
  if (status) return status;
  index = ECOallocvector(&status, max_color);
  if (status) return status;
  value = ECOallocvector(&status, max_color);
  if (status) return status;
  ordering = ECOallocvector(&status, max_color);
  if (status) return status;

  if (ei->isdynamic > 0) {
    /* compute the greater color used in the current coloring */
    max_color_in_curcoloring = 0;
    for (v = 1; v <= vertices; v++) {
          c = ECOvget(curcoloring, v);
      if (c > max_color_in_curcoloring) max_color_in_curcoloring = c;
    }
  }

  /* save the state of vertex "bestv" */
  for (k = 1; k <= max_color; k++) ECOvset(vstate, k, ECOmget(xvjstate, bestv, k));
  oldnumberofcolors = ei->curnumberofcolors;
  feasible = 0;
  degree = ECOgetdegree(ei, bestv);

  /* sort feasible color classes */
  counter = 0;
  switch (ei->isdynamic) {
    case 0: /* STATIC COLOR SELECTION: we simply consider all colors between 1 and UB-1 */
    for (j = 1; j <= ei->bestnumberofcolors-1; j++) {
      if (ECOvget(vstate, j) == 2) {
        if (ECOvget(size_classes, j) + 1 <= ECOceil(vertices, j > oldnumberofcolors ? j : oldnumberofcolors)) {
          /* now, "j" is a candidate for "bestv", so consider it */
          counter++;
          ECOvset(index, counter, j);
          ECOvset(value, counter, ECOvget(size_classes, j));
        }
      }
    }
    break;
    case 1: /* DYNAMIC COLOR SELECTION I: we consider colors from 1 to min{UB-1,max_color_in_curcoloring+1} */
    for (j = 1; j <= max_color_in_curcoloring; j++) {
      if (ECOvget(vstate, j) == 2) {
        if (ECOvget(size_classes, j) + 1 <= ECOceil(vertices, j > oldnumberofcolors ? j : oldnumberofcolors)) {
          /* now, "j" is a candidate for "bestv", so consider it */
          counter++;
          ECOvset(index, counter, j);
          ECOvset(value, counter, ECOvget(size_classes, j));
        }
      }
    }
    j = max_color_in_curcoloring + 1;
    if (j <= ei->bestnumberofcolors-1) {
      counter++;
      ECOvset(index, counter, j);
      ECOvset(value, counter, 999999);
    }
    break;
    case 2: /* DYNAMIC COLOR SELECTION II: we consider first max_color_in_curcoloring+1 and then 1,...,max_color_in_curcoloring */
    for (j = 1; j <= max_color_in_curcoloring; j++) {
      if (ECOvget(vstate, j) == 2) {
        if (ECOvget(size_classes, j) + 1 <= ECOceil(vertices, j > oldnumberofcolors ? j : oldnumberofcolors)) {
          /* now, "j" is a candidate for "bestv", so consider it */
          counter++;
          ECOvset(index, counter, j);
          ECOvset(value, counter, ECOvget(size_classes, j));
        }
      }
    }
    j = max_color_in_curcoloring + 1;
    if (j <= ei->bestnumberofcolors-1) {
      counter++;
      ECOvset(index, counter, j);
      ECOvset(value, counter, -1);
    }
    break;
  }

  if (counter > 0) {
    ECOvredim(index, counter);
    ECOvredim(value, counter);
    ECOvredim(ordering, counter);
    status = ECOsort(index, value, ordering, ECO_ASCENDING);
    if (status) {
      ECOfreevector(&ordering);
      ECOfreevector(&value);
      ECOfreevector(&index);
      ECOfreevector(&nstate);
      ECOfreevector(&vstate);
      return status;
    }

    /* we scan feasible colors for "bestv" recursively */
    for (i = 1; i <= counter; i++) {
      j = ECOvget(ordering, i);
      ei->curnumberofcolors = j > oldnumberofcolors ? j : oldnumberofcolors;
      for (k = 1; k <= max_color; k++) ECOmset(xvjstate, bestv, k, j == k ? 1 : 0);
      ECOvset(curcoloring, bestv, j);
      ECOvincr(size_classes, j);
      /* forbid this color for the neighbors of "bestv" and save these changes in "nstate" */
      ECOvfill(nstate, 0);
      for (n = 1; n <= degree; n++) {
        neighbor = ECOgetneighbor(ei, bestv, n);
        if (ECOmget(xvjstate, neighbor, j) == 2) {
          ECOmset(xvjstate, neighbor, j, 0);
          ECOvset(nstate, neighbor, 1);
        }
#ifdef ECODEBUGENUM
                    if (ECOmget(xvjstate, neighbor, j) == 1) {
                          printf("Vertex %d is already painted with color %d.\n", neighbor, j);
              return ECOERR_INTERNAL;
                    }
#endif
      }
      /* has all uncolored vertex at least one available color? */
      for (n = 1; n <= vertices; n++) {
        if (ECOvget(curcoloring, n) == 0) {
          c = 0;
          for (k = 1; k < ei->bestnumberofcolors; k++) {
            if (ECOmget(xvjstate, n, k) == 2) {
              c = 1;
              break;
            }
          }
          if (c == 0) goto node_fathomed;
        }
      }
#ifdef ECODEBUGENUM
//          printf("Set color %d to vertex %d. Now, class %d has %d vertices.\n", j, bestv, j, ECOvget(size_classes, j));
#endif
      status = ECOenumbranch(ei, lb);
      switch (status) {
        case ECOENUM_OK: /* a solution has found */
        feasible = 1;
        break;
        case ECOENUM_OPTIMAL: /* a solution of min_colors has found, exit with this solution */
        ECOfreevector(&ordering);
        ECOfreevector(&value);
        ECOfreevector(&index);
        ECOfreevector(&nstate);
        ECOfreevector(&vstate);
        return ECOENUM_OPTIMAL;
        case ECOENUM_TIMEEXHAUSTED: /* time limit reached, exit immediately */
        ECOfreevector(&ordering);
        ECOfreevector(&value);
        ECOfreevector(&index);
        ECOfreevector(&nstate);
        ECOfreevector(&vstate);
        return ECOENUM_TIMEEXHAUSTED;
        case ECOENUM_INFEASIBLE: /* no solutions found */
        break;
        default:
        ECOfreevector(&ordering);
        ECOfreevector(&value);
        ECOfreevector(&index);
        ECOfreevector(&nstate);
        ECOfreevector(&vstate);
        return status;
      }
node_fathomed:;
      ECOvdecr(size_classes, j);
      /* restore changes from "nstate" */
      for (n = 1; n <= degree; n++) {
        neighbor = ECOgetneighbor(ei, bestv, n);
        if (ECOvget(nstate, neighbor) == 1) ECOmset(xvjstate, neighbor, j, 2);
      }
    }
  }

  /* we restore the state of vertex "bestv" */
  for (k = 1; k <= max_color; k++) ECOmset(xvjstate, bestv, k, ECOvget(vstate, k));
  ECOvset(curcoloring, bestv, 0);
  ei->curnumberofcolors = oldnumberofcolors;

  ECOfreevector(&ordering);
  ECOfreevector(&value);
  ECOfreevector(&index);
  ECOfreevector(&nstate);
  ECOfreevector(&vstate);
  return feasible == 1 ? ECOENUM_OK : ECOENUM_INFEASIBLE;
}

/* This routine scans every solution of the formulation M in order to check if the inequality is valid and
 * compute the dimension of the face defined by it (and checking if the ineq defines a facet)
 * The input is array_v <-- indice v of variable x_vj
 *              array_j <-- if j>0, indice j of variable x_vj
 *                          if j<0, indice -j of variable w_j
 *           array_coef <-- coefficient of x_vj or w_j
 *           array_size <-- number of variables in the support
 *                  rhs <-- right hand side of the ineq */
int ECOcheckineq(ECOENVptr env, ECOGRPptr grp, int *array_v, int *array_j, int *array_coef, int array_size, int rhs)
{
  ECOVECptr curcoloring, size_classes, eqcol_admit;
  CPXENVptr cenv;
  struct ecointernal ei;
  int vertices, edges, status, k, flag, dimpoly;
  double start_t, pass_t;

  if (env == NULL) return ECOERR_NOENV;
  if (grp == NULL) return ECOERR_NOGRP;

  /* initialize */
  cenv = env->cplexenv;
  vertices = grp->vertices;
  edges = grp->edges;
  CPXmsg(env->cpxlog, "The graph has %d vertices and %d edges.\n", vertices, edges);

  /* check if G is isolated */
  if (edges == 0) {
    CPXmsg(env->cpxlog, "The graph only contains isolated vertices.\n");
    return ECOERR_OK;
  }

  /* check if G is K_n */
  if (edges == vertices * (vertices - 1) / 2) {
    CPXmsg(env->cpxlog, "The graph is complete.\n");
    return ECOERR_OK;
  }

  /* set internal environment (pointers) */
  ei.env = env;
  ei.grp = grp;
  ECOsetinternal(&ei);

  /* ask for memory: BE CAREFUL, we borrow some vectors from tabu search */
  ei.tabu_curcoloring = ECOallocvector(&status, vertices);
  if (status) return status;
  curcoloring = ei.tabu_curcoloring;
  ei.tabu_bigclasses = ECOallocvector(&status, vertices);
  if (status) return status;
  size_classes = ei.tabu_bigclasses;
  ei.tabu_coloredvert = ECOallocvector(&status, vertices);
  if (status) return status;
  ei.tabu_conflictvertexset = ECOallocvector(&status, vertices);
  if (status) return status;
  eqcol_admit = ei.tabu_conflictvertexset;

  /* get degree and neighborhoods ... */
  start_t = ECOclock();
  pass_t = start_t;
  status = ECOinitdegrees(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }

  /* Write the solutions in a file: BE CAREFUL, we borrow "rootstream" */
  ei.rootstream = fopen("points.poi", "wt");
  if (ei.rootstream == NULL) {
    ECOfreemem(&ei);
    return ECOERR_CREAT;
  }
  fprintf(ei.rootstream, "DIM = %d\n\nCONV_SECTION\n", vertices* vertices + vertices);

  CPXmsg(env->cpxlog, "Starting enumeration.\n");
  ei.numberofsolutions = 0;
  ei.numberofsolutionsface = 0;
  ei.array_v = array_v;
  ei.array_j = array_j;
  ei.array_coef = array_coef;
  ei.array_size = array_size;
  ei.rhs = rhs;
  ECOvfill(curcoloring, 0);
  ECOvfill(size_classes, 0);
  ECOvfill(eqcol_admit, 0);
  ei.curnumberofcolors = 0;
  status = ECOcheckineqenum(&ei, start_t, &pass_t);
  switch (status) {
    case ECOENUM_OK:
          CPXmsg(env->cpxlog, "The inequality is valid.\n");
          break;
        case ECOENUM_INFEASIBLE:
      CPXmsg(env->cpxlog, "Time elapsed in enumeration: %.2f seg.\n", ECOclock() - start_t);
      CPXmsg(env->cpxlog, "Solutions evaluated: %d.\n", ei.numberofsolutions);
          fclose(ei.rootstream);
          unlink("points.poi");
      ECOfreemem(&ei);
      return ECOERR_OK;
        default:
          fclose(ei.rootstream);
          unlink("points.poi");
      ECOfreemem(&ei);
      return status;
  }

  CPXmsg(env->cpxlog, "Time elapsed in enumeration: %.2f seg.\n", ECOclock() - start_t);
  CPXmsg(env->cpxlog, "Solutions evaluated: %d.\n", ei.numberofsolutions);
  CPXmsg(env->cpxlog, "Solutions in the face: %d.\n", ei.numberofsolutionsface);
  flag = 0;
  dimpoly = vertices * vertices - 1;
  for (k = 1; k <= vertices; k++) {
    if (flag == 0) {
          if (ECOvget(eqcol_admit, k) == 1) {
                /* found chromatic number */
        CPXmsg(env->cpxlog, "Eq. crom. number = %d.\nSkip set = {", k);
                flag = 1;
                dimpoly -= k;
          }
        }
        else {
          if (ECOvget(eqcol_admit, k) == 0) {
                /* found values of the skip-set */
            CPXmsg(env->cpxlog, " %d", k);
                dimpoly--;
          }
    }
  }
  fprintf(ei.rootstream, "\nEND\n");
  fclose(ei.rootstream);
  CPXmsg(env->cpxlog, " }.\nDimension of ECP = %d (minimal system has %d equations).\nSolutions saved in points.poi.\n", dimpoly, vertices * vertices + vertices - dimpoly);
  ECOfreemem(&ei);
  return ECOERR_OK;
}

int ECOcheckineqenum(struct ecointernal *ei, double start_t, double *pass_t)
{
  ECOVECptr curcoloring, size_classes, colority, eqcol_admit;
  int v, bestv, bestimpact, bestsubimpact, subimpact, flag, status;
  int vertices, n, c, j, i, lhs, degree, impact, k, oldnumberofcolors;
  double clk;

  /* initialize variables */
  vertices = ei->grp->vertices;
  curcoloring = ei->tabu_curcoloring;
  size_classes = ei->tabu_bigclasses;
  colority = ei->tabu_coloredvert;
  eqcol_admit = ei->tabu_conflictvertexset;

  /* find next vertex to paint */
  bestv = -1;
  bestimpact = -1;
  bestsubimpact = -1;
  for (v = 1; v <= vertices; v++) {
        if (ECOvget(curcoloring, v) == 0) {
          degree = ECOgetdegree(ei, v);
          /* compute impact of "v": we use Brelaz branching rule, based on colority of neighborhood of "v" */
      ECOvfill(colority, 0);
          impact = 0;
          for (n = 1; n <= degree; n++) {
                c = ECOvget(curcoloring, ECOgetneighbor(ei, v, n));
                if (c > 0) {
                  if (ECOvget(colority, c) == 0) {
                    ECOvset(colority, c, 1);
                        impact++;
                  }
                }
          }
          if (impact >= bestimpact) {
        /* compute subimpact: we use the degree of the vertex */
        subimpact = degree;
                if (impact > bestimpact || subimpact > bestsubimpact) {
          bestimpact = impact;
                  bestsubimpact = subimpact;
                  bestv = v;
                }
          }
        }
  }

  if (bestv == -1) {
    /* the coloring is complete, check if it is equitable. */
        flag = 0;
    k = ei->curnumberofcolors;
    for (j = 1; j <= k; j++) {
      c = ECOvget(size_classes, j);
          if (c < ECOfloor(vertices, k) || c > ECOceil(vertices, k)) {
            flag = 1;
                break;
          }
        }
        if (flag == 1) return ECOENUM_OK;
#ifdef ECODEBUGENUM
    printf("Leaf coloring:");
    ECOshowvector(curcoloring);
    printf("Size of classes:");
    ECOshowvector(size_classes);
#endif
        ei->numberofsolutions++;
        ECOvset(eqcol_admit, k, 1);

        /* log */
        clk = ECOclock();
        if (clk - *pass_t >= 10.0) {
          CPXmsg(ei->env->cpxlog, " total solutions: %d, sol. in the face: %d, time: %.2f seg.\n", ei->numberofsolutions, ei->numberofsolutionsface, clk - start_t);
          *pass_t = clk;
        }

        /* test if the inequality is valid */
        lhs = 0;
        for (i = 0; i < ei->array_size; i++) {
          v = ei->array_v[i];
          j = ei->array_j[i];
          c = ei->array_coef[i];
          if (j > 0) { /* check the value of x_vj */
        if (ECOvget(curcoloring, v) == j) lhs += c;
          }
          else { /* check the value of w_j */
            if (-j <= k) lhs += c;
          }
        }
        if (lhs > ei->rhs) {
      CPXmsg(ei->env->cpxlog, "The inequality is invalid for the %d-eqcol:", k);
      ECOshowvector(curcoloring);
          return ECOENUM_INFEASIBLE;
        }
        if (lhs == ei->rhs) {
          /* the solution is in the face */
          ei->numberofsolutionsface++;

          fprintf(ei->rootstream, "(%d) ", ei->numberofsolutionsface);
          for (v = 1; v <= vertices; v++) {
            for (j = 1; j <= vertices; j++) {
                  fprintf(ei->rootstream, "%d ", ECOvget(curcoloring, v) == j ? 1 : 0);
                }
          }
          for (j = 1; j <= vertices; j++) fprintf(ei->rootstream, "%d ", j <= k ? 1 : 0);
          fprintf(ei->rootstream, "\n");
        }
        return ECOENUM_OK;
  }

#ifdef ECODEBUGENUM
  printf("Vertex chosen: %d (impact=%d, subimpact=%d)\n", bestv, bestimpact, bestsubimpact);
#endif
  /* we scan feasible colors for "bestv" recursively */
  oldnumberofcolors = ei->curnumberofcolors;
  degree = ECOgetdegree(ei, bestv);
  for (j = 1; j <= vertices; j++) {
        flag = 0;
    for (n = 1; n <= degree; n++) {
      if (ECOvget(curcoloring, ECOgetneighbor(ei, bestv, n)) == j) {
        flag = 1;
                break;
          }
        }
        if (flag == 0) {
          ECOvset(curcoloring, bestv, j);
          ECOvincr(size_classes, j);
#ifdef ECODEBUGENUM
      printf("Set color %d to vertex %d. Now, class %d has %d vertices.\n", j, bestv, j, ECOvget(size_classes, j));
#endif
          ei->curnumberofcolors = j > oldnumberofcolors ? j : oldnumberofcolors;

          status = ECOcheckineqenum(ei, start_t, pass_t);
          switch (status) {
        case ECOENUM_OK: /* a solution has found */
                  break;
            case ECOENUM_INFEASIBLE: /* inequality is invalid */
                  return ECOENUM_INFEASIBLE;
        default:
          return status;
          }

          ECOvdecr(size_classes, j);
    }
  }

  /* we restore the state of vertex "bestv" */
  ECOvset(curcoloring, bestv, 0);
  ei->curnumberofcolors = oldnumberofcolors;
  return ECOENUM_OK;
}

