
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * files.c - write standard formulation in several formats
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static int writelp(char *filename, ECOGRPptr grp)
{
  FILE *stream;
  int vertices, max_color, v, j, feed, constraints;
  int e, v1, v2, status, k, coef;
  ECOVECptr isolated;

  vertices = grp->vertices;
  max_color = vertices;
  constraints = 0;
  isolated = ECOallocvector(&status, vertices);
  if (status) return status;

  stream = fopen(filename, "wt");
  if (stream == NULL) {
	ECOfreevector(&isolated);
    return ECOERR_CREAT;
  }

  /* objective function */
  fprintf(stream, "\nMinimize\n out:");
  feed = 0;
  for (j = 1; j <= max_color; j++) {
    fprintf(stream, " w%d", j);
	if (j < max_color) {
	  fprintf(stream, " +");
	  feed++;
	  if (feed == 10) {
	    fprintf(stream, "\n   ");
		feed = 0;
	  }
	}
  }

  /* assignment constraints. */
  fprintf(stream, "\nSubject To\n");
  for (v = 1; v <= vertices; v++) {
    fprintf(stream, " s%d:", ++constraints);
	feed = 0;
    for (j = 1; j <= max_color; j++) {
      fprintf(stream, " x%d_%d", v, j);
	  if (j < max_color) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " = 1\n");
  }

  /* edge constraints */
  ECOvfill(isolated, 1);
  for (e = 1; e <= grp->edges; e++) {
    v1 = ECOvget(grp->edge1, e);
    v2 = ECOvget(grp->edge2, e);
	ECOvset(isolated, v1, 0);
	ECOvset(isolated, v2, 0);
    for (j = 1; j <= max_color; j++) fprintf(stream, " s%d: x%d_%d + x%d_%d - w%d <= 0\n", ++constraints, v1, j, v2, j, j);
  }

  /* isolated vertices */
  for (v = 1; v <= vertices; v++) {
    if (ECOvget(isolated, v)) {
      for (j = 1; j <= max_color; j++) fprintf(stream, " s%d: x%d_%d - w%d <= 0\n", ++constraints, v, j, j);
	}
  }

  /* color order constraints */
  for (j = 1; j <= max_color - 1; j++) {
	fprintf(stream, " s%d: w%d - w%d >= 0\n", ++constraints, j, j+1);
  }

  /* equity constraints */
  for (j = 1; j <= max_color - 1; j++) {
	/* lower */
    fprintf(stream, " s%d:", ++constraints);
	feed = 0;
    for (v = 1; v <= vertices; v++) {
      fprintf(stream, " x%d_%d", v, j);
	  if (v < vertices) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " -");
    for (k = j; k <= max_color; k++) {
	  coef = ECOfloor(vertices, k);
	  if (k > j) coef = ECOfloor(vertices, k - 1) - coef;
	  fprintf(stream, " %d w%d", coef, k);
	  if (k < max_color) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " >= 0\n");
	/* upper */
    fprintf(stream, " s%d:", ++constraints);
	feed = 0;
    for (v = 1; v <= vertices; v++) {
      fprintf(stream, " x%d_%d", v, j);
	  if (v < vertices) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " -");
    for (k = j; k <= max_color; k++) {
	  coef = ECOceil(vertices, k);
	  if (k > j) coef = ECOceil(vertices, k - 1) - coef;
	  fprintf(stream, " %d w%d", coef, k);
	  if (k < max_color) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " <= 0\n");
  }

  /* variables. */
  fprintf(stream, "Binaries\n");
  feed = 0;
  for (v = 1; v <= vertices; v++) {
    for (j = 1; j <= max_color; j++) {
      fprintf(stream, " x%d_%d", v, j);
	  if (j < max_color) {
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n");
		  feed = 0;
	    }
	  }
	}
  }
  for (j = 1; j <= max_color; j++) {
    fprintf(stream, " w%d", j);
    if (j < max_color) {
	  feed++;
	  if (feed == 10) {
	    fprintf(stream, "\n");
	    feed = 0;
	  }
	}
  }
  fprintf(stream, "\nEnd\n");

  fclose(stream);
  return ECOERR_OK;
}

static int writeltx(char *filename, ECOGRPptr grp)
{
  FILE *stream;
  int vertices, max_color, v, j, feed, constraints;
  int e, v1, v2, status, k, coef;
  ECOVECptr isolated;

  vertices = grp->vertices;
  max_color = vertices;
  constraints = 0;
  isolated = ECOallocvector(&status, vertices);
  if (status) return status;

  stream = fopen(filename, "wt");
  if (stream == NULL) {
	ECOfreevector(&isolated);
    return ECOERR_CREAT;
  }

  /* objective function */
  fprintf(stream, "\nMIN ");
  feed = 0;
  for (j = 1; j <= max_color; j++) {
    fprintf(stream, " W%d", j);
	if (j < max_color) {
	  fprintf(stream, " +");
	  feed++;
	  if (feed == 10) {
	    fprintf(stream, "\n   ");
		feed = 0;
	  }
	}
  }

  /* assignment constraints. */
  fprintf(stream, "\nSUBJECT TO\n");
  for (v = 1; v <= vertices; v++) {
    fprintf(stream, " S%d)", ++constraints);
	feed = 0;
    for (j = 1; j <= max_color; j++) {
      fprintf(stream, " X%d_%d", v, j);
	  if (j < max_color) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " = 1\n");
  }

  /* edge constraints */
  ECOvfill(isolated, 1);
  for (e = 1; e <= grp->edges; e++) {
    v1 = ECOvget(grp->edge1, e);
    v2 = ECOvget(grp->edge2, e);
	ECOvset(isolated, v1, 0);
	ECOvset(isolated, v2, 0);
    for (j = 1; j <= max_color; j++) fprintf(stream, " S%d) X%d_%d + X%d_%d - W%d < 0\n", ++constraints, v1, j, v2, j, j);
  }

  /* isolated vertices */
  for (v = 1; v <= vertices; v++) {
    if (ECOvget(isolated, v)) {
      for (j = 1; j <= max_color; j++) fprintf(stream, " S%d) X%d_%d - W%d < 0\n", ++constraints, v, j, j);
	}
  }

  /* color order constraints */
  for (j = 1; j <= max_color - 1; j++) {
	fprintf(stream, " S%d) W%d - W%d >= 0\n", ++constraints, j, j+1);
  }

  /* equity constraints */
  for (j = 1; j <= max_color - 1; j++) {
	/* lower */
    fprintf(stream, " S%d)", ++constraints);
	feed = 0;
    for (v = 1; v <= vertices; v++) {
      fprintf(stream, " X%d_%d", v, j);
	  if (v < vertices) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " -");
    for (k = j; k <= max_color; k++) {
	  coef = ECOfloor(vertices, k);
	  if (k > j) coef = ECOfloor(vertices, k - 1) - coef;
	  fprintf(stream, " %d W%d", coef, k);
	  if (k < max_color) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " > 0\n");
	/* upper */
    fprintf(stream, " S%d)", ++constraints);
	feed = 0;
    for (v = 1; v <= vertices; v++) {
      fprintf(stream, " X%d_%d", v, j);
	  if (v < vertices) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " -");
    for (k = j; k <= max_color; k++) {
	  coef = ECOceil(vertices, k);
	  if (k > j) coef = ECOceil(vertices, k - 1) - coef;
	  fprintf(stream, " %d W%d", coef, k);
	  if (k < max_color) {
	    fprintf(stream, " +");
	    feed++;
	    if (feed == 10) {
	      fprintf(stream, "\n   ");
		  feed = 0;
	    }
	  }
	}
	fprintf(stream, " < 0\n");
  }

  /* variables. */
  fprintf(stream, "END\n");
  for (v = 1; v <= vertices; v++) {
    for (j = 1; j <= max_color; j++) {
      fprintf(stream, "INT X%d_%d\n", v, j);
	}
  }
  for (j = 1; j <= max_color; j++) {
    fprintf(stream, "INT W%d\n", j);
  }

  fclose(stream);
  return ECOERR_OK;
}

int ECOwriteformulation(ECOENVptr env, ECOGRPptr grp)
{
  char filename[ECO_MAXFILENAMESIZE];

  switch (env->par[ECO_PARAM_WRITEFILE]) {
    case ECO_FILE_NONE:
	  break;
	case ECO_FILE_LPFILE:
      strcpy(filename, env->filename);
      strcat(filename, ".lp");
      CPXmsg(env->cpxlog, "Writing LP formulation in %s.\n", filename);
	  writelp(filename, grp);
	  break;
	case ECO_FILE_LTXFILE:
      strcpy(filename, env->filename);
      strcat(filename, ".ltx");
      CPXmsg(env->cpxlog, "Writing LTX formulation in %s.\n", filename);
	  writeltx(filename, grp);
	  break;
	default:
	  return ECOERR_NOTIMPLEMENTED;
  }

  return ECOERR_OK;
}
