
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * formulation.c - problem generator
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

struct constraint {
  CPXENVptr cenv;
  CPXLPptr lp;
  struct ecointernal *ei;
  int constraints;
  int *rmatind;
  double *rmatval;
  double rhs;
  char sense;
  int nzcnt;
  char *name;
};

static void beginconstraint(struct constraint *co, char sense, int rhs, char *name)
{
  co->rhs = (double)rhs;
  co->sense = sense;
  co->nzcnt = 0;
  co->name = name;
}

static void addvar(struct constraint *co, int index, int coef)
{
  if (coef == 0) return; /* <-- var has a null coef. */
  if (index == 0) return; /* <-- var is fixed to zero. */
  if (index == 1) {
    /* if var is fixed to one, we pass it as a constant to the righthand side */
    co->rhs -= (double)coef;
    return;
  }

  /* var must be added */
  co->rmatind[co->nzcnt] = index - 2;
  co->rmatval[co->nzcnt] = (double)coef;
  co->nzcnt++;
}

static void addxvj(struct constraint *co, int v, int j, int coef)
{
  addvar(co, ECOmget(co->ei->xvar_state, v, j), coef);
}

static void addwj(struct constraint *co, int j, int coef)
{
  addvar(co, ECOvget(co->ei->wvar_state, j), coef);
}

static int endconstraint(struct constraint *co)
{
  int status, rmatbeg;
  char sense, rowname[32], *ptr[1];
  double cte;

  if (co->nzcnt == 0) {
    /* constraint empty */
#ifdef ECODEBUG
    switch (co->sense) {
      case 'E':
        if (co->rhs < -ECO_EPSILON || co->rhs > ECO_EPSILON) { /* rhs != 0 */
          printf("E) rhs = %lf\n", co->rhs);
          ECOabort(ECOERR_INTERNAL);
        }
        break;
      case 'L':
        if (co->rhs < -ECO_EPSILON) { /* rhs < 0 */
          printf("L) rhs = %lf\n", co->rhs);
          ECOabort(ECOERR_INTERNAL);
        }
        break;
      case 'G':
        if (co->rhs > ECO_EPSILON) { /* rhs > 0 */
          printf("G) rhs = %lf\n", co->rhs);
          ECOabort(ECOERR_INTERNAL);
        }
        break;
    }
#endif
    return ECOERR_OK;
  }
  if (co->nzcnt == 1) {
    cte = co->rhs / co->rmatval[0];
    sense = co->sense;
    if (co->rmatval[0] < -ECO_EPSILON) {
      /* we need to invert the sense if coef < 0 */
      switch (sense) {
        case 'E': break;
        case 'L': sense = 'G'; break;
        case 'G': sense = 'L'; break;
      }
    }
    switch (sense) {
      case 'E': /* var = cte --- normalize it */
#ifdef ECODEBUG
        if ((cte < -ECO_EPSILON || cte > ECO_EPSILON) && /* cte != 0 and cte != 1 */
            (cte < 1.0 - ECO_EPSILON || cte > 1.0 + ECO_EPSILON)) {
          printf("E) cte = %lf\n", cte);
          ECOabort(ECOERR_INTERNAL);
        }
#endif
        co->rmatval[0] = 1.0;
        co->rhs = cte;
        co->sense = 'E';
        break;
      case 'L': /* var <= cte --- if cte < 1 then var = 0 */
#ifdef ECODEBUG
        if (cte < -ECO_EPSILON) { /* cte < 0 */
          printf("L) cte = %lf\n", cte);
          ECOabort(ECOERR_INTERNAL);
        }
#endif
        if (cte >= 1.0) return ECOERR_OK;
        co->rmatval[0] = 1.0;
        co->rhs = 0.0;
        co->sense = 'E';
        break;
      case 'G': /* var >= cte --- if cte > 0 then var = 1 */
#ifdef ECODEBUG
        if (cte > 1.0 + ECO_EPSILON) { /* cte > 1 */
          printf("G) cte = %lf\n", cte);
          ECOabort(ECOERR_INTERNAL);
        }
#endif
        if (cte <= 0.0) return ECOERR_OK;
        co->rmatval[0] = 1.0;
        co->rhs = 1.0;
        co->sense = 'E';
        break;
    }
  }

  co->constraints++;
  rmatbeg = 0;
  sprintf(rowname, "%s%d", co->name, co->constraints);
  ptr[0] = rowname;
  status = CPXaddrows(co->cenv, co->lp, 0, 1, co->nzcnt, &co->rhs, &co->sense, &rmatbeg,
                      co->rmatind, co->rmatval, NULL, ptr);
  if (status) return status | ECOERR_CPLEX;
  return ECOERR_OK;
}

static int addconstraints(struct constraint *co)
{
  struct ecointernal *ei;
  int v, j, k, degree, vertices, max_color;
  int v1, v2, e, r, status;
  int u, vert;

  ei = co->ei;
  vertices = ei->grp->vertices;
  max_color = ei->max_color;

#ifdef ECODEBUGINIT
  printf("Generating assignment constraints...\n");
#endif
  for (v = 1; v <= vertices; v++) {
    beginconstraint(co, 'E', 1, "asg");
    for (j = 1; j <= max_color; j++) addxvj(co, v, j, 1);
    status = endconstraint(co);
    if (status) return status;
  }

  ei->adyac_substitutes = 0;
  if (ei->grp->edges <= ECO_MAXADYACSUST
	                     || ei->env->par[ECO_PARAM_FORM] == ECO_FORM_STANDARD
                         || ei->env->par[ECO_PARAM_FORM] == ECO_FORM_SYM1
						 || ei->env->par[ECO_PARAM_STRATEGY] != ECO_STRATEGY_BYC) { 
#ifdef ECODEBUGINIT
    printf("Generating edge constraints...\n");
#endif
    /* edge constraints */
    for (e = 1; e <= ei->grp->edges; e++) {
      v1 = ECOvget(ei->grp->edge1, e);
      v2 = ECOvget(ei->grp->edge2, e);

      /* for SYM1: use min(max_color, vertices/2) */
      for (j = 1; j <= (ei->env->par[ECO_PARAM_FORM] == ECO_FORM_SYM1 &&
                           ECOfloor(vertices, 2) < max_color ? ECOfloor(vertices, 2) : max_color); j++) {
        if (ECOmget(ei->xvar_state, v1, j) >= 2 && ECOmget(ei->xvar_state, v2, j) >= 2) {
		  /* we add the restriction only if two variables are not set, in order to avoid the following situation:
		   *   suppose we have edges (v1,v2), (v3, v2)
		   *   if x_v1j = 0, and x_v2j is not set, we add x_v2j - wj <= 0
		   *   if x_v3j = 0, and x_v2j is not set, we add x_v2j - wj <= 0
		   *   we added the same restriction two times!! */
          beginconstraint(co, 'L', 0, "ari");
          addxvj(co, v1, j, 1);
          addxvj(co, v2, j, 1);
          addwj(co, j, -1);
          status = endconstraint(co);
          if (status) return status;
	    }
      }
    }

    /* isolated vertices, or vertices that have a neighborhood with variables set to zero */
    for (v1 = 1; v1 <= vertices; v1++) {
      /* for SYM1: use min(max_color, vertices/2) */
      for (j = 1; j <= (ei->env->par[ECO_PARAM_FORM] == ECO_FORM_SYM1 &&
                           ECOfloor(vertices, 2) < max_color ? ECOfloor(vertices, 2) : max_color); j++) {
		if (ECOmget(ei->xvar_state, v1, j) >= 2) {
          degree = ECOgetdegree(ei, v1);
          k = 0;
          v2 = 1;
          while (v2 <= degree && k == 0) {
            if (ECOmget(ei->xvar_state, ECOgetneighbor(ei, v1, v2), j) != 0) k = 1;
#ifdef ECODEBUG
            if (ECOmget(ei->xvar_state, ECOgetneighbor(ei, v1, v2), j) == 1) {
			  printf("Variable x_%d%d must be set to zero because x_%d%d = 1.\n", v1, j, ECOgetneighbor(ei, v1, v2), j);
			  ECOabort(ECOERR_INTERNAL);
			}
#endif
            v2++;
	      }

          if (k == 0) {
			/* we add x_vj - w_j <= 0 */
            beginconstraint(co, 'L', 0, "iso");
            addxvj(co, v1, j, 1);
            addwj(co, j, -1);
            status = endconstraint(co);
            if (status) return status;
		  }
		}
      }
    }
  }
  else {
    ei->adyac_substitutes = 1;
    CPXmsg(ei->env->cpxlog, "Note: more than %d edges, activating alternative formulation.\n", ECO_MAXADYACSUST);
#ifdef ECODEBUGINIT
    printf("Generating neighborhood constraints...\n");
#endif
    for (v = 1; v <= vertices; v++) {
      /* we do not consider neighborhood constraints when x_vj is fixed */
      for (j = 1; j <= max_color; j++) {
        if (ECOmget(co->ei->xvar_state, v, j) > 1) {
          beginconstraint(co, 'L', 0, "nei");
          degree = ECOgetdegree(ei, v);
          /* if v is isolated, we make x_vj - w_j <= 0 */
          if (degree == 0) {
            addxvj(co, v, j, 1);
            addwj(co, j, -1);
          }
          else {
			/* TODO: quizas seria mejor si computamos theta(X) donde X son los vertices del vecindario de
			 *   v tales que x_vj no es cero. */
            r = ECOrvalue(ei, v, j, ei->min_color);
            addxvj(co, v, j, r);
            for (k = 1; k <= degree; k++) addxvj(co, ECOgetneighbor(ei, v, k), j, 1);
            addwj(co, j, -r);
          }
          status = endconstraint(co);
          if (status) return status;
        }
      }
    }
  }

#ifdef ECODEBUGINIT
  printf("Generating color order constraints...\n");
#endif
  for (j = 1; j <= max_color - 1; j++) {
    beginconstraint(co, 'G', 0, "ord");
    addwj(co, j, 1);
    addwj(co, j + 1, -1);
    status = endconstraint(co);
    if (status) return status;
  }

#ifdef ECODEBUGINIT
  printf("Generating equity constraints...\n");
#endif
  if (ei->env->par[ECO_PARAM_FORM] == ECO_FORM_SYM1) {
    /* |C_j| = tkj */
    for (j = 1; j <= max_color; j++) {
      beginconstraint(co, 'E', 0, "kqi");
      for (v = 1; v <= vertices; v++) addxvj(co, v, j, 1);
      addwj(co, j, -ECOtkj(j, j, vertices));
      for (k = j + 1; k <= max_color; k++)
        addwj(co, k, ECOtkj(k - 1, j, vertices) - ECOtkj(k, j, vertices));
      status = endconstraint(co);
      if (status) return status;
    }
  }
  else {
    /* floor(n/k) <= |C_j| <= ceil(n/k) */
    for (j = 1; j <= max_color; j++) {
      beginconstraint(co, 'G', 0, "kql");
      for (v = 1; v <= vertices; v++) addxvj(co, v, j, 1);
      addwj(co, j, -ECOfloor(vertices, j));
      for (k = j + 1; k <= max_color; k++)
        addwj(co, k, ECOfloor(vertices, k - 1) - ECOfloor(vertices, k));
      status = endconstraint(co);
      if (status) return status;

      beginconstraint(co, 'L', 0, "kqu");
      for (v = 1; v <= vertices; v++) addxvj(co, v, j, 1);
      addwj(co, j, -ECOceil(vertices, j));
      for (k = j + 1; k <= max_color; k++)
        addwj(co, k, ECOceil(vertices, k - 1) - ECOceil(vertices, k));
      status = endconstraint(co);
      if (status) return status;
    }
  }

  if (((ei->env->par[ECO_PARAM_STRATEGY] != ECO_STRATEGY_BYC) ||
	  (!(ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_EXTRA))) &&
	  (ei->env->par[ECO_PARAM_FORM] == ECO_FORM_SYM2A
      || ei->env->par[ECO_PARAM_FORM] == ECO_FORM_SYM2B 
      || ei->env->par[ECO_PARAM_FORM] == ECO_FORM_SYM2C)) {
	/* EXTRASYMMETRY: this constraints are added only if the formulation is M2 and we are not running a BYC or
	   we are running a BYC with the symmetry breakers disabled as cuts */
    CPXmsg(ei->env->cpxlog, "Note: extra symmetry constraints added in formulation.\n");
    for (v = 2; v <= vertices; v++) {
      vert = ECOvget(ei->ordering_symmetry, v);
      for (j = 2; j <= v - 1; j++) {
        if (ECOmget(ei->xvar_state, vert, j) != 0) {
          /* inequality: x_{vj} - sum_{u=j-1}^{v-1} x_{uj-1} <= 0, for v = 2,...,n and j = 2,...,v-1 */
          beginconstraint(co, 'L', 0, "xtr");
          addxvj(co, vert, j, 1);
          k = 0;
          for (u = j - 1; u <= v - 1; u++) {
			v2 = ECOvget(ei->ordering_symmetry, u);
			/* if x_uj-1 = 1 for some u, then the restriction is redundant */
			if (ECOmget(ei->xvar_state, v2, j - 1) == 1) k = 1;
            addxvj(co, v2, j - 1, -1);
          }
          if (k == 0) {
            status = endconstraint(co);
            if (status) return status;
		  }
		}
	  }
	}
  } 

  return ECOERR_OK;
}

int ECOprobgen(struct ecointernal *ei)
{
  CPXENVptr cenv;
  CPXLPptr lp;
  int v, j, status, variables, vertices, max_color;
  double obj, lb, ub;
  char ctype, colname[32], *ptr[1];
  struct constraint co;
  char lpfile[ECO_MAXFILENAMESIZE];

  /* creates a minimization problem */ 
  cenv = ei->env->cplexenv;
  lp = CPXcreateprob(cenv, &status, "formulation");
  if (lp == NULL) return status | ECOERR_CPLEX;
  CPXchgobjsen(cenv, lp, CPX_MIN);
#if CPX_VERSION >= 800
  CPXchgprobtype(cenv, lp, CPXPROB_MILP);
#else
  CPXchgprobtype(cenv, lp, CPXPROB_MIP);
#endif

  ctype = 'B';
  lb = 0.0;
  ub = 1.0;

  /* add variables x_vj */
  max_color = ei->max_color;
  vertices = ei->grp->vertices;
  variables = 0;
  for (v = 1; v <= vertices; v++) {
    for (j = 1; j <= vertices; j++) {
      if (ECOmget(ei->xvar_state, v, j) > 1) {
		if (j > max_color) {
          /* set x_vj = 0 if j > max_color
		   (perhaps, tabu serch improved the upper bound) */
          ECOmset(ei->xvar_state, v, j, 0);
	    }
	    else {
          /* add if x_vj is not fixed */
          obj = 0.0;
          sprintf(colname, "x%d_%d", v, j);
          ptr[0] = colname;
          status = CPXnewcols(cenv, lp, 1, &obj, &lb, &ub, &ctype, ptr);
          if (status) return status | ECOERR_CPLEX;
          ECOmset(ei->xvar_state, v, j, 2 + variables);
          variables++;
	    }
      }
    }
  }

  /* add variables w_j */
  for (j = 1; j <= vertices; j++) {
    if (ECOvget(ei->wvar_state, j) > 1) {
      if (j > max_color) {
        /* set w_j = 0 if j > max_color
		 (perhaps, tabu serch improved the upper bound) */
         ECOvset(ei->wvar_state, j, 0);      
	  }
	  else {
        /* add if w_j is not fixed */
        obj = 1.0;
        sprintf(colname, "w%d", j);
        ptr[0] = colname;
        status = CPXnewcols(cenv, lp, 1, &obj, &lb, &ub, &ctype, ptr);
        if (status) return status | ECOERR_CPLEX;
        ECOvset(ei->wvar_state, j, 2 + variables);
        variables++;
	  }
    }
  }

  /* note that w_j with j = 1, ..., min_color does not appear in the objective function,
   * so we must add to the objective function a variable with the value min_color */
  obj = 1.0;
  ctype = 'I';
  lb = (double)ei->min_color;
  ub = (double)ei->min_color;
  sprintf(colname, "min_color");
  ptr[0] = colname;
  status = CPXnewcols(cenv, lp, 1, &obj, &lb, &ub, &ctype, ptr);
  if (status) return status | ECOERR_CPLEX;
  variables++;

  /* add constraints */
  co.cenv = cenv;
  co.lp = lp;
  co.ei = ei;
  co.constraints = 0;
  co.rmatind = malloc(sizeof(int)*variables);
  if (co.rmatind == NULL) return ECOERR_MEMORY;
  co.rmatval = malloc(sizeof(double)*variables);
  if (co.rmatval == NULL) {
    free(co.rmatind);
    return ECOERR_MEMORY;
  }
  status = addconstraints(&co);
  if (status) {
    free(co.rmatval);
    free(co.rmatind);
    return status;
  }

  /* reserve memory for ei->x (we can use vector rmatval) */
  ei->x = co.rmatval;

  if (ei->env->par[ECO_PARAM_WRITEFILE] == ECO_FILE_LPFILE) {
    strcpy(lpfile, ei->env->filename);
    strcat(lpfile, ".lp");
    CPXwriteprob(cenv, lp, lpfile, NULL);
    CPXmsg(ei->env->cpxlog, "LP formulation written in %s.\n", lpfile);
  }
  CPXmsg(ei->env->cpxlog, "The formulation has %d constraints and %d variables.\n",
         co.constraints, variables);

  ei->lp = lp;
  ei->variables = variables;
  ei->constraints = co.constraints;
  free(co.rmatind);
  return ECOERR_OK;
}
