
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * general.c - general routines
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#ifdef ECONOWALL
  #ifdef ECOPARALLEL
    #include <omp.h>
  #else
    #include <unistd.h>
    #include <sys/times.h>
  #endif
#endif

static const char *errlist[] = {
  "All ok", /* ECOERR_OK */
  "Too many edges", /* ECOERR_MANYEDGES */
  "Not enough memory", /* ECOERR_MEMORY */
  "No environment", /* ECOERR_NOENV */
  "No graph", /* ECOERR_NOGRP */
  "The vertex does not exists", /* ECOERR_NOTVERTEX */
  "Loops are not allowed", /* ECOERR_LOOPS */
  "The graph is out of range", /* ECOERR_OUTOFRANGE */
  "Option not implemented yet", /* ECOERR_NOTIMPLEMENTED */
  "Error creating file", /* ECOERR_CREAT */
  "Internal error. Check source code", /* ECOERR_INTERNAL */
  "Trying to free a NULL pointer", /* ECOERR_NULLP */
  "Access out of bounds", /* ECOERR_BOUNDS */
  "Formulation incompatible with MIP strategy", /* ECOERR_INCOMPAT */
  "Unknown error" /* ECOERR_LASTCODE + 1 */
};

double divisor = 0.0;

double ECOclock()
{
#ifdef ECONOWALL
  #ifdef ECOPARALLEL
    /* on multi-threaded system in Linux, we need a precise wall-clock time */
    return omp_get_wtime();
  #else
    /* measure user-time: use it on single-threaded system in Linux (is more accurate) */
    struct tms buf;
    times(&buf);
    return ((double)buf.tms_utime) / divisor;
  #endif
#else
  /* measure standard wall-clock: use it on Windows
   * (Linux clock() function overflows every 72 minutes) */
  return ((double)clock()) / divisor;
#endif
}

void CPXPUBLIC CPXmsgcallback(void *handle, const char *str)
{
  ECOENVptr env = (ECOENVptr)handle;
  env->msgfunction(str);
}

ECOENVptr ECOopen(int *status, void(*msgfunction)(const char *), char *filename)
{
  ECOENVptr env;
  int cplexstatus;
  CPXENVptr cplexenv;

  env = malloc(sizeof(struct ecoenv));
  if (env == NULL) {
    *status = ECOERR_MEMORY;
    return NULL;
  }

  cplexenv = CPXopenCPLEX(&cplexstatus);
  if (cplexenv == NULL) {
    *status = cplexstatus | ECOERR_CPLEX;
    free(env);
    return NULL;
  }

  /* set params */
  env->cplexenv = cplexenv;
  env->par[ECO_PARAM_TILIM] = ECO_MAXTIME;
  env->par[ECO_PARAM_ITLIM] = ECO_MAXITER;
  env->par[ECO_PARAM_LOWERB] = ECO_ON;
  env->par[ECO_PARAM_STRATEGY] = ECO_STRATEGY_BYC;
  env->par[ECO_PARAM_UPPERB] = ECO_UPPERB_TABU;
  env->par[ECO_PARAM_FORM] = ECO_FORM_SYM2B;
  env->par[ECO_PARAM_ALPHA] = ECO_TABUALPHA;
  env->par[ECO_PARAM_BETA] = ECO_TABUBETA;
  env->par[ECO_PARAM_USERUB] = -1;
  env->par[ECO_PARAM_BRANCH] = ECO_BRANCH_CALLBACK;
  env->par[ECO_PARAM_NODESEL] = ECO_NODESEL_BESTBOUND;
  env->par[ECO_PARAM_MIPINTERVAL] = 100;
  env->par[ECO_PARAM_THREADS] = 1;
  env->par[ECO_PARAM_USERLB] = -1;
  env->par[ECO_PARAM_CUTPLANE] = 335;
  env->par[ECO_PARAM_PRIHEUR] = 7;
  env->par[ECO_PARAM_WRITEFILE] = ECO_FILE_NONE;
  env->par[ECO_PARAM_SKIPFACTORCUT] = ECO_SKIPFACTORCUT;
  env->par[ECO_PARAM_MAXROUNDSATROOT] = ECO_MAXROUNDSATROOT;
  env->par[ECO_PARAM_MAXROUNDSATNODE] = ECO_MAXROUNDSATNODE;
  env->par[ECO_PARAM_MAXDEPTHTREE] = ECO_MAXDEPTHTREE;
  env->par[ECO_PARAM_MAXTIMEHEUR] = ECO_MAXTIMEHEUR;
  env->par[ECO_PARAM_IMPENUM] = ECO_ON;
  env->par[ECO_PARAM_IMPVERT] = 0;

  env->bestobjval = 1.0;
  env->iter_result = 0;
  env->nodes_result = 0;
  env->timecut_result = 0;
  env->timeopt_result = 0;
  env->msgfunction = msgfunction;
  env->filename = filename;

  /* capture every message */
  CPXgetchannels(cplexenv, &env->cpxresults, &env->cpxwarning, &env->cpxerror, &env->cpxlog);
  CPXaddfuncdest(cplexenv, env->cpxresults, env, CPXmsgcallback);
  CPXaddfuncdest(cplexenv, env->cpxwarning, env, CPXmsgcallback);
  CPXaddfuncdest(cplexenv, env->cpxerror, env, CPXmsgcallback);
  CPXaddfuncdest(cplexenv, env->cpxlog, env, CPXmsgcallback);

  CPXsetintparam(cplexenv, CPX_PARAM_SCRIND, CPX_OFF);
#ifdef ECONOWALL
  #ifdef ECOPARALLEL
    divisor = 1.0 / omp_get_wtick();
  #else
    divisor = (double)sysconf(_SC_CLK_TCK);
  #endif
#else
  divisor = (double)CLOCKS_PER_SEC;
#endif
  *status = ECOERR_OK;
  return env;
}

int ECOclose(ECOENVptr *env)
{
  int cplexstatus;
  CPXENVptr cplexenv;

  if (*env == NULL) return ECOERR_NOENV;

  cplexenv = (*env)->cplexenv;
  free(*env);
  *env = NULL;

  cplexstatus = CPXcloseCPLEX(&cplexenv);
  if (cplexstatus) return cplexstatus | ECOERR_CPLEX;

  return ECOERR_OK;
}

void ECOgeterrorstring(ECOENVptr env, int errcode, char *errmsg)
{
  CPXENVptr cplexenv = NULL;

  if (errcode & ECOERR_CPLEX) {
    if (env != NULL) cplexenv = env->cplexenv;
    CPXgeterrorstring(cplexenv, errcode - ECOERR_CPLEX, errmsg);
    return;
  }

  if (errcode < 0 || errcode > ECOERR_LASTCODE) errcode = ECOERR_LASTCODE + 1;
  sprintf(errmsg, "ECOPT Error %d: %s.", errcode, errlist[errcode]);
}

void ECOabort(int errcode)
{
  char errmsg[1024];

  ECOgeterrorstring(NULL, errcode, errmsg);
  printf("ECOabort: %s\n", errmsg);
  abort();
}

int ECOgetparam(ECOENVptr env, int param)
{
  if (env == NULL) return -1;
  if (param < 0 || param > ECO_PARAM_LASTCODE) return -1;

  return env->par[param];
}

void ECOsetparam(ECOENVptr env, int param, int value)
{
  if (env == NULL) return;
  if (param < 0 || param > ECO_PARAM_LASTCODE) return;

  env->par[param] = value;
}

ECOGRPptr ECOcreategraph(ECOENVptr env, int *status, int vertices, int maxedges)
{
  ECOGRPptr grp;
  int v;

  if (env == NULL) {
    *status = ECOERR_NOENV;
    return NULL;
  }
  if (vertices < ECO_MINVERTICES || vertices > ECO_MAXVERTICES
      || maxedges < 0 || maxedges > ECO_MAXEDGES) {
    *status = ECOERR_OUTOFRANGE;
    return NULL;
  }

  /* alloc memory for structures */
  grp = malloc(sizeof(struct ecogrp));
  if (grp == NULL) {
    *status = ECOERR_MEMORY;
    return NULL;
  }

  grp->coloring = ECOallocvector(status, vertices);
  if (*status) {
    free(grp);
    return NULL;
  }

  if (maxedges > 0) {
    grp->edge1 = ECOallocvector(status, maxedges);
    if (*status) {
      ECOfreevector(&grp->coloring);
      free(grp);
      return NULL;
    }
    grp->edge2 = ECOallocvector(status, maxedges);
    if (*status) {
      ECOfreevector(&grp->edge1);
      ECOfreevector(&grp->coloring);
      free(grp);
      return NULL;
    }
    grp->hash_vertices = malloc(sizeof(struct ecolist) * vertices);
    if (grp->hash_vertices == NULL) {
      ECOfreevector(&grp->edge2);
      ECOfreevector(&grp->edge1);
      ECOfreevector(&grp->coloring);
      free(grp);
      return NULL;
    }
	for (v = 1; v <= vertices; v++) {
	  grp->hash_vertices[v-1].value = 0;
	  grp->hash_vertices[v-1].next = NULL;
	}
  }
  else {
    /* isolated graph */
    grp->edge1 = NULL;
    grp->edge2 = NULL;
    grp->hash_vertices = NULL;
  }

  /* set default values */
  grp->vertices = vertices;
  grp->maxedges = maxedges;
  grp->edges = 0;
  for (v = 1; v <= vertices; v++) ECOvset(grp->coloring, v, v);
  grp->colors_used = vertices;
  grp->is_optimal = 0;
  *status = ECOERR_OK;
  return grp;
}

int ECOfreegraph(ECOENVptr env, ECOGRPptr *grp)
{
  struct ecolist *hv, *hvnext;
  int v;
  if (env == NULL) return ECOERR_NOENV;
  if (*grp == NULL) return ECOERR_NOGRP;

  ECOfreevector(&(*grp)->coloring);
  if ((*grp)->maxedges > 0) {
    ECOfreevector(&(*grp)->edge1);
    ECOfreevector(&(*grp)->edge2);
    if ((*grp)->hash_vertices != NULL) {
      for (v = 1; v <= (*grp)->vertices; v++) {
	    hv = (*grp)->hash_vertices[v-1].next;
		if (hv != NULL) {
		  do {
            hvnext = hv->next;
			free(hv);
			hv = hvnext;
		  } while (hv != NULL);
		}
      }
	  free((*grp)->hash_vertices);
	}
  }

  free(*grp);
  *grp = NULL;

  return ECOERR_OK;
}

int ECOaddedge(ECOENVptr env, ECOGRPptr grp, int u, int v)
{
  int w, uu, vv, edges_loaded;
  struct ecolist *hv;

  if (env == NULL) return ECOERR_NOENV;
  if (grp == NULL) return ECOERR_NOGRP;
  if (u < 1 || u > grp->vertices) return ECOERR_NOTVERTEX;
  if (v < 1 || v > grp->vertices) return ECOERR_NOTVERTEX;
  if (u == v) return ECOERR_LOOPS;
  if (u > v) {
    w = u;
    u = v;
    v = w;
  }

  edges_loaded = grp->edges;
  /**********************************************
  ** THIS CODE IS SLOW!!
  ** test if the edge exists
  for (e = 1; e <= edges_loaded; e++) {
    uu = ECOvget(grp->edge1, e);
    vv = ECOvget(grp->edge2, e);
    if (u == uu && v == vv) return ECOERR_OK;
  }
  ***********************************************
   * Now we use a list "hash_vertices":
   * The first element is the number of vertices adjacent to the entry vertex */

  /* look for the entry with less elements */
  if (grp->hash_vertices[u-1].value < grp->hash_vertices[v-1].value) {
	uu = u;
	vv = v;
  }
  else {
	uu = v;
	vv = u;
  }
  /* test if the edge exists */
  hv = &(grp->hash_vertices[uu-1]);
  while (hv->next != NULL) {
	hv = hv->next;
	if (hv->value == vv) return ECOERR_OK;
  }

  /* add the new edge */
  if (edges_loaded >= grp->maxedges) return ECOERR_MANYEDGES;
  edges_loaded++;
  ECOvset(grp->edge1, edges_loaded, u);
  ECOvset(grp->edge2, edges_loaded, v);
  grp->edges = edges_loaded;

  /* add the first entry */
  hv = &(grp->hash_vertices[u-1]);
  hv->value++;
  while (hv->next != NULL) hv = hv->next;
  hv->next = malloc(sizeof(struct ecolist));
  if (hv->next == NULL) ECOabort(ECOERR_MEMORY);
  hv = hv->next;
  hv->value = v;
  hv->next = NULL;

  /* add the second entry */
  hv = &(grp->hash_vertices[v-1]);
  hv->value++;
  while (hv->next != NULL) hv = hv->next;
  hv->next = malloc(sizeof(struct ecolist));
  if (hv->next == NULL) ECOabort(ECOERR_MEMORY);
  hv = hv->next;
  hv->value = u;
  hv->next = NULL;

  return ECOERR_OK;
}

int ECOgetedges(ECOENVptr env, ECOGRPptr grp)
{
  if (env == NULL) return -1;
  if (grp == NULL) return -1;
  return grp->edges;
}

int ECOshowgraph(ECOGRPptr grp)
{
  int e, edges, carry, u, v;

  if (grp == NULL) return ECOERR_NOGRP;
  edges = grp->edges;

  printf("The graph has %d vertices and %d edges.\n", grp->vertices, edges);
  printf("The edges are:");
  carry = 0;
  for (e = 1; e <= edges; e++) {
    u = ECOvget(grp->edge1, e);
    v = ECOvget(grp->edge2, e);
    printf("  (%d,%d)", u, v);
    carry++;
    if (carry == 10) {
      printf("\n  ");
      carry = 0;
    }
  }
  printf(".\n");

  return ECOERR_OK;
}

int* ECOcolresult(ECOGRPptr grp, int *status, int *is_optimal, int *colors_used)
{
  int *coloring;
  int vertices, v;

  if (grp == NULL) {
    *status = ECOERR_NOGRP;
    return NULL;
  }

  vertices = grp->vertices;
  coloring = malloc(sizeof(int)*vertices);
  if (coloring == NULL) {
    *status = ECOERR_MEMORY;
    return NULL;
  }

  for (v = 1; v <= vertices; v++) coloring[v - 1] = ECOvget(grp->coloring, v);
  *status = ECOERR_OK;
  *is_optimal = grp->is_optimal;
  *colors_used = grp->colors_used;
  return coloring;
}

double ECOgetbestobjval(ECOENVptr env)
{
  if (env == NULL) return -1.0;
  return env->bestobjval;
}

long long ECOgetitcnt(ECOENVptr env)
{
  if (env == NULL) return -1;
  return env->iter_result;
}

long long ECOgetnodecnt(ECOENVptr env)
{
  if (env == NULL) return -1;
  return env->nodes_result;
}

int ECOgettimeopt(ECOENVptr env)
{
  if (env == NULL) return -1;
  return env->timeopt_result;
}

int ECOgettimecut(ECOENVptr env)
{
  if (env == NULL) return -1;
  return env->timecut_result;
}

void ECOsetinternal(struct ecointernal *ei)
{
  ei->degree = NULL;
  ei->ascending_by_degree = NULL;
  ei->descending_by_degree = NULL;
  ei->sl_ordering = NULL;
  ei->neighbormaxindep1_size = NULL;
  ei->neighbormaxindep2_size = NULL;
  ei->clique_set = NULL;
  ei->ordering_symmetry = NULL;
  ei->xvar_state = NULL;
  ei->wvar_state = NULL;
  ei->lp = NULL;
  ei->rootstream = NULL;
  ECOsetinternal_onlycallback(ei);
}

void ECOsetinternal_onlycallback(struct ecointernal *ei) {
  ei->tabu_bestcoloring = NULL;
  ei->tabu_curcoloring = NULL;
  ei->tabu_conflictvertexset = NULL;
  ei->tabu_conflictvertexlist = NULL;
  ei->tabu_bigclasses = NULL;
  ei->tabu_coloredvert = NULL;
  ei->tabu_tenure = NULL;
  ei->x = NULL;
  ei->cu.cutind = NULL;
  ei->cu.cutval = NULL;
  ei->cut_temp1 = NULL;
  ei->cut_temp2 = NULL;
  ei->cut_descending = NULL;
  ei->cut_forbidden = NULL;
  ei->clique_cut_set = NULL;
  ei->clique_ujks = NULL;
  ei->alpha2_cut_set = NULL;
  ei->alpha2clique_cut_set = NULL;
  ei->alpha2_cut_temp1 = NULL;
  ei->alpha2_cut_temp2 = NULL;
  ei->alpha2_cut_temp3 = NULL;
  ei->cs_cut_index = NULL;
  ei->cs_cut_value = NULL;
  ei->cs_cut_descending = NULL;
  ei->cs_cut_c_set = NULL;
  ei->cs_cut_s_set = NULL;
  ei->cs_cut_forbidden = NULL;
  ei->primal_coloring = NULL;
  ei->primal_vector1 = NULL;
  ei->primal_vector2 = NULL;
  ei->primal_vector3 = NULL;
  ei->primal_vector4 = NULL;
  ei->primal_vector5 = NULL;
  ei->primal_vector6 = NULL;
  ei->intfeas = NULL;
  ei->varind = NULL;
  ei->varlu = NULL;
  ei->varbd = NULL;
}

void ECOfreemem(struct ecointernal *ei)
{
  int vertices, v;

  ECOfreemem_onlycallback(ei);
  if (ei->lp != NULL) CPXfreeprob(ei->env->cplexenv, &ei->lp);
  ECOfreevector(&ei->ordering_symmetry);
  ECOfreevector(&ei->wvar_state);
  ECOfreematrix(&ei->xvar_state);
  ECOfreevector(&ei->ordering_symmetry);
  ECOfreevector(&ei->clique_set);
  ECOfreevector(&ei->neighbormaxindep2_size);
  ECOfreevector(&ei->neighbormaxindep1_size);
  ECOfreevector(&ei->sl_ordering);
  ECOfreevector(&ei->descending_by_degree);
  ECOfreevector(&ei->ascending_by_degree);

  /* free struct degree */
  if (ei->degree != NULL) {
    vertices = ei->grp->vertices;
    for (v = 1; v <= vertices; v++) ECOfreevector(&ei->degree[v - 1]);
    free(ei->degree);
  }

#ifdef ECODEF_WRITEROOTFILE
  if (ei->rootstream != NULL) fclose(ei->rootstream);
#endif
}

void ECOfreemem_onlycallback(struct ecointernal *ei)
{
  if (ei->varbd != NULL) free(ei->varbd);
  if (ei->varlu != NULL) free(ei->varlu);
  if (ei->varind != NULL) free(ei->varind);
  if (ei->intfeas != NULL) free(ei->intfeas);
  ECOfreevector(&ei->primal_vector6);
  ECOfreevector(&ei->primal_vector5);
  ECOfreevector(&ei->primal_vector4);
  ECOfreevector(&ei->primal_vector3);
  ECOfreevector(&ei->primal_vector2);
  ECOfreevector(&ei->primal_vector1);
  ECOfreevector(&ei->primal_coloring);
  ECOfreevector(&ei->cs_cut_forbidden);
  ECOfreevector(&ei->cs_cut_s_set);
  ECOfreevector(&ei->cs_cut_c_set);
  ECOfreevector(&ei->cs_cut_descending);
  ECOfreevector(&ei->cs_cut_value);
  ECOfreevector(&ei->cs_cut_index);
  ECOfreevector(&ei->alpha2_cut_temp3);
  ECOfreevector(&ei->alpha2_cut_temp2);
  ECOfreevector(&ei->alpha2_cut_temp1);
  ECOfreevector(&ei->alpha2clique_cut_set);
  ECOfreevector(&ei->alpha2_cut_set);
  ECOfreevector(&ei->clique_ujks);
  ECOfreevector(&ei->clique_cut_set);
  ECOfreevector(&ei->cut_forbidden);
  ECOfreevector(&ei->cut_descending);
  ECOfreevector(&ei->cut_temp2);
  ECOfreevector(&ei->cut_temp1);

  /* free struct cu */
  if (ei->cu.cutval != NULL) free(ei->cu.cutval);
  if (ei->cu.cutind != NULL) free(ei->cu.cutind);

  if (ei->x != NULL) free(ei->x);

  ECOfreematrix(&ei->tabu_tenure);
  ECOfreevector(&ei->tabu_coloredvert);
  ECOfreevector(&ei->tabu_bigclasses);
  ECOfreevector(&ei->tabu_conflictvertexlist);
  ECOfreevector(&ei->tabu_conflictvertexset);
  ECOfreevector(&ei->tabu_curcoloring);
  ECOfreevector(&ei->tabu_bestcoloring);
}
