
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * greedy.c - some greedy heuristics
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

/* This routine finds (or maximalizes) a clique on the graph by scanning vertices according
 * to 'ordering' but skipping those that appears in 'forbidden'. The clique is saved in
 * 'clique_set' (memory needs to be allocated before calling). */
int ECOgetclique(struct ecointernal *ei, ECOVECptr ordering, ECOVECptr forbidden,
                 ECOVECptr clique_set, int flag)
{
  int clique_size, i, v, k, vertices, degree, status;
  ECOVECptr candidates, noneighbor;

  /* ask for memory */
  vertices = ei->grp->vertices;
  candidates = ECOallocvector(&status, vertices);
  if (status) return status;
  noneighbor = ECOallocvector(&status, vertices);
  if (status) {
    ECOfreevector(&candidates);
    return status;
  }

  /* determine size of the current clique */
  clique_size = ECOvsize(clique_set);
  ECOvmaxsize(clique_set);

  /* make a list of possible candidates */
  ECOvfill(candidates, 1);
  for (i = 1; i <= clique_size; i++) {
    v = ECOvget(clique_set, i);
    degree = ECOgetdegree(ei, v);
    ECOvfill(noneighbor, 1);
//    ECOvset(noneighbor, v, 0);
    for (k = 1; k <= degree; k++) ECOvset(noneighbor, ECOgetneighbor(ei, v, k), 0);
    /* now, noneighbor = V - N(v) --> these vertices are not candidates */
    for (k = 1; k <= vertices; k++) {
      if (ECOvget(noneighbor, k) == 1) ECOvset(candidates, k, 0);
    }
  }

  /* add candidates to the clique */
  for (i = 1; i <= vertices; i++) {
    v = ECOvget(ordering, i);
    if (ECOvget(forbidden, v) == flag && ECOvget(candidates, v) == 1) {
      clique_size++;
      ECOvset(clique_set, clique_size, v);
      /* reduce candidates */
      degree = ECOgetdegree(ei, v);
      ECOvfill(noneighbor, 1);
      ECOvset(noneighbor, v, 0);
      for (k = 1; k <= degree; k++) ECOvset(noneighbor, ECOgetneighbor(ei, v, k), 0);
      /* now, noneighbor = V - N[v] --> these vertices are not candidates */
      for (k = 1; k <= vertices; k++) {
        if (ECOvget(noneighbor, k) == 1) ECOvset(candidates, k, 0);
      }
    }
  }
  ECOvredim(clique_set, clique_size);

  ECOfreevector(&noneighbor);
  ECOfreevector(&candidates);
  return ECOERR_OK;
}

/* This routine finds a maximal independent set on the graph according to 'ordering'
 * but skipping those that appears in 'forbidden'. The indep. set is saved in 'indep_set'
 * (memory needs to be allocated before calling) and the size is returned. */
int ECOgetindepset(struct ecointernal *ei, ECOVECptr ordering, ECOVECptr forbidden,
                   ECOVECptr indep_set)
{
  int i, v, k, l, indep_size, degree, vertices, flag;

  indep_size = 0;
  vertices = ei->grp->vertices;
  for (i = 1; i <= vertices; i++) {
    v = ECOvget(ordering, i);
    if (ECOvget(forbidden, v) == 1) continue;
    degree = ECOgetdegree(ei, v);
    k = 1;
    flag = 1;
    while (k <= indep_size && flag == 1) {
      l = 1;
      while (l <= degree && flag == 1) {
        /* is 'v' adjacent to the kth. vertex of the independent set? */
        if (ECOgetneighbor(ei, v, l) == ECOvget(indep_set, k)) flag = 0;
        l++;
      }
      k++;
    }
    if (flag == 1) {
      /* 'v' can be added to the indep. set */
      indep_size++;
      ECOvset(indep_set, indep_size, v);
    }
  }
  return indep_size;
}

/* Finds an independent set from a list of vertices using a greedy heuristic,
 * useful for knowing a lower bound of alpha(list). */
int ECOgreedyindepset(struct ecointernal *ei, ECOVECptr list_set,
                      ECOVECptr indep_set, int *indep_size)
{
  ECOVECptr current_indep_set, forbidden;
  int vertices, current_indep_size, v, flag, status;

  *indep_size = 0;
  if (ECOvsize(list_set) == 0) return ECOERR_OK;

  vertices = ei->grp->vertices;
  current_indep_set = ECOallocvector(&status, vertices);
  if (status) return status;
  forbidden = ECOallocvector(&status, vertices);
  if (status) {
    ECOfreevector(&current_indep_set);
    return status;
  }

  /* set only the vertices in the list */
  ECOvfill(forbidden, 1);
  for (v = 1; v <= ECOvsize(list_set); v++) ECOvset(forbidden, ECOvget(list_set, v), 0);

  /* computes the first indepedent set (using ascending order) */
  *indep_size = ECOgetindepset(ei, ei->ascending_by_degree, forbidden, indep_set);
  for (v = 1; v <= *indep_size; v++) ECOvset(forbidden, ECOvget(indep_set, v), 1);

  /* computes other indep. sets and request the greatest one */
  do {
    current_indep_size = ECOgetindepset(ei, ei->ascending_by_degree, forbidden, current_indep_set);
    if (current_indep_size > *indep_size) {
      for (v = 1; v <= current_indep_size; v++)
        ECOvset(indep_set, v, ECOvget(current_indep_set, v));
      *indep_size = current_indep_size;
    }
    for (v = 1; v <= current_indep_size; v++) ECOvset(forbidden, ECOvget(current_indep_set, v), 1);
    /* does exists some allowed vertices yet? */
    flag = 1;
    for (v = 1; v <= vertices; v++) flag &= ECOvget(forbidden, v);
  } while (flag == 0);

  ECOfreevector(&forbidden);
  ECOfreevector(&current_indep_set);
  return ECOERR_OK;
}

/* Finds a partition of vertices by cliques from a list of vertices using a greedy heuristic,
 * useful for knowing an upper bound of alpha(list). */
int ECOgreedypartition(struct ecointernal *ei, ECOVECptr list_set, int *partition_size)
{
  ECOVECptr clique_set, forbidden;
  int vertices, clique_size, v, flag, status;

  *partition_size = 0;
  if (ECOvsize(list_set) == 0) return ECOERR_OK;

  vertices = ei->grp->vertices;
  clique_set = ECOallocvector(&status, vertices);
  if (status) return status;
  forbidden = ECOallocvector(&status, vertices);
  if (status) {
    ECOfreevector(&clique_set);
    return status;
  }

  /* set only the vertices in the list */
  ECOvfill(forbidden, 1);
  for (v = 1; v <= ECOvsize(list_set); v++) ECOvset(forbidden, ECOvget(list_set, v), 0);

  /* compute cliques until all vertices are consumed */
  do {
    ECOvredim(clique_set, 0);
    status = ECOgetclique(ei, ei->descending_by_degree, forbidden, clique_set, 0);
    if (status) {
      ECOfreevector(&forbidden);
      ECOfreevector(&clique_set);
      return status;
    }
    clique_size = ECOvsize(clique_set);
    for (v = 1; v <= clique_size; v++) ECOvset(forbidden, ECOvget(clique_set, v), 1);
    (*partition_size)++;
    /* does exists some allowed vertices yet? */
    flag = 1;
    for (v = 1; v <= vertices; v++) flag &= ECOvget(forbidden, v);
  } while (flag == 0);

  ECOfreevector(&forbidden);
  ECOfreevector(&clique_set);
  return ECOERR_OK;
}
