
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * lowerbound.c - initial heuristics for generating a lower bound of chieq
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOlowerbound(struct ecointernal *ei)
{
  int status, min_color, max_clique, userlb;

  CPXmsg(ei->env->cpxlog, "Initial lower bound: ");
  if (ei->env->par[ECO_PARAM_LOWERB] == ECO_ON) {
    CPXmsg(ei->env->cpxlog, "ON.\n");
    /* computes a bound */
    status = ECOalternatebound(ei, &min_color);
    if (status) return status;
    /* computes a maximal clique */
    status = ECOgreedyclique(ei, &max_clique);
    if (status) return status;
    /* use the biggest one as lower bound */
    if (max_clique > min_color) min_color = max_clique;
  }
  else {
    CPXmsg(ei->env->cpxlog, "OFF.\n");
    /* use an edge as initial clique */
    ei->clique_set = ECOallocvector(&status, 2);
    if (status) return status;
    ECOvset(ei->clique_set, 1, ECOvget(ei->grp->edge1, 1));
    ECOvset(ei->clique_set, 2, ECOvget(ei->grp->edge2, 1));
    max_clique = 2;
    min_color = 2;
  }

#ifdef ECODEBUGINIT
  printf("Maximal clique found:");
  ECOshowvector(ei->clique_set);
#endif

  /* LB is defined by user */
  userlb = ei->env->par[ECO_PARAM_USERLB];
  if (userlb > min_color) {
    CPXmsg(ei->env->cpxlog, "Using lower bound supplied by user.\n");
    min_color = userlb;
  }

  CPXmsg(ei->env->cpxlog, "Size of maximal clique: %d.\nLower bound: %d.\n", max_clique, min_color);
  ei->min_color = min_color;
  ei->env->bestobjval = (double)min_color;
  return ECOERR_OK;
}

int ECOalternatebound(struct ecointernal *ei, int *min_color)
{
  ECOVECptr list_set;
  int vertices, v, i, d, degree, theta, minval, flag;
  int list_size, mincol, minvert, status;

  vertices = ei->grp->vertices;
  list_set = ECOallocvector(&status, vertices);
  if (status) return status;

  minval = vertices;
  minvert = -1;
  for (v = 1; v <= vertices; v++) {
    degree = ECOgetdegree(ei, v);
    /* list_set = V - N[v] */
    list_size = 0;
    for (i = 1; i <= vertices; i++) {
      if (i != v) {
        flag = 1;
        for (d = 1; d <= degree; d++) if (i == ECOgetneighbor(ei, v, d)) flag = 0;
        if (flag) {
          list_size++;
          ECOvset(list_set, list_size, i);
        }
      }
    }
    /* theta is an upper bound of alpha(G - N[v]) */
    ECOvredim(list_set, list_size);
    status = ECOgreedypartition(ei, list_set, &theta);
    if (status) {
      ECOfreevector(&list_set);
      return status;
    }
#ifdef ECODEBUGINIT
    printf("theta(G - N[%d]) = %d, G - N[%d] = ", v, theta, v);
    ECOshowvector(list_set);
#endif
    ECOvmaxsize(list_set);
    /* we need the least one */
    if (theta < minval) {
      minval = theta;
      minvert = v;
    }
  }

  /* remember that ceil((n + 1) / (theta + 2)) <= chieq */
  mincol = (vertices + 1) / (minval + 2);
  if ((vertices + 1) % (minval + 2) > 0) mincol++;
  *min_color = mincol;

  CPXmsg(ei->env->cpxlog, "Fact: alpha(G - N[%d]) <= %d, so chieq >= %d.\n",
         minvert, minval, mincol);
  ECOfreevector(&list_set);
  return ECOERR_OK;
}

int ECOgreedyclique(struct ecointernal *ei, int *max_clique)
{
  ECOVECptr clique_set1, clique_set2, forbidden;
  int vertices, clique_size, v, count, status, sumneigh, bestsumneigh;

  /* request some memory */
  vertices = ei->grp->vertices;
  clique_set2 = ECOallocvector(&status, vertices);
  if (status) return status;
  clique_set1 = ECOallocvector(&status, vertices);
  if (status) {
    ECOfreevector(&clique_set2);
    return status;
  }
  forbidden = ECOallocvector(&status, vertices);
  if (status) {
    ECOfreevector(&clique_set1);
    ECOfreevector(&clique_set2);
    return status;
  }

  /* computes the first clique */
  ECOvfill(forbidden, 0);
  ECOvredim(clique_set1, 0);
  status = ECOgetclique(ei, ei->descending_by_degree, forbidden, clique_set1, 0);
  if (status) {
    ECOfreevector(&forbidden);
    ECOfreevector(&clique_set1);
    ECOfreevector(&clique_set2);
    return status;
  }
  clique_size = ECOvsize(clique_set1);
  sumneigh = 0;
  for (v = 1; v <= clique_size; v++) sumneigh += ECOgetdegree(ei, ECOvget(clique_set1, v));
  ECOvredim(clique_set2, clique_size);
  ECOvcopy(clique_set1, clique_set2);
  bestsumneigh = sumneigh;
  ECOvset(forbidden, ECOvget(clique_set1, 1), 1); /* forbids the first vertex */
#ifdef ECODEBUGINIT
  printf("First clique is (sumneigh=%d):", bestsumneigh);
  ECOshowvector(clique_set2);
#endif

  /* computes other cliques and request the greatest one */
  for (count = 1; count <= vertices-2; count++) {
    ECOvredim(clique_set1, 0);
    status = ECOgetclique(ei, ei->descending_by_degree, forbidden, clique_set1, 0);
    if (status) {
      ECOfreevector(&forbidden);
      ECOfreevector(&clique_set1);
      ECOfreevector(&clique_set2);
      return status;
    }
	/* maximalize the clique found only if the current has at least 2 vertices. */
	if (ECOvsize(clique_set1) >= 2) {
      status = ECOgetclique(ei, ei->sl_ordering, forbidden, clique_set1, 1);
      if (status) {
        ECOfreevector(&forbidden);
        ECOfreevector(&clique_set1);
        ECOfreevector(&clique_set2);
        return status;
      }
      clique_size = ECOvsize(clique_set1);
      if (clique_size >= ECOvsize(clique_set2)) {
	    sumneigh = 0;
        for (v = 1; v <= clique_size; v++) sumneigh += ECOgetdegree(ei, ECOvget(clique_set1, v));
	    if ((clique_size > ECOvsize(clique_set2)) || ((clique_size == ECOvsize(clique_set2)) && sumneigh > bestsumneigh)) {
          ECOvredim(clique_set2, clique_size);
          ECOvcopy(clique_set1, clique_set2);
		  bestsumneigh = sumneigh;
#ifdef ECODEBUGINIT
          printf("Next clique is (sumneigh=%d):", bestsumneigh);
          ECOshowvector(clique_set2);
#endif
	    }
      }
	}

    ECOvset(forbidden, ECOvget(clique_set1, 1), 1);
  }

  ei->clique_set = clique_set2;
  *max_clique = ECOvsize(clique_set2);
  ECOfreevector(&forbidden);
  ECOfreevector(&clique_set1);
  /* note: do not free clique_set2, it is ei->clique_set */
  return ECOERR_OK;
}
