
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * main.c - program startup
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define MAX_ARRAY 100

ECOENVptr env = NULL;
ECOGRPptr grp = NULL;
int vertices = 0;
int perform = 1;

void msgfunction(const char *str)
{
  printf(str);
}

int readgraph(char *name)
{
  char errmsg[1024], buffer[128], c;
  FILE *stream;
  int status, u, v, line, edges;

  stream = fopen(name, "rt");
  if (stream == NULL) {
    printf("Error opening %s file.\n", name);
    return 1;
  }

  line = 1;
  while(fgets(buffer, 128, stream) != NULL) {
    c = buffer[0];
    if (c == 'p') { /* p format */
      if (vertices > 0) {
        printf("Format error in %s file, line %d.\n", name, line);
        return 1;
      }
      sscanf(buffer, "%c %s %d %d", &c, errmsg, &vertices, &edges);
      if (vertices <= 0 || edges < 0) {
        printf("Invalid number of vertices or edges in %s file, line %d.\n", name, line);
        return 1;
      }
      grp = ECOcreategraph(env, &status, vertices, edges);
      if (grp == NULL) {
        ECOgeterrorstring(env, status, errmsg);
        printf("ECOcreategraph: %s\n", errmsg);
        return 1;
      }
    }
    if (c == 'e') { /* e format */
      if (vertices == 0) {
        printf("Format error in %s file, line %d.\n", name, line);
        return 1;
      }
      sscanf(buffer, "%c %d %d", &c, &u, &v);
      if (u <= 0 || v <= 0 || u > vertices || v > vertices) {
        printf("Out of range error in %s file, line %d.\n", name, line);
        return 1;
      }
      status = ECOaddedge(env, grp, u, v);
      if (status) {
        ECOgeterrorstring(env, status, errmsg);
        printf("ECOaddedge: %s\n", errmsg);
        return 1;
      }
    }
    line++;
  }
  if (vertices == 0) {
    printf("Error reading %s file.\n", name);
    return 1;
  }

  fclose(stream);
  printf("%d lines readed from %s file.\n", line - 1, name);
  return 0;
}

int writecoloring(char *name)
{
  char errmsg[1024];
  FILE *stream;
  int status, v, is_optimal, colors_used;
  int timeopt_result, timecut_result, maxtime;
  long long nodes_result, iter_result;
  double gap, bestobj;
  int *coloring;

  /* obtain best coloring and other variables */
  coloring = ECOcolresult(grp, &status, &is_optimal, &colors_used);
  if (coloring == NULL) {
    ECOgeterrorstring(env, status, errmsg);
    printf("ECOcolresult: %s\n", errmsg);
    return 1;
  }
  iter_result = ECOgetitcnt(env);
  nodes_result = ECOgetnodecnt(env);
  timeopt_result = ECOgettimeopt(env);
  timecut_result = ECOgettimecut(env);

  /* normalize timeopt */
  maxtime = ECOgetparam(env, ECO_PARAM_TILIM);
  if (timeopt_result > maxtime) {
    printf("Ouch... timeopt > maxtime. Set timeopt to %d.\n", maxtime);
    timeopt_result = maxtime;
  }

  /* compute the gap as CPLEX does */
  bestobj = ceil(ECOgetbestobjval(env) - ECO_WJTOL);
  gap = ((double)colors_used - bestobj) * 100.0 / (double)colors_used;

  /* write everything in disk ... */
  stream = fopen(name, "wt");
  if (stream == NULL) {
    printf("Error creating %s file.\n", name);
    return 1;
  }

  fprintf(stream, "%d:%d:%d:%d:%lld:%lld:%d:%d:%d\n", vertices, ECOgetedges(env, grp), colors_used,
                  is_optimal, iter_result, nodes_result, timeopt_result, timecut_result, (int)gap);
  for (v = 1; v <= vertices; v++) fprintf(stream, "%d\n", coloring[v - 1]);
  fclose(stream);

  printf("Output saved in %s.\n", name);

  /* ... and in screen */
  if (is_optimal) {
    printf("Best solution found is optimal and has %d colors.\n", colors_used);
  }
  else {
    printf("Best solution found has %d colors.\n", colors_used);
    printf("Best Node Objval = %f\nGap = %.2f%%\n", bestobj, gap);
  }

  free(coloring);
  return 0;
}

int setargs(char *arg)
{
  int option, value;

  if (arg[0] != '-') return 1;
  if (arg[1] == 0) return 1;
  if (arg[2] < '0' || arg[2] > '9') return 1;
  option = arg[2] - '0';

  switch (arg[1]) {
    case 'p': /* cutting-plane value */
      if (arg[3] != '=') return 1;
      if (arg[4] < '0' || arg[4] > '9') return 1;
      sscanf(&arg[4], "%d", &value);
      switch (option) {
        case 1: /* cutting-plane routines */
          if (value < 0 || value > 1023) {
            printf("Oops... num is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_CUTPLANE, value);
          break;
        case 2: /* skip factor */
          if (value < 1 || value > 100) {
            printf("Oops... n is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_SKIPFACTORCUT, value);
          break;
        case 3: /* rounds at root node */
          if (value < 1 || value > 50) {
            printf("Oops... n is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_MAXROUNDSATROOT, value);
          break;
        case 4: /* rounds at other nodes */
          if (value < 0 || value > 10) {
            printf("Oops... n is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_MAXROUNDSATNODE, value);
          break;
        case 5: /* depth of tree */
          if (value < 3 || value > 1000) {
            printf("Oops... depth is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_MAXDEPTHTREE, value);
          break;
        default: /* error */
          return 1;
      }
	  break;
    case 'x': /* other value */
      if (arg[3] != '=') return 1;
      if (arg[4] < '0' || arg[4] > '9') return 1;
      sscanf(&arg[4], "%d", &value);
      switch (option) {
#ifdef ECOPARALLEL	
        case 0: /* number of threads */
          if (value < 0 || value > 16) {
            printf("Oops... thr is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_THREADS, value);
          break;
#endif
        case 1: /* time */
          if (value <= 0 || value > 30758400) {
            printf("Oops... time is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_TILIM, value);
          break;
        case 2: /* iterations (tabu) */
          if (value <= 0 || value > 10000000) {
            printf("Oops... iterations are out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_ITLIM, value);
          break;
        case 3: /* alpha (tabu) */
          if (value < 0 || value > 5000) {
            printf("Oops... alpha is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_ALPHA, value);
          break;
        case 4: /* beta (tabu) */
          if (value < 0 || value > 5000) {
            printf("Oops... beta is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_BETA, value);
          break;
        case 5: /* upper bound */
          if (value <= 1 || value > 10000) {
            printf("Oops... UB is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_USERUB, value);
          break;
        case 6: /* lower bound */
          if (value <= 1 || value > 10000) {
            printf("Oops... LB is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_USERLB, value);
          break;
        case 7: /* frequency of log */
          if (value < 1 || value > 1000) {
            printf("Oops... n is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_MIPINTERVAL, value);
          break;
        case 8: /* max time (heuristic) */
          if (value < 5 || value > 99999) {
            printf("Oops... time is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_MAXTIMEHEUR, value);
          break;
        case 9: /* primal heuristics */
          if (value < 0 || value > 63) {
            printf("Oops... num is out of range.\n");
            return 1;
          }
          ECOsetparam(env, ECO_PARAM_PRIHEUR, value);
          break;
        default: /* error */
          return 1;
      }
      break;
    case 'l': /* lower bound of chieq */
      switch (option) {
        case 1: ECOsetparam(env, ECO_PARAM_LOWERB, ECO_OFF); break;
        case 2: ECOsetparam(env, ECO_PARAM_LOWERB, ECO_ON); break;
        default: return 1;
      }
      break;
    case 'u': /* upper bound of chieq */
      switch (option) {
        case 1: ECOsetparam(env, ECO_PARAM_UPPERB, ECO_UPPERB_DELTA); break;
        case 2: ECOsetparam(env, ECO_PARAM_UPPERB, ECO_UPPERB_GREEDY); break;
        case 3: ECOsetparam(env, ECO_PARAM_UPPERB, ECO_UPPERB_TABU); break;
#ifdef ECOPARALLEL	
        case 4: ECOsetparam(env, ECO_PARAM_UPPERB, ECO_UPPERB_PTABU); break;
#endif
        default: return 1;
      }
      break;
    case 's': /* strategy */
      switch (option) {
        case 1: ECOsetparam(env, ECO_PARAM_STRATEGY, ECO_STRATEGY_OFF); break;
        case 2: ECOsetparam(env, ECO_PARAM_STRATEGY, ECO_STRATEGY_BYB); break;
        case 3: ECOsetparam(env, ECO_PARAM_STRATEGY, ECO_STRATEGY_CPLEX); break;
        case 4: ECOsetparam(env, ECO_PARAM_STRATEGY, ECO_STRATEGY_BYC); break;
        case 5: ECOsetparam(env, ECO_PARAM_STRATEGY, ECO_STRATEGY_ENUM); break;
        case 6: ECOsetparam(env, ECO_PARAM_STRATEGY, ECO_STRATEGY_ENUMDYN1); break;
        case 7: ECOsetparam(env, ECO_PARAM_STRATEGY, ECO_STRATEGY_ENUMDYN2); break;
        case 8: ECOsetparam(env, ECO_PARAM_STRATEGY, ECO_STRATEGY_CPLEXUB); break;
        default: return 1;
      }
      break;
    case 'e': /* execution */
      switch (option) {
        case 1: ECOsetparam(env, ECO_PARAM_WRITEFILE, ECO_FILE_NONE); perform = 1; break;
        case 2: ECOsetparam(env, ECO_PARAM_WRITEFILE, ECO_FILE_LPFILE); perform = 1; break;
        case 3: ECOsetparam(env, ECO_PARAM_WRITEFILE, ECO_FILE_LPFILE); perform = 0; break;
        case 4: ECOsetparam(env, ECO_PARAM_WRITEFILE, ECO_FILE_LTXFILE); perform = 0; break;
        case 5: perform = 2; break;
        default: return 1;
      }
      break;
    case 'f': /* formulation */
      switch (option) {
        case 1: ECOsetparam(env, ECO_PARAM_FORM, ECO_FORM_STANDARD); break;
        case 2: ECOsetparam(env, ECO_PARAM_FORM, ECO_FORM_SYM1); break;
        case 3: ECOsetparam(env, ECO_PARAM_FORM, ECO_FORM_SYM2A); break;
        case 4: ECOsetparam(env, ECO_PARAM_FORM, ECO_FORM_SYM2B); break;
        case 5: ECOsetparam(env, ECO_PARAM_FORM, ECO_FORM_SYM2C); break;
        default: return 1;
      }
      break;
    case 'b': /* branching variable */
      switch (option) {
        case 1: ECOsetparam(env, ECO_PARAM_BRANCH, ECO_BRANCH_CPLEXDEF);
                            ECOsetparam(env, ECO_PARAM_NODESEL, ECO_NODESEL_DFS);
                            break;
        case 2: ECOsetparam(env, ECO_PARAM_BRANCH, ECO_BRANCH_PSEUDOREDUCED);
                            ECOsetparam(env, ECO_PARAM_NODESEL, ECO_NODESEL_DFS);
                            break;
        case 3: ECOsetparam(env, ECO_PARAM_BRANCH, ECO_BRANCH_CALLBACK);
                            ECOsetparam(env, ECO_PARAM_NODESEL, ECO_NODESEL_DFS);
                            break;
        case 4: ECOsetparam(env, ECO_PARAM_BRANCH, ECO_BRANCH_CPLEXDEF);
                            ECOsetparam(env, ECO_PARAM_NODESEL, ECO_NODESEL_BESTBOUND);
                            break;
        case 5: ECOsetparam(env, ECO_PARAM_BRANCH, ECO_BRANCH_PSEUDOREDUCED);
                            ECOsetparam(env, ECO_PARAM_NODESEL, ECO_NODESEL_BESTBOUND);
                            break;
        case 6: ECOsetparam(env, ECO_PARAM_BRANCH, ECO_BRANCH_CALLBACK);
                            ECOsetparam(env, ECO_PARAM_NODESEL, ECO_NODESEL_BESTBOUND);
                            break;
        default: return 1;
      }
      break;
    case 'i': /* implicit enumeration */
      switch (option) {
        case 1: ECOsetparam(env, ECO_PARAM_IMPENUM, ECO_OFF);
                break;
        case 2: ECOsetparam(env, ECO_PARAM_IMPENUM, ECO_ON);
                ECOsetparam(env, ECO_PARAM_IMPVERT, 40);
                break;
        case 3: ECOsetparam(env, ECO_PARAM_IMPENUM, ECO_ON);
                ECOsetparam(env, ECO_PARAM_IMPVERT, 50);
                break;
        case 4: ECOsetparam(env, ECO_PARAM_IMPENUM, ECO_ON);
                ECOsetparam(env, ECO_PARAM_IMPVERT, 60);
                break;
        case 5: ECOsetparam(env, ECO_PARAM_IMPENUM, ECO_ON);
                ECOsetparam(env, ECO_PARAM_IMPVERT, 0);
                break;
		default: return 1;
	  }
	  break;
    default: /* error */
      return 1;
  }
  return 0;
}

void check()
{
  int array_v[MAX_ARRAY], array_j[MAX_ARRAY], array_coef[MAX_ARRAY], array_size;
  int coef, rhs, i, v, j, flag;
  char option[128];

  array_size = 0;
  rhs = 0;

  do {
	printf("Inequality is:");
	if (array_size == 0) printf(" empty\n");
	else {
	  for (i = 0; i < array_size; i++) {
		coef = array_coef[i];
		j = array_j[i];
	    if (j > 0) printf(" %+d x(%d,%d)", coef, array_v[i], j); /* variable is x_vj */
		else printf(" %+d w(%d)", coef, -j); /* variable is wj */
	  }
	  printf(" <= %+d\n", rhs);
	}
	printf("Choose an option:\n");
	printf(" x) Change coefficient of x(v,j)\n");
	printf(" w) Change coefficient of w(j)\n");
	printf(" r) Change right hand side\n");
	printf(" v) Write x(v,1) + x(v,2) + ... + x(v,n) \n");
	printf(" j) Write x(1,j) + x(2,j) + ... + x(n,j) \n");
	printf(" c) Perform check\n");
	printf(" q) Quit\n");
	printf("> ");
	scanf("%s", &option);
	switch (option[0]) {
	  case 'x': case 'X':
	    printf("v > ");
	    scanf("%d", &v);
		if (v < 1 || v > vertices) {
		  printf("v must be between 1 and %d\n", vertices);
		  break;
		}
	    printf("j > ");
	    scanf("%d", &j);
		if (j < 1 || j > vertices) {
		  printf("j must be between 1 and %d\n", vertices);
		  break;
		}
	    printf("coef > ");
	    scanf("%d", &coef);
		if (coef < -10000 || coef > 10000) {
		  printf("coef must be between -10000 and 10000\n");
		  break;
		}
	    printf("Coefficient of x(%d,%d) is now %+d.\n", v, j, coef);
		flag = 0;
	    for (i = 0; i < array_size; i++) {
	      if (array_j[i] == j && array_v[i] == v) {
			array_coef[i] = coef;
			flag = 1;
		    break;
		  }
		}
		if (flag == 0) {
		  array_v[array_size] = v;
		  array_j[array_size] = j;
		  array_coef[array_size] = coef;
		  array_size++;
		}
		break;
	  case 'w': case 'W':
	    printf("j > ");
	    scanf("%d", &j);
		if (j < 1 || j > vertices) {
		  printf("j must be between 1 and %d\n", vertices);
		  break;
		}
	    printf("coef > ");
	    scanf("%d", &coef);
		if (coef < -10000 || coef > 10000) {
		  printf("coef must be between -10000 and 10000\n");
		  break;
		}
	    printf("Coefficient of w(%d) is now %+d.\n", j, coef);
		flag = 0;
	    for (i = 0; i < array_size; i++) {
	      if (array_j[i] == -j) {
			array_coef[i] = coef;
			flag = 1;
		    break;
		  }
		}
		if (flag == 0) {
		  array_v[array_size] = 0;
		  array_j[array_size] = -j;
		  array_coef[array_size] = coef;
		  array_size++;
		}
		break;
	  case 'r': case 'R':
	    printf("rhs > ");
	    scanf("%d", &coef);
		if (coef < -10000 || coef > 10000) {
		  printf("rhs must be between -10000 and 10000\n");
		  break;
		}
		rhs = coef;
		break;
	  case 'v': case 'V':
	    printf("v > ");
	    scanf("%d", &v);
		if (v < 1 || v > vertices) {
		  printf("v must be between 1 and %d\n", vertices);
		  break;
		}
	    printf("coef > ");
	    scanf("%d", &coef);
		if (coef < -10000 || coef > 10000) {
		  printf("coef must be between -10000 and 10000\n");
		  break;
		}
	    printf("Coefficients are now %+d.\n", coef);
		for (j = 1; j <= vertices; j++) {
		  flag = 0;
	      for (i = 0; i < array_size; i++) {
	        if (array_j[i] == j && array_v[i] == v) {
			  array_coef[i] = coef;
			  flag = 1;
		      break;
		    }
		  }
		  if (flag == 0) {
		    array_v[array_size] = v;
		    array_j[array_size] = j;
		    array_coef[array_size] = coef;
		    array_size++;
		  }
		}
		break;
	  case 'j': case 'J':
	    printf("j > ");
	    scanf("%d", &j);
		if (j < 1 || j > vertices) {
		  printf("j must be between 1 and %d\n", vertices);
		  break;
		}
	    printf("coef > ");
	    scanf("%d", &coef);
		if (coef < -10000 || coef > 10000) {
		  printf("coef must be between -10000 and 10000\n");
		  break;
		}
	    printf("Coefficients are now %+d.\n", coef);
		for (v = 1; v <= vertices; v++) {
		  flag = 0;
	      for (i = 0; i < array_size; i++) {
	        if (array_j[i] == j && array_v[i] == v) {
			  array_coef[i] = coef;
			  flag = 1;
		      break;
		    }
		  }
		  if (flag == 0) {
		    array_v[array_size] = v;
		    array_j[array_size] = j;
		    array_coef[array_size] = coef;
		    array_size++;
		  }
		}
		break;
	  case 'c': case 'C':
        ECOcheckineq(env, grp, array_v, array_j, array_coef, array_size, rhs);
		break;
	  default: break;
	}
  } while (option[0] != 'q' && option[0] != 'Q');
}

int main(int argc, char *argv[])
{
  char errmsg[1024];
  char colfile[ECO_MAXFILENAMESIZE], outfile[ECO_MAXFILENAMESIZE];
  int status, args;

  printf("ecopt - Equitable Coloring Optimizer\n");
  printf("Made in 2008-2012 by Daniel E. Severin (mail: daniel@fceia.unr.edu.ar).\n\n");
  printf("Partially supported by grants:\n   ANPCyT PICT 38036 (2007-2009)\n   UBACyT X143 (2008-2010)\n   CONICET PID 204 (2010-2012)\n\n");
  if (argc < 2) {
    printf("Synopsis: ./ecopt file [options]\n\n");
    printf("The file must be given without extension.\n");
    printf("Input: file.col is an undirected simple graph in DIMACS format.\n");
    printf("Output: file.out contains the following output information.\n");
    printf("     vert:edges:colused:isopt:iter:nodes:tottime:cuttime:gap\n");
    printf("     color of the first vertex\n");
    printf("     color of the second vertex\n");
    printf("     ...\n");
    printf("Options:\n");
    printf("1) Execution/writing options:\n");
    printf("    -e1           Only perform optimization (default).\n");
    printf("    -e2           Perform optimization and write formulation in CPLEX LP format, if necessary.\n");
    printf("    -e3           Write standard formulation M in CPLEX LP format.\n");
    printf("    -e4           Write standard formulation M in LINDO LTX format.\n");
    printf("    -e5           Compute validity and dimension of the face of an inequality.\n");
    printf("2) Initial lower bound options:\n");
    printf("    -l1           Do not set a lower bound.\n");
    printf("    -l2           Find a lower bound greedily (default).\n");
    printf("3) Initial upper bound options:\n");
    printf("    -u1           Set an upper bound with Delta + 1.\n");
    printf("    -u2           Use Furmanczyk's O(n^4) Naive algorithm.\n");
    printf("    -u3           Use tabu search - TABUEQCOL (default).\n");
#ifdef ECOPARALLEL
    printf("    -u4           Use Parallel TABUEQCOL.\n");
#endif
    printf("4) MIP strategy:\n");
    printf("    -s1           Do not optimize, return initial solution instead.\n");
    printf("    -s2           Use pure Branch and Bound strategy.\n");
    printf("    -s3           Use CPLEX default strategy.\n");
    printf("    -s4           Use Branch and Cut strategy - ECOPT (default).\n");
	printf("    -s5           Use enumerative algorithm with static color selection.\n");
	printf("    -s6           Use enumerative algorithm with dynamic color selection.\n");
	printf("    -s7           Use enumerative algorithm with dynamic color selection and check last color first.\n");
	printf("    -s8           Default CPLEX satrategy, upper bound set to UB-1.\n");
    printf("5) Formulation selection:\n");
    printf("    -f1           Use standard formulation M.\n");
    printf("    -f2           Use formulation M1: y_j >= y_j+1.\n");
    printf("    -f3           Use M2, ascending order of vertices by degree.\n");
    printf("    -f4           Use M2, descending order of vertices by degree (default).\n");
    printf("    -f5           Use M2, SL ordering of vertices.\n");
    printf("6) Branching variable selection and node selection:\n");
    printf("    -b1           CPLEX default branching strategy with Depth First Search.\n");
    printf("    -b2           Branch based on pseudo reduced costs with Depth First Search.\n");
    printf("    -b3           Perform ECOPT branching rules with Depth First Search.\n");
    printf("    -b4           Same as -b1, but use Best Bound Search.\n");
    printf("    -b5           Same as -b2, but use Best Bound Search.\n");
    printf("    -b6           Same as -b3, but use Best Bound Search (default).\n");
    printf("7) Cutting-plane value parameters:\n");
	printf("    -p1=num       Enable/disable several cutting-plane routines. Default is 335.\n");
	printf("        bit 0 - Cutting-plane routines (default: ON)\n");
	printf("                If it is OFF, the cut-callback is disabled completely!\n");
	printf("        bit 1 - Clique cuts (default: ON)\n");
	printf("        bit 2 - 2-rank cuts (default: ON)\n");
	printf("        bit 3 - Block cuts (default: ON)\n");
	printf("        bit 4 - Neighborhood cuts (default: OFF)\n");
	printf("        bit 5 - Outside-neighborhood cuts (default: OFF)\n");
	printf("        bit 6 - S-color cuts (default: ON)\n");
	printf("        bit 7 - ujks-clique-neighborhood cuts (default: OFF)\n");
	printf("        bit 8 - Extra symmetry breakers (default: ON)\n");
	printf("                If it is OFF, these constraints are added into the formulation.\n");
	printf("        bit 9 - CPLEX cuts (default: OFF)\n");
	printf("                Remember that CPLEX cuts are independent from the cut-callback.\n");
	printf("    -p2=n       Execute cutting-plane routines every n nodes. Default is 2.\n");
	printf("    -p3=n       Execute n cutting-plane rounds at root node. Default is 20.\n");
	printf("    -p4=n       Execute n cutting-plane rounds at other nodes. Default is 1.\n");
	printf("                Use 0 for a Cut-and-Branch execution.\n");
	printf("    -p5=depth   Max depth of tree where routines are exec. Default is 50.\n");
    printf("8) Implicit enumeration parameters:\n");
	printf("    -i1         Do not use implicit enumeration.\n");
	printf("    -i2         Use implicit enumeration when 40 vertices are uncolored.\n");
	printf("    -i3         Use implicit enumeration when 50 vertices are uncolored.\n");
	printf("    -i4         Use implicit enumeration when 60 vertices are uncolored.\n");
	printf("    -i5         Use implicit enumeration when vert.uncolored * UB <= %d (default).\n", ECO_ENUMFACTOR);
    printf("9) Other value parameters:\n");
#ifdef ECOPARALLEL	
    printf("    -x0=thr       Number of threads (0=max). Default is 1 (experimental).\n");
#endif
    printf("    -x1=time      Set max. time limit in seconds. Default is 7200.\n");
    printf("    -x2=iter      Set max. iterations for tabu search. Default is 50000.\n");
    printf("    -x3=alpha     Set alpha coef. * 100. Default is 90.\n");
    printf("    -x4=beta      Set beta coef. * 100. Default is 900.\n");
    printf("    -x5=ub        Set upper bound for chieq.\n");
	printf("    -x6=lb        Set lower bound for chieq.\n");
    printf("    -x7=n         Display log every n nodes. Default is 100.\n");
	printf("    -x8=time      Set max. time in seconds devoted to primal heuristics. Default is 300.\n");
	printf("    -x9=num       Enable/disable several primal heuristics. Default is 7.\n");
	printf("        bit 0 - Heuristics routines (default: ON)\n");
	printf("                If it is OFF, the heur-callback is disabled completely!\n");
	printf("                (no initial solution will be provided to the optimization)\n");
	printf("                Also the enumeration routine will be disabled\n");
	printf("        bit 1 - Custom Heuristic 1 (default: ON)\n");
	printf("        bit 2 - Custom Heuristic 2 (default: ON)\n");
	printf("        bit 3 - If ON then execute heur-callback ONLY in root node (default: OFF)\n");
	printf("        bit 4 - CPLEX heuristics (default: OFF)\n");
	printf("                Remember that CPLEX h. are independent from the heur-callback.\n");
	printf("        bit 5 - Tabu search (default: OFF)\n");
    return 1;
  }

  env = ECOopen(&status, msgfunction, argv[1]);
  if (env == NULL) {
    ECOgeterrorstring(env, status, errmsg);
    printf("ECOopen: %s\n", errmsg);
    return 1;
  }

  for (args = 2; args < argc; args++) {
    if (setargs(argv[args])) {
      printf("Error in options argument.\n");
      ECOclose(&env);
      return 1;
    }
  }

  strcpy(colfile, argv[1]);
  strcat(colfile, ".col");

  if (readgraph(colfile)) {
    ECOclose(&env);
    return 1;
  }

  switch (perform) {
    case 0: /* write formulation */
      status = ECOwriteformulation(env, grp);
      if (status) {
        ECOgeterrorstring(env, status, errmsg);
        printf("ECOwriteformulation: %s\n", errmsg);
        ECOclose(&env);
        return 1;
      }
	  break;
	case 1: /* perform optimization */
      status = ECOoptimize(env, grp);
      if (status) {
        ECOgeterrorstring(env, status, errmsg);
        printf("ECOoptimize: %s\n", errmsg);
        ECOclose(&env);
        return 1;
      }

	  strcpy(outfile, argv[1]);
      strcat(outfile, ".out");

      if (writecoloring(outfile)) {
        ECOclose(&env);
        return 1;
      }
	  break;
	case 2: /* Check validity and dimension of an inequality */
	  check();
      break;
  }

  ECOfreegraph(env, &grp);

  status = ECOclose(&env);
  if (status) {
    ECOgeterrorstring(env, status, errmsg);
    printf("ECOclose: %s\n", errmsg);
    return 1;
  }
  printf("Ready!\n");
  return 0;
}
