
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2009 by Daniel E. Severin
 *
 * memory.c - memory manager
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

ECOVECptr ECOallocvector(int *status, int size)
{
  ECOVECptr vec;

  vec = malloc(sizeof(struct ecovec));
  if (vec == NULL) {
    *status = ECOERR_MEMORY;
    return NULL;
  }

  if (size > 0) {
    vec->elem = malloc(sizeof(int)*size);
    if (vec->elem == NULL) {
      *status = ECOERR_MEMORY;
      free(vec);
      return NULL;
    }
  }
  else vec->elem = NULL;

  vec->size = size;
  vec->maxsize = size;
  *status = ECOERR_OK;
  return vec;
}

void ECOfreevector(ECOVECptr *vec)
{
  if (*vec == NULL) return;
  if ((*vec)->maxsize > 0) free((*vec)->elem);
  free(*vec);
  *vec = NULL;
}

int ECOvsize(ECOVECptr vec)
{
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
#endif

  return vec->size;
}

void ECOvmaxsize(ECOVECptr vec)
{
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
#endif

  vec->size = vec->maxsize;
}

void ECOvredim(ECOVECptr vec, int newsize)
{
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
  if (newsize < 0 || newsize > vec->maxsize) ECOabort(ECOERR_BOUNDS);
#endif

  vec->size = newsize;
}

int ECOvget(ECOVECptr vec, int i)
{
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
  if (i < 1 || i > vec->size) ECOabort(ECOERR_BOUNDS);
#endif

  return vec->elem[i - 1];
}

void ECOvset(ECOVECptr vec, int i, int val)
{
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
  if (i < 1 || i > vec->size) ECOabort(ECOERR_BOUNDS);
#endif

  vec->elem[i - 1] = val;
}

void ECOvincr(ECOVECptr vec, int i)
{
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
  if (i < 1 || i > vec->size) ECOabort(ECOERR_BOUNDS);
#endif

  vec->elem[i - 1]++;
}

void ECOvdecr(ECOVECptr vec, int i)
{
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
  if (i < 1 || i > vec->size) ECOabort(ECOERR_BOUNDS);
#endif

  vec->elem[i - 1]--;
}

void ECOvcopy(ECOVECptr vecsrc, ECOVECptr vecdest)
{
  int i;
#ifdef ECODEBUG
  if (vecsrc == NULL || vecdest == NULL) ECOabort(ECOERR_NULLP);
  if (vecsrc->size != vecdest->size) ECOabort(ECOERR_BOUNDS);
#endif

  for (i = 0; i < vecsrc->size; i++) vecdest->elem[i] = vecsrc->elem[i];
}

void ECOvfill(ECOVECptr vec, int val)
{
  int i;
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
#endif

  for (i = 0; i < vec->size; i++) vec->elem[i] = val;
}

void ECOvflip(ECOVECptr vec)
{
  int i, size, middle, u;
#ifdef ECODEBUG
  if (vec == NULL) ECOabort(ECOERR_NULLP);
#endif

  size = vec->size;
  middle = size / 2;
  for (i = 0; i < middle; i++) {
    u = vec->elem[i];
    vec->elem[i] = vec->elem[size - i - 1];
    vec->elem[size - i - 1] = u;
  }
}

void ECOshowvector(ECOVECptr vec)
{
  int i, size, carry;

  carry = 0;
  size = vec->size;
  for (i = 0; i < size; i++) {
    printf(" %d", vec->elem[i]);
    carry++;
    if (carry == 20) {
      printf("\n  ");
      carry = 0;
    }
  }
  printf(" (size = %d).\n", size);
}

struct ostruct {
  int index;
  int value;
};

static int compare(const void *op1, const void *op2)
{
  const struct ostruct *o1, *o2;

  o1 = (const struct ostruct *)op1;
  o2 = (const struct ostruct *)op2;
  if (o1->value < o2->value) return -1;
  if (o1->value > o2->value) return 1;
  return 0;
}

int ECOsort(ECOVECptr index, ECOVECptr value, ECOVECptr ordering, int ordertype)
{
  int size, i;
  struct ostruct *ord;

  size = ECOvsize(index);
#ifdef ECODEBUG
  if (ECOvsize(value) != size || ECOvsize(ordering) != size) ECOabort(ECOERR_INTERNAL);
#endif

  /* alloc internal memory and fill it */
  ord = malloc(sizeof(struct ostruct)*size);
  if (ord == NULL) return ECOERR_MEMORY;
  for (i = 0; i < size; i++) {
    ord[i].index = index->elem[i];
    /* for randomize order, we take random values instead of original ones */
    if (ordertype == ECO_RANDOMIZE) ord[i].value = rand();
    else ord[i].value = value->elem[i];
  }

  /* sort according to 'value' */
  qsort(ord, size, sizeof(struct ostruct), compare);
  if (ordertype == ECO_DESCENDING) {
    /* descending order */
    for (i = 0; i < size; i++) ordering->elem[i] = ord[size - i - 1].index;
  }
  else {
    /* ascending/randomize order */
    for (i = 0; i < size; i++) ordering->elem[i] = ord[i].index;
  }

  free(ord);
  return ECOERR_OK;
}

ECOMTXptr ECOallocmatrix(int *status, int size1, int size2)
{
  ECOMTXptr mtx;

  mtx = malloc(sizeof(struct ecomtx));
  if (mtx == NULL) {
    *status = ECOERR_MEMORY;
    return NULL;
  }

  if (size1 > 0 && size2 > 0) {
    mtx->elem = malloc(sizeof(int)*size1*size2);
    if (mtx->elem == NULL) {
      *status = ECOERR_MEMORY;
      free(mtx);
      return NULL;
    }
  }
  else mtx->elem = NULL;

  mtx->size1 = size1;
  mtx->size2 = size2;
  *status = ECOERR_OK;
  return mtx;
}

void ECOfreematrix(ECOMTXptr *mtx)
{
  if (*mtx == NULL) return;

  if ((*mtx)->size1 > 0 && (*mtx)->size2 > 0) free((*mtx)->elem);
  free(*mtx);
  *mtx = NULL;
}

int ECOmsize1(ECOMTXptr mtx)
{
#ifdef ECODEBUG
  if (mtx == NULL) ECOabort(ECOERR_NULLP);
#endif

  return mtx->size1;
}

int ECOmsize2(ECOMTXptr mtx)
{
#ifdef ECODEBUG
  if (mtx == NULL) ECOabort(ECOERR_NULLP);
#endif

  return mtx->size2;
}

int ECOmget(ECOMTXptr mtx, int i, int j)
{
#ifdef ECODEBUG
  if (mtx == NULL) ECOabort(ECOERR_NULLP);
  if (i < 1 || i > mtx->size1) ECOabort(ECOERR_BOUNDS);
  if (j < 1 || j > mtx->size2) ECOabort(ECOERR_BOUNDS);
#endif

  return mtx->elem[(i - 1) + mtx->size1 * (j - 1)];
}

void ECOmset(ECOMTXptr mtx, int i, int j, int val)
{
#ifdef ECODEBUG
  if (mtx == NULL) ECOabort(ECOERR_NULLP);
  if (i < 1 || i > mtx->size1) ECOabort(ECOERR_BOUNDS);
  if (j < 1 || j > mtx->size2) ECOabort(ECOERR_BOUNDS);
#endif

  mtx->elem[(i - 1) + mtx->size1 * (j - 1)] = val;
}

void ECOmfill(ECOMTXptr mtx, int val)
{
  int i, max;
#ifdef ECODEBUG
  if (mtx == NULL) ECOabort(ECOERR_NULLP);
#endif

  max = mtx->size1 * mtx->size2;
  for (i = 0; i < max; i++) mtx->elem[i] = val;
}
