
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * optimize.c - main optimization routines
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOoptimize(ECOENVptr env, ECOGRPptr grp)
{
  CPXENVptr cenv;
  struct ecointernal ei;
  int vertices, edges, status, v;
  double timeopt_result, bov, cutoff;
  double start_t, presolved_t, remaining_t;
  char longnumber1[32], longnumber2[32], longnumber3[32];
  long long branchnodes;

  if (env == NULL) return ECOERR_NOENV;
  if (grp == NULL) return ECOERR_NOGRP;

  /* initialize */
  env->iter_result = 0;
  env->nodes_result = 0;
  env->timecut_result = 0;
  env->timeopt_result = 0;

  cenv = env->cplexenv;
  vertices = grp->vertices;
  edges = grp->edges;
  ei.density = (100 * edges) / ((vertices * (vertices-1)) / 2);
  CPXmsg(env->cpxlog, "The graph has %d vertices and %d edges (density = %d%%).\n", vertices, edges, ei.density);
#ifdef ECODEBUG
  ECOshowgraph(grp);
#endif

  /* check if G is isolated */
  env->bestobjval = 1.0;
  if (edges == 0) {
    CPXmsg(env->cpxlog, "The graph only contains isolated vertices: using trivial coloring.\n");
    ECOvfill(grp->coloring, 1);
    grp->colors_used = 1;
    grp->is_optimal = 1;
    return ECOERR_OK;
  }

  /* start with trivial coloring */
  for (v = 1; v <= vertices; v++) ECOvset(grp->coloring, v, v);
  grp->colors_used = vertices;
  grp->is_optimal = 0;

  /* check if G is K_n */
  if (edges == vertices * (vertices - 1) / 2) {
    CPXmsg(env->cpxlog, "The graph is complete: using trivial coloring.\n");
    grp->is_optimal = 1;
    env->bestobjval = (double)vertices;
    return ECOERR_OK;
  }

  /* set internal environment (pointers) */
  ei.env = env;
  ei.grp = grp;
  ECOsetinternal(&ei);

  /* set internal environment (counters) */
  ei.callbacks_made = 0;
  ei.branchcallback_time = 0.0;
  ei.cutcallback_time = 0.0;
  ei.heurcallback_time = 0.0;
  ei.cuts_applied = 0;
  ei.clique_cuts_applied = 0;
  ei.alpha2_cuts_applied = 0;
  ei.block_cuts_applied = 0;
  ei.extra_cuts_applied = 0;
  ei.neighborhood_cuts_applied = 0;
  ei.outsideneighbor_cuts_applied = 0;
  ei.ujks_cuts_applied = 0;
  ei.cs_cuts_applied = 0;
  ei.cut_executed = 0;
  ei.primal_executed = 0;
  ei.enum_found_sol = 0;
  ei.totaliter = 0;
  ei.enumactivated = 0;
  ei.enuminittime = ECOclock();
  ei.enumavailabletime = ECO_ENUMTIMELIMIT;
  ei.totalnodeprocessed = 0;

  /* preprocess: degree, lower and upper bounds ... */
  start_t = ECOclock();
  status = ECOinitdegrees(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }
  status = ECOordervertices(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }
  status = ECOlowerbound(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }
  status = ECOinitcol(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }

  /* check if min_color <= max_color */
  if (ei.min_color > ei.max_color) {
    CPXmsg(env->cpxlog, "Oops... upper bound is smaller than lower bound.\n");
	return ECOERR_INTERNAL;
  }

  /* check if the incumbent is optimal */
  if (ei.min_color == grp->colors_used) {
    CPXmsg(env->cpxlog, "Initial solution is optimal.\n");
    ECOfreemem(&ei);
    grp->is_optimal = 1;
    return ECOERR_OK;
  }

  /* generates an ordering of vertices */
  status = ECOordersymmetry(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }

#ifdef ECOPARALLEL
  ECOthreadinit(&ei);
#endif

  /* initialize structure for tabu search and enumeration algorithm */
  status = ECOtabuinit(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }
  /* if "tabu search" is demanded, then execute it */
  if (env->par[ECO_PARAM_UPPERB] == ECO_UPPERB_TABU ||
      env->par[ECO_PARAM_UPPERB] == ECO_UPPERB_PTABU) {
    if (env->par[ECO_PARAM_STRATEGY] == ECO_STRATEGY_OFF) {
      /* give all the time if there is no MIP optimization. */ 
      presolved_t = ECOclock() - start_t;
      remaining_t = (double)env->par[ECO_PARAM_TILIM] - presolved_t;
      status = ECOtabueqcol(&ei, remaining_t);
    }
    else status = ECOtabueqcol(&ei, ECO_MAXTIMEINITABU);
    if (status) {
      ECOfreemem(&ei);
      return status;
    }
    if (ei.min_color == grp->colors_used) {
      timeopt_result = ECOclock() - start_t;
      env->timeopt_result = (int)timeopt_result;
      CPXmsg(env->cpxlog, "Total time elapsed: %.2f seg.\n", timeopt_result);
      ECOfreemem(&ei);
      return ECOERR_OK;
    }
  }

  /* stop if "no optimization" is requested */
  if (env->par[ECO_PARAM_STRATEGY] == ECO_STRATEGY_OFF) {
    CPXmsg(env->cpxlog, "No MIP optimization. Returning initial coloring.\n");
    timeopt_result = ECOclock() - start_t;
    env->timeopt_result = (int)timeopt_result;
    CPXmsg(env->cpxlog, "Total time elapsed: %.2f seg.\n", timeopt_result);
    ECOfreemem(&ei);
    return ECOERR_OK;
  }

  /* use enumerative algorithm if requested */
  if (env->par[ECO_PARAM_STRATEGY] == ECO_STRATEGY_ENUM
	   || env->par[ECO_PARAM_STRATEGY] == ECO_STRATEGY_ENUMDYN1
	   || env->par[ECO_PARAM_STRATEGY] == ECO_STRATEGY_ENUMDYN2) {
    ei.start_t = ECOclock(); /* <-- also used for enumeration routines! */
    presolved_t = ei.start_t - start_t;
    CPXmsg(env->cpxlog, "Time consumed in preprocess: %.2f seg.\n", presolved_t);
    remaining_t = (double)env->par[ECO_PARAM_TILIM] - presolved_t;
    if (remaining_t <= 0.0) {
      CPXmsg(env->cpxlog, "Too much time at presolve stage.\nGiving initial solution.\n");
      ECOfreemem(&ei);
      return ECOERR_OK;
    }
    CPXmsg(env->cpxlog, "Remaining time: %.2f seg.\n", remaining_t);
	ei.maxtim_t = remaining_t;
	ECOenum(&ei);
    timeopt_result = ECOclock() - start_t;
    env->timeopt_result = (int)timeopt_result;
    CPXmsg(env->cpxlog, "Total time elapsed: %.2f seg.\n", timeopt_result);
    ECOfreemem(&ei);
    return ECOERR_OK;
  }

  /* more preprocessing... (required by formulation.c and some cutting-plane routines) */
  status = ECOmaxindepneighborhood(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }

  /* makes the IP formulation */
  status = ECOprobgen(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }

  /* have we more time? */
  ei.start_t = ECOclock(); /* <-- also used for enumeration routines! */
  presolved_t = ei.start_t - start_t;
  CPXmsg(env->cpxlog, "Time consumed in preprocess: %.2f seg.\n", presolved_t);
  remaining_t = (double)env->par[ECO_PARAM_TILIM] - presolved_t;
  if (remaining_t <= 0.0) {
    CPXmsg(env->cpxlog, "Too much time at presolve stage.\nGiving initial solution.\n");
    ECOfreemem(&ei);
    return ECOERR_OK;
  }
  CPXmsg(env->cpxlog, "Remaining time: %.2f seg.\n", remaining_t);
  ei.maxtim_t = remaining_t;

  /* set CPLEX parameters */
  CPXsetdefaults(cenv);
  CPXsetintparam(cenv, CPX_PARAM_SCRIND, CPX_OFF);
  CPXsetdblparam(cenv, CPX_PARAM_TRELIM, 1024.0);
  CPXsetdblparam(cenv, CPX_PARAM_WORKMEM, 128.0);
  CPXsetdblparam(cenv, CPX_PARAM_TILIM, remaining_t);
  CPXsetintparam(cenv, CPX_PARAM_MIPDISPLAY, 3);
  CPXsetintparam(cenv, CPX_PARAM_MIPINTERVAL, env->par[ECO_PARAM_MIPINTERVAL]);
  CPXsetdblparam(cenv, CPX_PARAM_EPGAP, 0.0);
  CPXsetdblparam(cenv, CPX_PARAM_EPAGAP, 0.0);
  CPXsetdblparam(cenv, CPX_PARAM_EPINT, ECO_WJTOL); /* do not use a lower tolerance. it may cause to drop optimal solutions! */
#if CPX_VERSION >= 1200
  #ifdef ECOPARALLEL
  CPXsetintparam(cenv, CPX_PARAM_THREADS, ei.threads);
  CPXsetintparam(cenv, CPX_PARAM_PARALLELMODE, CPX_PARALLEL_DETERMINISTIC);
  #else
  CPXmsg(env->cpxlog, "Note: Using one thread! (multithreads not implemented yet)\n");
  CPXsetintparam(cenv, CPX_PARAM_THREADS, 1);
  #endif
#endif

  /* Note about presolver:
     If one needs to know the "real" objective value of the relaxation at the root note, one should:
       1. Change in formulation.c the value of min_color to 2.
       2. Change in symmetry.c that only w1 and w2 are set to 1.
       3. Turn off the presolver.

  CPXsetintparam(cenv, CPX_PARAM_PREIND, CPX_OFF); */

  /* set additional parameters according to strategy */
  switch (env->par[ECO_PARAM_STRATEGY]) {
    case ECO_STRATEGY_BYB:
      /* PURE BRANCH AND BOUND */
      CPXmsg(env->cpxlog, "Branch and Bound method selected.\n");
      CPXsetintparam(cenv, CPX_PARAM_PREIND, CPX_OFF);
      CPXsetintparam(cenv, CPX_PARAM_LBHEUR, CPX_OFF);
      CPXsetintparam(cenv, CPX_PARAM_CLIQUES, -1);
      CPXsetintparam(cenv, CPX_PARAM_COVERS, -1);
      CPXsetintparam(cenv, CPX_PARAM_DISJCUTS, -1);
      CPXsetintparam(cenv, CPX_PARAM_FRACCUTS, -1);
      CPXsetintparam(cenv, CPX_PARAM_GUBCOVERS, -1);
      CPXsetintparam(cenv, CPX_PARAM_FLOWCOVERS, -1);
      CPXsetintparam(cenv, CPX_PARAM_FLOWPATHS, -1);
      CPXsetintparam(cenv, CPX_PARAM_IMPLBD, -1);
      CPXsetintparam(cenv, CPX_PARAM_MIRCUTS, -1);
      CPXsetintparam(cenv, CPX_PARAM_PROBE, -1);
      CPXsetintparam(cenv, CPX_PARAM_HEURFREQ, -1);
      CPXsetintparam(cenv, CPX_PARAM_RINSHEUR, -1);
#if CPX_VERSION >= 1100
      CPXsetintparam(cenv, CPX_PARAM_ZEROHALFCUTS, -1);
      CPXsetintparam(cenv, CPX_PARAM_FPHEUR, -1);
      CPXsetintparam(cenv, CPX_PARAM_REPEATPRESOLVE, 0);
#endif
#if CPX_VERSION >= 1200
      CPXsetintparam(cenv, CPX_PARAM_MCFCUTS, -1);
#endif
      break;
    case ECO_STRATEGY_CPLEXUB:
      /* CPLEX DEFAULT + use UB-1, so check if we have computed an UB-eqcol already */
      if (ei.max_color != grp->colors_used) {
        CPXmsg(env->cpxlog, "A %d-eqcol must be known in order to continue. Sorry!\n", ei.max_color);
        CPXmsg(env->cpxlog, "(try executing ecopt with options: -u3 -x5=%d)\n", ei.max_color);
        return ECOERR_INTERNAL;
      }
	  cutoff = (double)ei.max_color - 1.0 + ECO_WJTOL;
	  CPXmsg(env->cpxlog, "Set cutoff = %.4f\n", cutoff);
      CPXsetdblparam(cenv, CPX_PARAM_CUTUP, cutoff);
	  /* do not put a "break" here */
	case ECO_STRATEGY_CPLEX:
      /* CPLEX DEFAULT */
      CPXmsg(env->cpxlog, "CPLEX default method selected.\n");
      break;
    case ECO_STRATEGY_BYC:
      /**************************/
      /* ECOPT - BRANCH AND CUT */
      /**************************/
#ifdef ECODEF_CUTANDBRANCHINLOWDENSITY
      if (ei.density < 40) {
	    /* if density is less than 40%, use Cut-and-Branch instead! */
        CPXmsg(env->cpxlog, "Density < 40. Set Cut and Branch method.\n");
        env->par[ECO_PARAM_MAXROUNDSATNODE] = 0;
      }
#endif
	  if (env->par[ECO_PARAM_MAXROUNDSATNODE] == 0) {
		/* A Cut-and-Branch is requested. */
        CPXmsg(env->cpxlog, "Cut and Branch method selected.\n");
	  }
	  else CPXmsg(env->cpxlog, "Branch and Cut method selected.\n");
      if (env->par[ECO_PARAM_FORM] == ECO_FORM_STANDARD ||
                  env->par[ECO_PARAM_FORM] == ECO_FORM_SYM1) {
        ECOfreemem(&ei);
        return ECOERR_INCOMPAT;
      }
      /* variable selection */
      CPXmsg(env->cpxlog, "Branching strategy: ");
      switch (env->par[ECO_PARAM_BRANCH]) {
        case ECO_BRANCH_CPLEXDEF:
          CPXmsg(env->cpxlog, "CPLEX.\n");
          break;
        case ECO_BRANCH_PSEUDOREDUCED:
          CPXmsg(env->cpxlog, "Pseudo-reduced costs.\n");
          CPXsetintparam(cenv, CPX_PARAM_VARSEL, CPX_VARSEL_PSEUDOREDUCED);
		  break;
        case ECO_BRANCH_CALLBACK:
          CPXmsg(env->cpxlog, "Custom (ECOPT).\n");
          CPXsetintparam(cenv, CPX_PARAM_VARSEL, CPX_VARSEL_PSEUDOREDUCED);
          break;
        default: return ECOERR_INTERNAL;
      }
      /* node selection */
      CPXmsg(env->cpxlog, "Node selection: ");
      switch (env->par[ECO_PARAM_NODESEL]) {
        case ECO_NODESEL_DFS:
          CPXmsg(env->cpxlog, "DFS.\n");
          CPXsetintparam(cenv, CPX_PARAM_NODESEL, CPX_NODESEL_DFS);
          break;
        case ECO_NODESEL_BESTBOUND:
          CPXmsg(env->cpxlog, "BBS.\n");
          CPXsetintparam(cenv, CPX_PARAM_NODESEL, CPX_NODESEL_BESTBOUND);
          break;
        default: return ECOERR_INTERNAL;
      }
      /* implicit enumeration: active ONLY if density is higher than 20% */
      ei.enumactivated = 0;
      CPXmsg(env->cpxlog, "Run implicit enumeration: ");
      if (env->par[ECO_PARAM_IMPENUM] == ECO_ON) {
		if (ei.density >= 20) {
		  if (env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_ON) ei.enumactivated = 1;
		  else CPXmsg(env->cpxlog, "disabled due to heur-callback not available.\n");
		}
		else CPXmsg(env->cpxlog, "disabled due to low density.\n");
        if (ei.enumactivated == 1) {
	      if (env->par[ECO_PARAM_IMPVERT] == 0) CPXmsg(env->cpxlog, "when vert.uncolored * UB <= %d\n", ECO_ENUMFACTOR);
		  else CPXmsg(env->cpxlog, "when %d vertices are uncolored.\n", env->par[ECO_PARAM_IMPVERT]);
		}
	  }
      else CPXmsg(env->cpxlog, "never.\n");
#ifdef ECOPARALLEL
      if (ei.enumactivated == 1 && ei.threads > 1) {
        CPXmsg(env->cpxlog,"Enumeration routine can not run on multithread environments, sorry\n");
        ECOfreemem(&ei);
        return ECOERR_NOTIMPLEMENTED;
      }
#endif
      /* cplex cuts */
      if (!(env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_CPLEX)) { 
        CPXsetintparam(cenv, CPX_PARAM_CLIQUES, -1);
        CPXsetintparam(cenv, CPX_PARAM_COVERS, -1);
        CPXsetintparam(cenv, CPX_PARAM_DISJCUTS, -1);
        CPXsetintparam(cenv, CPX_PARAM_FRACCUTS, -1);
        CPXsetintparam(cenv, CPX_PARAM_GUBCOVERS, -1);
        CPXsetintparam(cenv, CPX_PARAM_FLOWCOVERS, -1);
        CPXsetintparam(cenv, CPX_PARAM_FLOWPATHS, -1);
        CPXsetintparam(cenv, CPX_PARAM_IMPLBD, -1);
        CPXsetintparam(cenv, CPX_PARAM_MIRCUTS, -1);
#if CPX_VERSION >= 1100
        CPXsetintparam(cenv, CPX_PARAM_ZEROHALFCUTS, -1);
#endif
#if CPX_VERSION >= 1200
        CPXsetintparam(cenv, CPX_PARAM_MCFCUTS, -1);
#endif
      }
      /* cplex heuristics */
      if (!(env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_CPLEX)) {
        CPXsetintparam(cenv, CPX_PARAM_HEURFREQ, -1);
        CPXsetintparam(cenv, CPX_PARAM_RINSHEUR, -1);
        CPXsetintparam(cenv, CPX_PARAM_LBHEUR, CPX_OFF);
#if CPX_VERSION >= 1100
        CPXsetintparam(cenv, CPX_PARAM_FPHEUR, -1);
#endif
      }
      else {
        CPXmsg(env->cpxlog, "CPLEX heuristics enabled.\n");
      }
      /* show own heuristics enabled */
      if (env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_ON) {
        if (env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_SCANXVJ) CPXmsg(env->cpxlog, "Heuristic 1 enabled.\n");
        if (env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_SCANCOLOR) CPXmsg(env->cpxlog, "Heuristic 2 enabled.\n");
        if (env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_TABU) CPXmsg(env->cpxlog, "Heuristic tabueqcol enabled.\n");
      }
      /* other parameters */
      CPXsetintparam(cenv, CPX_PARAM_STARTALG, CPX_ALG_BARRIER);
      CPXsetintparam(cenv, CPX_PARAM_SUBALG, CPX_ALG_PRIMAL);
      CPXsetintparam(cenv, CPX_PARAM_PRESLVND, 1);
      CPXsetintparam(cenv, CPX_PARAM_PROBE, -1);
#if CPX_VERSION >= 1100
      CPXsetintparam(cenv, CPX_PARAM_REPEATPRESOLVE, 0);
#endif
      /* set callback routines */
      status = ECOcallbackinit(&ei);
      if (status) {
        ECOfreemem(&ei);
        return status;
      }
#ifdef ECODEF_EXITAFTERROOT
      ei.bestobjvalroot = 1;
      ei.bestroundroot = 0;
      ei.besttimeroot = 0.0;
      ei.firsttimeroot = ECOclock();
      ei.firstobjvalroot = 1;
#endif
      break;
	default:
	  /* BAD OPTION */
      return ECOERR_INTERNAL;
  }

  /* optimize */
  status = CPXmipopt(cenv, ei.lp);
  timeopt_result = ECOclock() - start_t;
#ifdef ECOPARALLEL
  ECOfreememeis(&ei);
#endif

  if (status) {
#ifdef ECODEF_EXITAFTERROOT
    if (status != CPXERR_CALLBACK) {
      ECOfreemem(&ei);
      return status | ECOERR_CPLEX;
    }
#else
    ECOfreemem(&ei);
    return status | ECOERR_CPLEX;
#endif
  }
  if (!CPXgetbestobjval(cenv, ei.lp, &bov)) {
    if (bov > ei.min_color) env->bestobjval = bov;
  }

  status = CPXgetstat(cenv, ei.lp);
  if (status != CPXMIP_OPTIMAL) {
    /* optimality not achieved. */
    CPXmsg(env->cpxlog, "No optimal solution found. Reason: ");
    switch (status) {
      case CPXMIP_TIME_LIM_INFEAS:
        CPXmsg(env->cpxlog, "time limit reached.\nGiving initial solution.\n");
        ECOshowcutlog(&ei);
        ECOfreemem(&ei);
        return ECOERR_OK;
      case CPXMIP_TIME_LIM_FEAS:
        CPXmsg(env->cpxlog, "time limit reached.\nGiving current incumbent.\n");
        ECOshowcutlog(&ei);
        status = ECOgetsolution(&ei);
        ECOfreemem(&ei);
        return status;
      case CPXMIP_INFEASIBLE:
      case CPXMIP_INForUNBD:
        CPXmsg(env->cpxlog, "infeasible.\n");
		if (env->par[ECO_PARAM_STRATEGY] != ECO_STRATEGY_CPLEXUB) return ECOERR_INTERNAL;
        CPXmsg(env->cpxlog, "Initial solution is optimal.\n");
		grp->is_optimal = 1;
		goto results;
      case CPXMIP_ABORT_INFEAS:
        CPXmsg(env->cpxlog, "callback aborted the optimization.\nGiving initial solution.\n");
        ECOfreemem(&ei);
        return ECOERR_OK;
      case CPXMIP_ABORT_FEAS:
        CPXmsg(env->cpxlog, "callback aborted the optimization.\nGiving current incumbent.\n");
        status = ECOgetsolution(&ei);
        ECOfreemem(&ei);
        return status;
    }
#ifdef ECODEF_EXITAFTERROOT
    CPXmsg(env->cpxlog, "callback finish optimization at root node.\n");
    CPXmsg(env->cpxlog, "Best (fractional) lower bound achieved: %.4lf\n", env->bestobjval);
    CPXmsg(env->cpxlog, "First lower bound: %d\n", ei.firstobjvalroot);
    CPXmsg(env->cpxlog, "Best lower bound: %d     Round: %d     Time: %.2f\n", ei.bestobjvalroot, ei.bestroundroot, ei.besttimeroot);
    CPXmsg(env->cpxlog, "GAP between first and best: %.2lf%%\n", 100.0 * ((double)ei.bestobjvalroot - (double)ei.firstobjvalroot) / ((double)ei.firstobjvalroot));
    timeopt_result = ECOclock() - start_t;
    env->timeopt_result = (int)timeopt_result;
    CPXmsg(env->cpxlog, "Total time elapsed: %.2f seg.\n", timeopt_result);
    ECOshowcutlog(&ei);
    ECOfreemem(&ei);
    return ECOERR_OK;
#else
    CPXmsg(env->cpxlog, "unknown.\n");
    ECOfreemem(&ei);
    return status | ECOERR_CPLEX;
#endif
  }

  /* solution is optimal, convert it to a switchable coloring */
  grp->is_optimal = 1;
  status = ECOgetsolution(&ei);
  if (status) {
    ECOfreemem(&ei);
    return status;
  }

  /* set results */
results:;
  env->timeopt_result = (int)timeopt_result;
  env->timecut_result = (int)(ei.branchcallback_time + ei.cutcallback_time + ei.heurcallback_time);
  env->iter_result = CPXgetmipitcnt(cenv, ei.lp);
  branchnodes = CPXgetnodecnt(cenv, ei.lp);
  env->nodes_result = branchnodes + ei.totalnodeprocessed; /* contabilize as "Nodes" the sum of branch nodes and enumeration nodes */

  /* show messages in screen */
  CPXmsg(env->cpxlog, "Time elapsed in optimization: %.2f seg.\n", timeopt_result);
  CPXmsg(env->cpxlog, "Callback routines were executed %d times.\n", ei.callbacks_made);
  CPXmsg(env->cpxlog, "Details of the optimization:\n");
  sprintf(longnumber1, "%lld", env->iter_result);
  CPXmsg(env->cpxlog, " *) LP Iterations: %s\n", longnumber1);
  sprintf(longnumber1, "%lld", env->nodes_result);
  sprintf(longnumber2, "%lld", branchnodes);
  sprintf(longnumber3, "%lld", ei.totalnodeprocessed);
  CPXmsg(env->cpxlog, " *) Nodes evaluated: %s (branch=%s, enum=%s)\n", longnumber1, longnumber2, longnumber3);
  if (env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_ON)
    CPXmsg(env->cpxlog, " *) Cut callback executed: %d\n", ei.cut_executed);
  if (env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_ON)
    CPXmsg(env->cpxlog, " *) Primal heuristic executed: %d\n", ei.primal_executed);
  if (env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_TABU)
    CPXmsg(env->cpxlog, " *) Tabu iterations during MIP optimization = %d\n", ei.totaliter);
  ECOshowcutlog(&ei);
  ECOfreemem(&ei);
  return ECOERR_OK;
}
