
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * partabu.c - Parallel TABUEQCOL algorithm (only with Linux pthreads)
 * Note: if changes are introduced, they must also be introduced in tabu.c!
 * Also ECOPARALLEL must be defined !!!
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <omp.h>
#include <semaphore.h>

struct commondata_t {
  sem_t access;
  struct ecointernal *ei;
  int current_k, best_k_found;
  double start_t, maxtim_t;
};

struct mydata_t {
  int thr;
  struct commondata_t *commondata;
  ECOVECptr bestcoloring, curcoloring, bigclasses;
  ECOMTXptr tenure;
  ECOVECptr conflictvertexset, conflictvertexlist;
  int iter;
};

void *mythread(void *arg)
{
  ECOVECptr bestcoloring, curcoloring, bigclasses;
  ECOMTXptr tenure;
  ECOVECptr conflictvertexset, conflictvertexlist;

  int neighbor, fbest, schema, jmin, alpha, beta, ul;
  int vertices, edges, fcur, fnext, fmin, i, j, u, v, d, degree;
  int conflictcount, tabu, umin, imin, vmin, xfloor, xceil, iter;
  int maxiter, thr, current_k, exitcondition;
  double start_t, maxtim_t,pass_t, clk;
  struct mydata_t *mydata;
  struct commondata_t *commondata;
  struct ecointernal *ei;

  /* set variables */
  mydata = (struct mydata_t*)arg;
  commondata = mydata->commondata;
  ei = commondata->ei;
  start_t = commondata->start_t;
  maxtim_t = commondata->maxtim_t;
  thr = mydata->thr;
  umin = -1;
  imin = -1;
  jmin = -1;
  vertices = ei->grp->vertices;
  edges = ei->grp->edges;
  bestcoloring = mydata->bestcoloring;
  curcoloring = mydata->curcoloring;
  conflictvertexset = mydata->conflictvertexset;
  conflictvertexlist = mydata->conflictvertexlist;
  bigclasses = mydata->bigclasses;
  tenure = mydata->tenure;
  alpha = ei->env->par[ECO_PARAM_ALPHA];
  beta = ei->env->par[ECO_PARAM_BETA];
  maxiter = ei->env->par[ECO_PARAM_ITLIM];

  for(;;) {
    /* get value of current_k */
    sem_wait(&commondata->access); /* BEGIN CRITICAL SECTION */

    current_k = commondata->current_k;
    if (current_k <= ei->min_color) {
      CPXmsg(ei->env->cpxlog, "Thread %d: no more \"k\" needed to explore, exiting.\n", thr + 1);
      sem_post(&commondata->access); /* <-- free semaphore before quit */
      return NULL;
    }
    current_k--;
    commondata->current_k = current_k;

    CPXmsg(ei->env->cpxlog, "Thread %d: we are going to find a %d-eqcol.\n", thr + 1, current_k);
    ECOgeninitialsol(ei, ei->grp->coloring, ei->grp->colors_used, current_k);
    ECOvcopy(ei->tabu_curcoloring, bestcoloring);
    ECOvcopy(ei->tabu_curcoloring, curcoloring);
    ECOvcopy(ei->tabu_bigclasses, bigclasses);

    sem_post(&commondata->access); /* END CRITICAL SECTION */

    /* start tabueqcol algorithm */
    xfloor = ECOfloor(vertices, current_k);
    xceil = ECOceil(vertices, current_k);
    ECOmfill(tenure, 0);

    /* computes the conflictive edges */
    ECOvfill(conflictvertexset, 0);
    ECOvredim(conflictvertexlist, vertices);
    fcur = 0;
    conflictcount = 0;
    for (i = 1; i <= edges; i++) {
      u = ECOvget(ei->grp->edge1, i);
      v = ECOvget(ei->grp->edge2, i);
      if (ECOvget(curcoloring, u) == ECOvget(curcoloring, v)) {
        if (ECOvget(conflictvertexset, u) == 0) {
          ECOvset(conflictvertexset, u, 1);
          conflictcount++;
          ECOvset(conflictvertexlist, conflictcount, u);
        }
        if (ECOvget(conflictvertexset, v) == 0) {
          ECOvset(conflictvertexset, v, 1);
          conflictcount++;
	  ECOvset(conflictvertexlist, conflictcount, v);
        }
        fcur++;
      }
    }
    ECOvredim(conflictvertexlist, conflictcount);
    fbest = fcur;

#ifdef ECODEBUGTABU
    printf("Thread %d: initial solution has %d conflictive edges:", thr + 1, fcur);
    ECOshowvector(curcoloring);
    printf("and the conflictive vertices are:");
    ECOshowvector(conflictvertexlist);
#endif

    iter = 0;
    exitcondition = 0;
    while (exitcondition == 0) {
      /* make iterations until: 1) time is exhausted, 2) limit of iterations is reached
                             or 3) a coloring is found, 4) 10 seconds passed */
      pass_t = ECOclock();
      clk = pass_t;
      while (fbest > 0 && clk - pass_t < 10.0) {
        if (iter >= maxiter) {
          sem_wait(&commondata->access); /* BEGIN CRITICAL SECTION */
          CPXmsg(ei->env->cpxlog, "Thread %d: Max iteration limit reached, exiting.\n", thr + 1);
          sem_post(&commondata->access); /* END CRITICAL SECTION */
	  return NULL;
	}
        iter++;
        mydata->iter++;
      
        /* local search */
        schema = 0;
        fmin = edges;
        for (ul = 1; ul <= conflictcount; ul++) {
          u = ECOvget(conflictvertexlist, ul);
          i = ECOvget(curcoloring, u);

          /* schema 1: movement of a vertex from one class to another */
          if (xfloor != xceil) {
            if (ECOvget(bigclasses, i) == 1) {
              for (j = 1; j <= current_k; j++) {
                if (ECOvget(bigclasses, j) == 0) {
                  /* computes f of next solution
                   * we suppose that vertex u is painted with color j */
                  fnext = fcur;
                  degree = ECOgetdegree(ei, u);
                  for (d = 1; d <= degree; d++) {
                    neighbor = ECOgetneighbor(ei, u, d);
                    if (ECOvget(curcoloring, neighbor) == j) fnext++;
                    if (ECOvget(curcoloring, neighbor) == i) fnext--;
                  }

                  /* check if the solution is tabu */
                  tabu = 0;
                  if (fnext >= fbest) {
                    if (ECOmget(tenure, u, j) != 0) tabu = 1;
                  }

                  /* update the solution */
                  if (fnext < fmin && tabu == 0) {
                    schema = 1;
                    fmin = fnext;
                    umin = u;
                    imin = i;
                    jmin = j;
                  }
                }
              }
            }
          }

          /* schema 2: exchange of vertices from different classes */
          for (v = 1; v <= vertices; v++) {
            j = ECOvget(curcoloring, v);

            /* the following condition ensures that a pair (u, v) is not visited twice */
            if (j > i || (j < i && ECOvget(conflictvertexset, v) == 0)) {
              /* computes f of next solution
               * we suppose that vertex u is painted with color j, and v with color i */
              fnext = fcur;
              degree = ECOgetdegree(ei, u);
              for (d = 1; d <= degree; d++) {
                neighbor = ECOgetneighbor(ei, u, d);
                if (neighbor != v) {
                  if (ECOvget(curcoloring, neighbor) == j) fnext++;
                  if (ECOvget(curcoloring, neighbor) == i) fnext--;
                }
              }
              degree = ECOgetdegree(ei, v);
              for (d = 1; d <= degree; d++) {
                neighbor = ECOgetneighbor(ei, v, d);
                if (neighbor != u) {
                  if (ECOvget(curcoloring, neighbor) == i) fnext++;
                  if (ECOvget(curcoloring, neighbor) == j) fnext--;
                }
              }

              /* check if the solution is tabu */
              tabu = 0;
              if (fnext >= fbest) {
                if (ECOmget(tenure, u, j) != 0) tabu = 1;
                if (ECOmget(tenure, v, i) != 0) tabu = 1;
              }

              /* update the solution */
              if (fnext < fmin && tabu == 0) {
                schema = 2;
                fmin = fnext;
                umin = u;
                vmin = v;
                imin = i;
                jmin = j;
             }
            }
          }
        }

        /* generate new solution */
        switch (schema) {
          case 0:
	    /* tenure is too high, exit now */
            sem_wait(&commondata->access); /* BEGIN CRITICAL SECTION */
            CPXmsg(ei->env->cpxlog, "Thread %d: cannot find solutions (maybe the tabu tenure is too high), exiting.\n", thr + 1);
            sem_post(&commondata->access); /* END CRITICAL SECTION */
	    return NULL;
          case 1:
            ECOvset(curcoloring, umin, jmin);
	    break;
          case 2:
            ECOvset(curcoloring, umin, jmin);
            ECOvset(curcoloring, vmin, imin);
	    break;
        }
        fcur = fmin;

        /* update tabu tenure */
        for (v = 1; v <= vertices; v++) {
          for (j = 1; j <= current_k; j++) {
            tabu = ECOmget(tenure, v, j);
            if (tabu > 0 && tabu <= iter) ECOmset(tenure, v, j, 0);
          }
        }
        tabu = iter + (alpha * conflictcount + beta) / 100;
        ECOmset(tenure, umin, imin, tabu);

        if (fcur < fbest) {
          /* update best solution */
          ECOvcopy(curcoloring, bestcoloring);
          fbest = fcur;
        }
        if (schema == 1) {
          /* update class sizes */
          ECOvset(bigclasses, imin, 0);
          ECOvset(bigclasses, jmin, 1);
        }

        /* computes the conflictive edges again */
        ECOvfill(conflictvertexset, 0);
        ECOvredim(conflictvertexlist, vertices);
        conflictcount = 0;
        for (i = 1; i <= edges; i++) {
          u = ECOvget(ei->grp->edge1, i);
          v = ECOvget(ei->grp->edge2, i);
          if (ECOvget(curcoloring, u) == ECOvget(curcoloring, v)) {
            if (ECOvget(conflictvertexset, u) == 0) {
              ECOvset(conflictvertexset, u, 1);
              conflictcount++;
              ECOvset(conflictvertexlist, conflictcount, u);
            }
            if (ECOvget(conflictvertexset, v) == 0) {
              ECOvset(conflictvertexset, v, 1);
              conflictcount++;
              ECOvset(conflictvertexlist, conflictcount, v);
            }
          }
        }
        ECOvredim(conflictvertexlist, conflictcount);

#ifdef ECODEBUGTABU
        printf("Thread %d: solution in iter %d has %d conflictive edges. ", thr + 1, iter, fcur);
        printf("Vertex %d with color %d is marked tabu, with t = %d.\n", umin, imin, tabu - iter);
#endif

        clk = ECOclock();
        if (clk - start_t >= maxtim_t) {
          sem_wait(&commondata->access); /* BEGIN CRITICAL SECTION */
          CPXmsg(ei->env->cpxlog, "Thread %d: time exhausted, exiting.\n", thr + 1);
          sem_post(&commondata->access); /* END CRITICAL SECTION */
	  return NULL;
	}
      }

      sem_wait(&commondata->access); /* BEGIN CRITICAL SECTION */

      if (fbest > 0) {
	/* -->10 seconds passes
	 * if k is greather than a coloring found by other thread,
	 * do not explore for k-eqcols any more */
        CPXmsg(ei->env->cpxlog, " thread: %d, iter: %d, fbest: %d, time: %.2f seg.\n", thr + 1, iter,
	                                                             fbest, clk - start_t);
#ifdef ECODEBUGTABU
          printf("Thread %d: time to check if a better coloring has been found.", thr + 1);
#endif
	if (current_k >= commondata->best_k_found) {
	  exitcondition = 1;
#ifdef ECODEBUGTABU
          printf("A coloring of %d colors has been found.", commondata->best_k_found);
#endif
	}
      }
      else {
	/* -->a k-eqcol was found
	 * save it in ecointernal */
        CPXmsg(ei->env->cpxlog, "Thread %d: coloring found in %d iterations.\n", thr + 1, iter);
	if (current_k < commondata->best_k_found) {
	  commondata->best_k_found = current_k;
          ECOvcopy(bestcoloring, ei->grp->coloring);
          ei->grp->colors_used = current_k;
          if (current_k < ei->max_color) ei->max_color = current_k;
          if (current_k <= ei->min_color) {
            CPXmsg(ei->env->cpxlog, "Optimality reached!\n");
            ei->grp->is_optimal = 1;
            sem_post(&commondata->access); /* <-- free semaphore before quit */
	    return NULL;
          }
	}
	exitcondition = 1;
      }

      sem_post(&commondata->access); /* END CRITICAL SECTION */
    }
  }

  /* this code is never reached */
  return NULL;
}

int ECOparalleltabueqcol(struct ecointernal *ei, double maxtim_t)
{
  int status, threads, t, vertices, current_k, totaliter;
  pthread_t *thread;
  struct mydata_t *mydata;
  struct commondata_t commondata;
  cpu_set_t cpuset;
  double start_t, lapse_t;

  /* selects k = min(bestsol - 1, maxcolor),
   * note that we know a (bestsol)-eqcol but, possibly, we do not know a (maxcolor)-eqcol. */
  if (ei->grp->colors_used > ei->max_color) current_k = ei->max_color;
  else current_k = ei->grp->colors_used - 1;

  vertices = ei->grp->vertices;
  start_t = ECOclock();

  /* initialize common structure */
  commondata.ei = ei;
  commondata.current_k = current_k + 1; 
  commondata.best_k_found = current_k + 1;
  commondata.start_t = start_t;
  commondata.maxtim_t = maxtim_t;
  sem_init(&commondata.access, 0, 0);   /* semaphore is locked */

  /* initialize thread structure */
  thread = NULL;
  mydata = NULL;

  threads = ei->threads;
  thread = malloc(sizeof(pthread_t) * threads);
  if (thread == NULL) {
    status = ECOERR_MEMORY;
    goto free_and_exit;
  }
  mydata = malloc(sizeof(struct mydata_t) * threads);
  if (mydata == NULL) {
    status = ECOERR_MEMORY;
    goto free_and_exit;
  }

  for (t = 0; t < threads; t++) {
    mydata[t].thr = t;
    mydata[t].commondata = &commondata;
    mydata[t].iter = 0;
    mydata[t].bestcoloring = NULL;
    mydata[t].curcoloring = NULL;
    mydata[t].conflictvertexset = NULL;
    mydata[t].conflictvertexlist = NULL;
    mydata[t].bigclasses = NULL;
    mydata[t].tenure = NULL;
  }

  /* ask memory for each thread */
  for (t = 0; t < threads; t++) {
    mydata[t].bestcoloring = ECOallocvector(&status, vertices);
    if (status) {
      status = ECOERR_MEMORY;
      goto free_and_exit;
    }
    mydata[t].curcoloring = ECOallocvector(&status, vertices);
    if (status) {
      status = ECOERR_MEMORY;
      goto free_and_exit;
    }
    mydata[t].conflictvertexset = ECOallocvector(&status, vertices);
    if (status) {
      status = ECOERR_MEMORY;
      goto free_and_exit;
    }
    mydata[t].conflictvertexlist = ECOallocvector(&status, vertices);
    if (status) {
      status = ECOERR_MEMORY;
      goto free_and_exit;
    }
    mydata[t].bigclasses = ECOallocvector(&status, vertices);
    if (status) {
      status = ECOERR_MEMORY;
      goto free_and_exit;
    }
    mydata[t].tenure = ECOallocmatrix(&status, vertices, vertices);
    if (status) {
      status = ECOERR_MEMORY;
      goto free_and_exit;
    }
  }

  /* create threads */
  for (t = 0; t < threads; t++) {
    if (pthread_create(&thread[t], NULL, mythread, (void*)&mydata[t])) {
      CPXmsg(ei->env->cpxlog, "Error creating thread %d!\n", t + 1);
      ECOabort(ECOERR_INTERNAL);
    }
    CPU_ZERO(&cpuset);
    CPU_SET(t, &cpuset);
    if (pthread_setaffinity_np(thread[t], sizeof(cpu_set_t), &cpuset)) {
      CPXmsg(ei->env->cpxlog, "Warning: unable to set affinity of thread %d!\n", t + 1);
    }
  }

  /* free semaphore */
  sem_post(&commondata.access);

  /* wail threads */
  totaliter = 0;
  for (t = 0; t < threads; t++) {
    if (pthread_join(thread[t], NULL)) {
      CPXmsg(ei->env->cpxlog, "Tabueqcol: error joining thread %d!\n", t + 1);
      ECOabort(ECOERR_INTERNAL);
    }
    totaliter += mydata[t].iter;
  }

  /* show messages in screen */
  CPXmsg(ei->env->cpxlog, "Total iterations: %d\n", totaliter);
  lapse_t = ECOclock() - start_t;
  CPXmsg(ei->env->cpxlog, "Time consumed by Parallel TABUEQCOL: %.2f\n", lapse_t);
  CPXmsg(ei->env->cpxlog, "Average efficiency: %.2f it/s\n", (double)totaliter / lapse_t);
  status = ECOERR_OK;

free_and_exit:
  /* free memory and exit */
  if (mydata != NULL) {
    for (t = 0; t < threads; t++) {
      if (mydata[t].bestcoloring != NULL) ECOfreevector(&mydata[t].bestcoloring);
      if (mydata[t].curcoloring != NULL) ECOfreevector(&mydata[t].curcoloring);
      if (mydata[t].conflictvertexset != NULL) ECOfreevector(&mydata[t].conflictvertexset);
      if (mydata[t].conflictvertexlist != NULL) ECOfreevector(&mydata[t].conflictvertexlist);
      if (mydata[t].bigclasses != NULL) ECOfreevector(&mydata[t].bigclasses);
      if (mydata[t].tenure != NULL) ECOfreematrix(&mydata[t].tenure);
    }
    free(mydata);
  }
  if (thread != NULL) free(thread);

  return status;
}
