
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * primal1.c - primal heuristic callbacks
 ***/

#include <ecopt.h>
#include <stdio.h>

int ECOinitheurcallback(struct ecointernal *ei)
{
  int vertices, status;

#ifdef ECODEBUGPRIMAL
  printf("Primal heuristic enabled.\n");
#endif
  vertices = ei->grp->vertices;
  ei->primal_coloring = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->primal_vector1 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->primal_vector2 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->primal_vector3 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->primal_vector4 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->primal_vector5 = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->primal_vector6 = ECOallocvector(&status, vertices);
  if (status) return status;

  ei->primal_enabled = 1;
  ei->primal_enabled_node = 1;
  ei->primal_executed = 0;
  ei->make_tabu = 0;
  ei->totaliter = 0;

  return ECOERR_OK;
}

/* makes the coloring feasible, according to the symmetry established.
 * Note: we need to execute this routine EVEN if "checkfeas" is turned off because
 * ECOevalx may not work if the coloring does not follow rules given by symmetry. */
int ECOmakefeasible(struct ecointernal *ei)
{
  ECOVECptr index, value, ascending, colclass, coloring;
  int vertices, v, vert, j, status, colors_used;

#ifdef ECODEBUGPRIMAL
  printf("Making the solution feasible...\n");
#endif
  vertices = ei->grp->vertices;
  coloring = ei->primal_coloring;
  colors_used = ei->primal_colors_used;
  index = ei->primal_vector1;
  value = ei->primal_vector2;
  ascending = ei->primal_vector3;
  colclass = ei->primal_vector4;
  ECOvredim(index, colors_used);
  ECOvredim(value, colors_used);
  ECOvredim(ascending, colors_used);
  ECOvredim(colclass, colors_used);

  for (j = 1; j <= colors_used; j++) ECOvset(index, j, j);
  ECOvfill(value, 999999999);

  for (v = 1; v <= vertices; v++) {
    vert = ECOvget(ei->ordering_symmetry, v);
    j = ECOvget(coloring, vert);
    if (ECOvget(value, j) > v) ECOvset(value, j, v);
  }

  status = ECOsort(index, value, ascending, ECO_ASCENDING);
  if (status) return status;

  for (j = 1; j <= colors_used; j++) ECOvset(colclass, ECOvget(ascending, j), j);
  for (v = 1; v <= vertices; v++) {
    j = ECOvget(coloring, v);
    ECOvset(coloring, v, ECOvget(colclass, j));
  }

  return ECOERR_OK;
}

/* gives the tuple (x,w) corresponding to the coloring */
void ECOgivex(struct ecointernal *ei, double *x)
{
  ECOVECptr coloring;
  int vertices, max_color, j, v, painted, index, colors_used;
#ifdef ECODEBUG
  int class_size;
#endif

  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  coloring = ei->primal_coloring;
  colors_used = ei->primal_colors_used;

  for (j = 1; j <= max_color; j++) {
#ifdef ECODEBUG
    class_size = 0;
#endif
    for (v = 1; v <= vertices; v++) {
      /* determine if v is painted with color j */
      painted = (ECOvget(coloring, v) == j ? 1 : 0);
      /* set x_vj */
      index = ECOmget(ei->xvar_state, v, j);
#ifdef ECODEBUG
      if (index == 0 && painted == 1) {
        printf("Vertex %d can not be painted with color %d.\n", v, j);
        ECOabort(ECOERR_INTERNAL);
      }
      if (index == 1 && painted == 0) {
        printf("Vertex %d must be painted with color %d.\n", v, j);
        ECOabort(ECOERR_INTERNAL);
      }
      class_size += painted;
#endif
      if (index >= 2) x[index - 2] = (double)painted;
    }
    /* set w_j */
    index = ECOvget(ei->wvar_state, j);
    if (index >= 2) x[index - 2] = (j <= colors_used ? 1.0 : 0.0);
#ifdef ECODEBUG
    if (j <= colors_used) {
      if (class_size < ECOfloor(vertices, colors_used)
		  || class_size > ECOceil(vertices, colors_used)) {
        printf("The coloring is not equitable. See class j = %d.\n", j);
        ECOabort(ECOERR_INTERNAL);
      }
    }
	else {
	  if (class_size > 0) {
        printf("The coloring is not valid (j = %d).\n", j);
        ECOabort(ECOERR_INTERNAL);
	  }
	}
#endif
  }
}

int ECOheurcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei,
                              double *objval_p, double *x, int *useraction_p)
{
  int status, max_color, exists_upper_bound, current_k, fail, i;

  /* if the enumerative routine has a solution, post it! */
  if (ei->enum_found_sol != 0) {
    status = ECOmakefeasible(ei);
    if (status) return status;

    ECOgivex(ei, x);
    *useraction_p = CPX_CALLBACK_SET;
    *objval_p = (double)ei->primal_colors_used;

    CPXmsg(ei->env->cpxlog, "Coloring of %d colors injected.\n", ei->primal_colors_used);
	ei->enum_found_sol = 0;
    return ECOERR_OK;
  }

  if (ei->primal_enabled == 0 || ei->primal_enabled_node == 0) return ECOERR_OK;

  /* if we are in a subproblem, disable heurs in the same node next time */
  if (ei->isroot == 0) ei->primal_enabled_node = 0;

  /* obtain best value */
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_REMAINING, &ei->LB_current);
  if (status) return status | ECOERR_CPLEX;
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_INTEGER, &ei->UB_current);
  if (status) return status | ECOERR_CPLEX;
  /* check if optimality is reached */
  if (ei->LB_current + ECO_WJTOL > ei->UB_current) return ECOERR_OK;

  ei->primal_executed++;
#ifdef ECODEBUGPRIMAL
  printf("ECOheurcallback: exec = %d, LB = %lf, UB = %lf\n", ei->primal_executed,
	                                     ei->LB_current, ei->UB_current);
#endif

  /* initialize some vars */
  max_color = ei->max_color;
  exists_upper_bound = 1;
  if ((double)max_color < ei->UB_current - ECO_WJTOL) exists_upper_bound = 0;

  /* give first coloring if it is possible: note that colors_used may be greater that max_color,
       e.g. max_color = delta + 1, but colors_used = delta + 2. */
  if (exists_upper_bound == 0 && ei->grp->colors_used == max_color) {
    ECOvcopy(ei->grp->coloring, ei->primal_coloring);
    ei->primal_colors_used = ei->grp->colors_used;

    status = ECOmakefeasible(ei);
    if (status) return status;

    ECOgivex(ei, x);
    *useraction_p = CPX_CALLBACK_SET;
    *objval_p = (double)ei->primal_colors_used;

    CPXmsg(ei->env->cpxlog, "Initial coloring injected.\n");
    return ECOERR_OK;
  }

  if (exists_upper_bound) current_k = (int)(ei->UB_current + ECO_WJTOL);
  else current_k = max_color + 1; /* <-- because colors_used > max_color */

  /* try to compute a better coloring, i.e. less than current_k */
  for (i = 0; i < ei->variables; i++) ei->x[i] = x[i];
  status = ECOprigreedy(ei, current_k, &fail);
  if (status) return status;

  if (fail == 0) {
    status = ECOmakefeasible(ei);
    if (status) return status;

    ECOgivex(ei, x);
    *useraction_p = CPX_CALLBACK_SET;
    *objval_p = (double)ei->primal_colors_used;
	return ECOERR_OK;
  }

  if ((ei->env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_TABU) && ei->make_tabu) {
    /* make a tabu search now! */
    ei->make_tabu = 0;
    status = ECOpritabu(ei, &fail);
    if (status) return status;

    if (fail == 0) {
      status = ECOmakefeasible(ei);
      if (status) return status;

      ECOgivex(ei, x);
      *useraction_p = CPX_CALLBACK_SET;
      *objval_p = (double)ei->primal_colors_used;
    }
  }

  return ECOERR_OK;
}

int ECOincumcallback(CPXCENVptr cenv, void *cbdata, int wherefrom,
                     struct ecointernal *ei, double *x, double objval)
{
  int status, lb, ub, i, max_color, exists_upper_bound;

  /* obtain best lower bound: lb = ceil(LB_current),
             new upper bound: ub = floor(objval),
			 and old upper bound.*/
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_REMAINING, &ei->LB_current);
  if (status) return status | ECOERR_CPLEX;
  status = CPXgetcallbackinfo(cenv, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_INTEGER, &ei->UB_current);
  if (status) return status | ECOERR_CPLEX;
  lb = (int)(ei->LB_current + 1.0 - ECO_WJTOL);
  ub = (int)(objval + ECO_WJTOL);
  /* check if optimality is reached */
  if (lb >= ub) return ECOERR_OK;

  /* initialize some vars */
  max_color = ei->max_color;
  exists_upper_bound = 1;
  if ((double)max_color < ei->UB_current - ECO_WJTOL) exists_upper_bound = 0;

  if (exists_upper_bound == 0 && ei->grp->colors_used == max_color) return ECOERR_OK;

  /* if we reach this point, it is because the new incumbent is not the initial solution. */
  ei->make_tabu = 1;

#ifdef ECODEBUG
  if (ei->primal_colors_used < ub) {
    printf("Error: primal_colors_used = %d, ub = %d\n", ei->primal_colors_used, ub);
	return ECOERR_INTERNAL;
  }
#endif
  if (ub >= ei->primal_colors_used) return ECOERR_OK;

  /* if we reach this point, it is because the new incumbent comes from CPLEX. */
  for (i = 0; i < ei->variables; i++) ei->x[i] = x[i];
  ei->primal_colors_used = ECOevalx(ei, ei->primal_coloring);

#ifdef ECODEBUG
  if (ei->primal_colors_used != ub) {
    printf("Error: primal_colors_used = %d, ub = %d\n", ei->primal_colors_used, ub);
	return ECOERR_INTERNAL;
  }
#endif
  CPXmsg(ei->env->cpxlog, "CPLEX has found a %d-eqcol.\n", ub);

  return ECOERR_OK;
}
