
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * simplecut.c - generate cuts corresponding to polynomial families of ineq.
 * 
 * Note: since all cuts are scanned each time the cutting-plane is executed, we apply them as LOCAL CUTS. 
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

/* block cuts: generate inequalities of the form sum_{k = j}^n xvk - wj <= 0 */
int ECOblockseparation(struct ecointernal *ei)
{
  int status, vertices, first_j_frac, last_j_frac, max_color, cutadded, j, v, k;
  struct ecocut *cu;

  /* set variables */
  cu = &ei->cu;
  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  first_j_frac = ei->first_j_frac;
  last_j_frac = ei->last_j_frac;

  ei->block_cuts_applied_round = 0;

  /* in order to continue, we need at least one fractional variable wj */
  if (ei->is_wj_integer) return ECOERR_OK;

  /* if wj = 0 or wj = 1, the constraint is not violated. */
  for (j = first_j_frac; j <= last_j_frac; j++) {
    for (v = 1; v <= vertices; v++) {
      ECObegincut(cu, 'L', 0);
#ifdef ECODEBUGCUTS
      printf("block cut:");
#endif
      for (k = j; k <= max_color; k++) {
        ECOaddxcut(cu, v, k, 1);
#ifdef ECODEBUGCUTS
        printf(" + x%d_%d", v, k);
#endif
      }
      ECOaddwcut(cu, j, -1);
#ifdef ECODEBUGCUTS
      printf(" - w%d <= 0\n", j);
#endif
      status = ECOendcut(cu, &cutadded, ECO_VIOLATION / 10.0, ECO_OFF); /* ten times less than clique inequalities */
      if (status) return status;
      if (cutadded) {
        ei->block_cuts_applied++;
		ei->block_cuts_applied_round++;
        if (ei->block_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;
	  }
    }
  }

  return ECOERR_OK;
}

/* neighborhood cuts: generate inequalities of the form:
   r x_vj + sum_{u in N(v)} xuj + sum_{k = beta+1}^n (r - ceil(n/k)) x_vk - r wj <= 0
 * NOTE: These cuts must be **LOCAL CUTS** if we use min_color = ceil(obj.func.) */
int ECOneighborhoodseparation(struct ecointernal *ei)
{
  int status, vertices, max_color, min_color, last_j_frac;
  int degree, j, v, k, u, r, beta, cutadded;
  struct ecocut *cu;

  /* set variables */
  cu = &ei->cu;
  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  min_color = (int)(ei->objective + 1.0 - ECO_WJTOL);
  last_j_frac = ei->last_j_frac;

  ei->neighborhood_cuts_applied_round = 0;

  for (v = 1; v <= vertices; v++) {
    if (ECOvget(ei->neighbormaxindep1_size, v) >= 3) {
      for (j = 1; j <= last_j_frac; j++) {
        /* generate neighborhood cut */
        ECObegincut(cu, 'L', 0);
        r = ECOrvalue(ei, v, j, min_color);
#ifdef ECODEBUGCUTS
        printf("neighborhood cut: + %d x%d_%d", r, v, j);
#endif
        ECOaddxcut(cu, v, j, r);

	    degree = ECOgetdegree(ei, v);
	    for (u = 1; u <= degree; u++) {
	      ECOaddxcut(cu, ECOgetneighbor(ei, v, u), j, 1);
#ifdef ECODEBUGCUTS
          printf(" + x%d_%d", ECOgetneighbor(ei, v, u), j);
#endif
	    }
	    beta = 1;
	    while (ECOceil(vertices, beta + 1) >= r) beta++;
	    for (k = beta + 1; k <= max_color; k++) {
          ECOaddxcut(cu, v, k, r - ECOceil(vertices, k));
#ifdef ECODEBUGCUTS
          printf(" + %d x%d_%d", r - ECOceil(vertices, k), v, k);
#endif
	    }
	    ECOaddwcut(cu, j, -r);
#ifdef ECODEBUGCUTS
        printf(" - %d w%d <= 0\n", r, j);
#endif
        status = ECOendcut(cu, &cutadded, ECO_VIOLATION / 10.0, ECO_ON);  /* ten times less than clique inequalities */
	    if (status) return status;
        if (cutadded) {
		  ei->neighborhood_cuts_applied++;
          ei->neighborhood_cuts_applied_round++;
          if (ei->neighborhood_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;
		}
      }
    }
  }

  return ECOERR_OK;
}

/* outsideneighbor cuts: see (v,j)-outside-neighborhood inequalities...
 * NOTE: These cuts must be **LOCAL CUTS** if we use min_color = ceil(obj.func.) */
int ECOoutsideneighborseparation(struct ecointernal *ei)
{
  int status, vertices, max_color, min_color, cutadded;
  int degree, j, v, k, u, hatj, floorj, ceilj, lastcol;
  struct ecocut *cu;
  ECOVECptr noneighborhood;

  /* set variables */
  cu = &ei->cu;
  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  min_color = (int)(ei->objective + 1.0 - ECO_WJTOL);

  ei->outsideneighbor_cuts_applied_round = 0;

  noneighborhood = ECOallocvector(&status, vertices);
  if (status) return status;

  /* lastcol = min{ floor(n/2), last_j_frac } */
  lastcol = ECOfloor(vertices, 2);
  if (lastcol > ei->last_j_frac) lastcol = ei->last_j_frac;

  for (v = 1; v <= vertices; v++) {
    if (ECOvget(ei->neighbormaxindep1_size, v) >= 3) {
      /* first, we compute set V \ N[u] */
      ECOvfill(noneighborhood, 1);
      ECOvset(noneighborhood, v, 0);
      degree = ECOgetdegree(ei, v);
      for (u = 1; u <= degree; u++) ECOvset(noneighborhood, ECOgetneighbor(ei, v, u), 0);
      /* then, we scan j = 1...min{floor(n/2),last_j_frac} */
      for (j = 1; j <= lastcol; j++) {
        hatj = j;
        if (hatj < min_color) hatj = min_color;
        floorj = ECOfloor(vertices, hatj);
        ceilj = ECOceil(vertices, hatj);
        if (floorj != ceilj && ceilj >= 3 && ECOvget(ei->neighbormaxindep1_size, v) >= floorj) {
          ECObegincut(cu, 'L', 0);
#ifdef ECODEBUGCUTS
          printf("outsideneighbor cut: + %d x%d_%d", floorj - 1, v, j);
#endif
          ECOaddxcut(cu, v, j, floorj - 1);

	      for (u = 1; u <= vertices; u++) {
            if (ECOvget(noneighborhood, u)) {
	          ECOaddxcut(cu, u, j, -1);
#ifdef ECODEBUGCUTS
              printf(" - x%d_%d", u, j);
#endif
	        }
	      }
	      for (k = hatj + 1; k <= max_color; k++) {
            ECOaddxcut(cu, v, k, floorj - ECOfloor(vertices, k));
#ifdef ECODEBUGCUTS
            printf(" + %d x%d_%d", floorj - ECOfloor(vertices, k), v, k);
#endif
          }
	      for (k = hatj + 1; k <= max_color; k++) {
            ECOaddwcut(cu, k, -(ECOfloor(vertices, k-1) - ECOfloor(vertices, k)));
#ifdef ECODEBUGCUTS
            printf(" - %d w%d", ECOfloor(vertices, k-1) - ECOfloor(vertices, k), k);
#endif
          }
#ifdef ECODEBUGCUTS
          printf(" <= 0\n");
#endif
          status = ECOendcut(cu, &cutadded, ECO_VIOLATION / 10.0, ECO_ON); /* ten times less than clique inequalities */
          if (status) {
	        ECOfreevector(&noneighborhood);
	        return status;
	      }
	      if (cutadded) {
		    ei->outsideneighbor_cuts_applied++;
			ei->outsideneighbor_cuts_applied_round++;
            if (ei->outsideneighbor_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;
	      }
        }
      }
    }
  }

  ECOfreevector(&noneighborhood);
  return ECOERR_OK;
}

/* extra symmetry breakers: generate inequalities of the form given in formulation.c (EXTRASYMMETRY) */
int ECOextraseparation(struct ecointernal *ei)
{
  int status, vertices, cutadded, j, v, v2, k, vert, u;
  struct ecocut *cu;

  /* set variables */
  cu = &ei->cu;
  vertices = ei->grp->vertices;

  ei->extra_cuts_applied_round = 0;

  for (v = 2; v <= vertices; v++) {
    vert = ECOvget(ei->ordering_symmetry, v);
    for (j = 2; j <= v - 1; j++) {
      if (ECOxval(vert, j, ei) > ECO_EPSILON) {
        /* inequality: x_{vj} - sum_{u=j-1}^{v-1} x_{uj-1} <= 0, for v = 2,...,n and j = 2,...,v-1 */
        ECObegincut(cu, 'L', 0);
#ifdef ECODEBUGCUTS
        printf("extra cut: + x%d_%d", vert, j);
#endif
        ECOaddxcut(cu, vert, j, 1);
        k = 0;
        for (u = j - 1; u <= v - 1; u++) {
          v2 = ECOvget(ei->ordering_symmetry, u);
		  /* if x_uj-1 = 1 for some u, then the restriction is redundant */
          if (ECOxval(v2, j - 1, ei) > 1.0 - ECO_EPSILON) k = 1;
          ECOaddxcut(cu, v2, j - 1, -1);
#ifdef ECODEBUGCUTS
          printf(" - x%d_%d", v2, j - 1);
#endif
        }
#ifdef ECODEBUGCUTS
        printf(" <= 0\n");
#endif
        if (k == 0) {
          status = ECOendcut(cu, &cutadded, ECO_VIOLATION / 10.0, ECO_ON); /* ten times less than clique inequalities */
          if (status) return status;
          if (cutadded) {
		    ei->extra_cuts_applied++;
			ei->extra_cuts_applied_round++;
            if (ei->extra_cuts_applied_round >= ECO_MAXFAMILYCUTROUND) return ECOERR_OK;
		  }
        }
      }
    }
  }

  return ECOERR_OK;
}
