
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * symmetry.c - find an order of vertices that breaks some symmetries
 ***/

#include <ecopt.h>
#include <stdio.h>

int ECOordersymmetry(struct ecointernal *ei)
{
  int v, vert, index, vertices, status, clique_size;
  int z, flag, j, degree, max_color;

  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  ei->ordering_symmetry = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->xvar_state = ECOallocmatrix(&status, vertices, vertices);
  if (status) return status;
  ei->wvar_state = ECOallocvector(&status, vertices);
  if (status) return status;

  /* first vertices are used by the initial clique */
  clique_size = ECOvsize(ei->clique_set);
  for (v = 1; v <= clique_size; v++)
    ECOvset(ei->ordering_symmetry, v, ECOvget(ei->clique_set, v));

  /* last vertices are ordered according to formulation */
  index = 0;
  for (v = clique_size + 1; v <= vertices; v++) {
    index++;
    do {
      switch (ei->env->par[ECO_PARAM_FORM]) {
        case ECO_FORM_STANDARD:
        case ECO_FORM_SYM1:
        case ECO_FORM_SYM2B:
          vert = ECOvget(ei->descending_by_degree, index);
		  break;
        case ECO_FORM_SYM2A:
          vert = ECOvget(ei->ascending_by_degree, index);
		  break;
        case ECO_FORM_SYM2C:
          vert = ECOvget(ei->sl_ordering, index);
		  break;
	  }

      /* if vert belongs to the initial clique, discard it */
      z = 0;
      flag = 1;
      do {
        z++;
        if (vert == ECOvget(ei->clique_set, z)) {
          flag = 0;
          index++;
        }
      } while (flag == 1 && z < clique_size);
    } while (flag == 0);
    ECOvset(ei->ordering_symmetry, v, vert);
  }

#ifdef ECODEF_DONOTORDER
  /* overwrite ordering with trivial one! */
  for (v = 1; v <= vertices; v++) ECOvset(ei->ordering_symmetry, v, v);
#endif

#ifdef ECODEBUGINIT
  printf("Symmetry ordering:");
  ECOshowvector(ei->ordering_symmetry);
#endif

  /* fill values of x_vj matrix */
  for (v = 1; v <= vertices; v++) {
    vert = ECOvget(ei->ordering_symmetry, v);
    for (j = 1; j <= v; j++) ECOmset(ei->xvar_state, vert, j, 2);
	for (j = v + 1; j <= vertices; j++) {
      if (ei->env->par[ECO_PARAM_FORM] == ECO_FORM_STANDARD
		  || ei->env->par[ECO_PARAM_FORM] == ECO_FORM_SYM1)
	    ECOmset(ei->xvar_state, vert, j, 2);
	  else ECOmset(ei->xvar_state, vert, j, 0); /* <-- symmetry breaking */
	}
  }

  /* fill values of w_j vector: w_j = 1 if j <= lowerbound, w_j = 0 if j > upperbound
   * also fill values of x_vj = 0 if j > upperbound */
  for (j = 1; j <= ei->min_color; j++) ECOvset(ei->wvar_state, j, 1);
  for (j = ei->min_color + 1; j <= max_color; j++) ECOvset(ei->wvar_state, j, 2);
  for (j = max_color + 1; j <= vertices; j++) {
	for (v = 1; v <= vertices; v++) ECOmset(ei->xvar_state, v, j, 0);
    ECOvset(ei->wvar_state, j, 0);
  }

  /* fix variables according to the clique: remember that model SYM1 must not
   *               fix a clique in order to not delete factible solutions
   * do not fix if the ordering is not given! */
#ifdef ECODEF_FIXCLIQUE
#ifndef ECODEF_DONOTORDER
  if (ei->env->par[ECO_PARAM_FORM] != ECO_FORM_SYM1) {
    for (v = 1; v <= clique_size; v++) {
      vert = ECOvget(ei->ordering_symmetry, v);
      /* fix x_vv = 1 for v in the clique */
	  for (j = 1; j <= max_color; j++) ECOmset(ei->xvar_state, vert, j, 0);
      ECOmset(ei->xvar_state, vert, v, 1);
      /* fix x_uv = 0 for u adyacent to v */
      degree = ECOgetdegree(ei, vert);
      for (z = 1; z <= degree; z++) ECOmset(ei->xvar_state, ECOgetneighbor(ei, vert, z), v, 0);
    }
  }
  else {
    CPXmsg(ei->env->cpxlog, "Warning: Clique not fixed due to the formulation selected!\n");
  }
#endif
#endif

  return ECOERR_OK;
}

static double getval(int index, struct ecointernal *ei)
{
  if (index == 0) return 0.0;
  if (index == 1) return 1.0;
  return ei->x[index - 2];
}

double ECOxval(int v, int j, struct ecointernal *ei)
{
  return getval(ECOmget(ei->xvar_state, v, j), ei);
}

double ECOwval(int j, struct ecointernal *ei)
{
  return getval(ECOvget(ei->wvar_state, j), ei);
}

int ECOevalx(struct ecointernal *ei, ECOVECptr coloring)
{
  int v, j, vertices, max_color, color, last_color;

  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  last_color = 0;
  for (v = 1; v <= vertices; v++) {
    color = -1;
    for (j = 1; j <= max_color; j++) {
      if (ECOxval(v, j, ei) > 0.5) {
        color = j;
        continue;
      }
    }
    ECOvset(coloring, v, color);
    if (color > last_color) last_color = color;
  }
  return last_color;
}

int ECOgetsolution(struct ecointernal *ei)
{
  int status;

#if CPX_VERSION >= 1100
  status = CPXgetx(ei->env->cplexenv, ei->lp, ei->x, 0, ei->variables - 1);
#else
  status = CPXgetmipx(ei->env->cplexenv, ei->lp, ei->x, 0, ei->variables - 1);
#endif
  if (status) return status | ECOERR_CPLEX;

  ei->grp->colors_used = ECOevalx(ei, ei->grp->coloring);
  return ECOERR_OK;
}
