
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * tabu.c - TABUEQCOL algorithm
 * Note: if changes are introduced, they must also be introduced in partabu.c!
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOtabuinit(struct ecointernal *ei)
{
  int vertices, status;

  /* ask for memory (it is used by ECOtabueqcol, ECOpritabu and ECOenum) */
  vertices = ei->grp->vertices;
  ei->tabu_bestcoloring = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->tabu_curcoloring = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->tabu_conflictvertexset = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->tabu_conflictvertexlist = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->tabu_bigclasses = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->tabu_coloredvert = ECOallocvector(&status, vertices);
  if (status) return status;
  ei->tabu_tenure = ECOallocmatrix(&status, vertices, vertices);
  if (status) return status;

  return ECOERR_OK;
}

int ECOtabueqcol(struct ecointernal *ei, double maxtim_t)
{
  int current_k, stopcondition, iter, maxiter, totaliter, status;
  double start_t, lapse_t;

#ifdef ECOPARALLEL
  if (ei->env->par[ECO_PARAM_UPPERB] == ECO_UPPERB_PTABU && ei->threads != 1) {
    CPXmsg(ei->env->cpxlog, "Executing Parallel TABUEQCOL.\n");
    return ECOparalleltabueqcol(ei, maxtim_t);
  }
#endif
  CPXmsg(ei->env->cpxlog, "Executing TABUEQCOL.\n");

  /* selects k = min(bestsol - 1, maxcolor),
   * note that we know a (bestsol)-eqcol but, possibly, we do not know a (maxcolor)-eqcol. */
  if (ei->grp->colors_used > ei->max_color) current_k = ei->max_color;
  else current_k = ei->grp->colors_used - 1;

  /* main loop */
  start_t = ECOclock();
  maxiter = ei->env->par[ECO_PARAM_ITLIM];
  totaliter = 0;
  stopcondition = 0;

  do {
    CPXmsg(ei->env->cpxlog, "We are going to find a %d-eqcol.\n", current_k);

    /* obtain initial solution */
    ECOgeninitialsol(ei, ei->grp->coloring, ei->grp->colors_used, current_k);
    ECOvcopy(ei->tabu_curcoloring, ei->tabu_bestcoloring);

    /* make the tabu search */
    status = ECOtabusearch(ei, &iter, maxiter, start_t, maxtim_t, current_k);
    totaliter += iter;

    switch (status) {
      case ECOTABU_TENURETOOHIGH:/* first stop condition: tabu tenure too high */
        CPXmsg(ei->env->cpxlog, "Cannot find solutions! Maybe the tabu tenure is too high.\n");
        stopcondition = 1;
        break;
      case ECOTABU_TIMEEXHAUSTED: /* second stop condition: time limit */
        CPXmsg(ei->env->cpxlog, "Time exhausted! Giving best solution found.\n");
        stopcondition = 1;
        break;
      case ECOTABU_ITERLIMIT: /* third stop condition: iter limit */
        CPXmsg(ei->env->cpxlog, "Max iteration limit reached! Giving best solution found.\n");
        stopcondition = 1;
        break;
      case ECOTABU_OK:
        CPXmsg(ei->env->cpxlog, "Coloring found in %d iterations.\n", iter);
        ECOvcopy(ei->tabu_bestcoloring, ei->grp->coloring);
        ei->grp->colors_used = current_k;
        if (current_k < ei->max_color) ei->max_color = current_k;

        if (current_k <= ei->min_color) { /* fourth stop condition: optimal solution */
          CPXmsg(ei->env->cpxlog, "Optimality reached!\n");
          stopcondition = 1;
          ei->grp->is_optimal = 1;
        }

        /* if UB is defined by user, we stop here since we have the coloring requested by the user */
		if (current_k == ei->env->par[ECO_PARAM_USERUB]) stopcondition = 1;
        break;
      default: /* an error has ocurred */
	    return status;
    }

    /* next k */
    current_k--;
  } while (stopcondition == 0);

  /* show messages in screen */
  CPXmsg(ei->env->cpxlog, "Total iterations: %d\n", totaliter);
  lapse_t = ECOclock() - start_t;
  CPXmsg(ei->env->cpxlog, "Time consumed by TABUEQCOL: %.2f\n", lapse_t);
  CPXmsg(ei->env->cpxlog, "Average efficiency: %.2f it/s\n", (double)totaliter / lapse_t);
  return ECOERR_OK;
}

/* generates the initial solution:
 *   1) if we do not have an initial (k+1)-eqcol, then we use the order
        previously computed.
 *   2) if we have an initial (k+1)-eqcol, then we scan classes 1..k of that
        coloring. */
void ECOgeninitialsol(struct ecointernal *ei, ECOVECptr coloring, int colors_used, int current_k)
{
  ECOVECptr curcoloring, coloredvert, bigclasses, ordering;
  int vertices, v, j, i;

  vertices = ei->grp->vertices;
  curcoloring = ei->tabu_curcoloring;
  coloredvert = ei->tabu_coloredvert;
  bigclasses = ei->tabu_bigclasses;
  ordering = ei->ordering_symmetry;

//  if (current_k == colors_used - 1) {
    /* first case */
    ECOvfill(coloredvert, 0);
    j = 1;
	do {
	  v = -1;
      for (i = 1; i <= vertices; i++) {
        if (ECOvget(coloring, i) == j && ECOvget(coloredvert, i) == 0) {
	      v = i;
  		  break;
		}
	  }
      if (v != -1) {
	    ECOvset(coloredvert, v, 1);
        ECOvset(curcoloring, v, j);
        j++;
        if (j > current_k) j = 1;
	  }
	} while (v != -1);

    do {
	  v = -1;
      for (i = 1; i <= vertices; i++) {
        if (ECOvget(coloredvert, ECOvget(ordering, i)) == 0) {
		  v = ECOvget(ordering, i);
		  break;
		}
	  }
	  if (v != -1) {
	    ECOvset(coloredvert, v, 1);
        ECOvset(curcoloring, v, j);
        j++;
        if (j > current_k) j = 1;
	  }
	} while (v != -1);
//  }
//  else {
    /* second case */
//    CPXmsg(ei->env->cpxlog, "Generating partial solution from scratch.\n");
//    j = 1;
//    for (i = 1; i <= vertices; i++) {
//      ECOvset(curcoloring, ECOvget(ordering, i), j);
//      j++;
//      if (j > current_k) j = 1;
//	}
//  }
  ECOvfill(bigclasses, 0);
  for (i = 1; i <= vertices % current_k; i++) ECOvset(bigclasses, i, 1);
}

int ECOtabusearch(struct ecointernal *ei, int *iter, int maxiter, double start_t,
                  double maxtim_t, int current_k)
{
  ECOVECptr bestcoloring, curcoloring, bigclasses;
  ECOMTXptr tenure;
  ECOVECptr conflictvertexset, conflictvertexlist;

  int neighbor, fbest, schema, jmin, alpha, beta, ul;
  int vertices, edges, fcur, fnext, fmin, i, j, u, v, d, degree;
  int conflictcount, tabu, umin, imin, vmin, xfloor, xceil;
  double pass_t, clk;

  /* set variables */
  umin = -1;
  imin = -1;
  jmin = -1;
  vertices = ei->grp->vertices;
  edges = ei->grp->edges;
  bestcoloring = ei->tabu_bestcoloring;
  curcoloring = ei->tabu_curcoloring;
  conflictvertexset = ei->tabu_conflictvertexset;
  conflictvertexlist = ei->tabu_conflictvertexlist;
  bigclasses = ei->tabu_bigclasses;
  tenure = ei->tabu_tenure;
  xfloor = ECOfloor(vertices, current_k);
  xceil = ECOceil(vertices, current_k);
  alpha = ei->env->par[ECO_PARAM_ALPHA];
  beta = ei->env->par[ECO_PARAM_BETA];
  *iter = 0;

  /* we start with an empty list */
  ECOmfill(tenure, 0);

  /* computes the conflictive edges */
  ECOvfill(conflictvertexset, 0);
  ECOvredim(conflictvertexlist, vertices);
  fcur = 0;
  conflictcount = 0;
  for (i = 1; i <= edges; i++) {
    u = ECOvget(ei->grp->edge1, i);
    v = ECOvget(ei->grp->edge2, i);
    if (ECOvget(curcoloring, u) == ECOvget(curcoloring, v)) {
      if (ECOvget(conflictvertexset, u) == 0) {
        ECOvset(conflictvertexset, u, 1);
        conflictcount++;
        ECOvset(conflictvertexlist, conflictcount, u);
      }
      if (ECOvget(conflictvertexset, v) == 0) {
        ECOvset(conflictvertexset, v, 1);
        conflictcount++;
	ECOvset(conflictvertexlist, conflictcount, v);
      }
      fcur++;
    }
  }
  ECOvredim(conflictvertexlist, conflictcount);
  fbest = fcur;

#ifdef ECODEBUGTABU
  printf("Initial solution has %d conflictive edges:", fcur);
  ECOshowvector(curcoloring);
  printf("and the conflictive vertices are:");
  ECOshowvector(conflictvertexlist);
#endif

  /* make iterations until we reach a valid k-eqcol */
  pass_t = ECOclock();
  while (fbest > 0) {
    if (*iter >= maxiter) return ECOTABU_ITERLIMIT;
    (*iter)++;

    /* local search */
    schema = 0;
    fmin = edges;
    for (ul = 1; ul <= conflictcount; ul++) {
      u = ECOvget(conflictvertexlist, ul);
      i = ECOvget(curcoloring, u);

      /* schema 1: movement of a vertex from one class to another */
      if (xfloor != xceil) {
        if (ECOvget(bigclasses, i) == 1) {
          for (j = 1; j <= current_k; j++) {
            if (ECOvget(bigclasses, j) == 0) {
              /* computes f of next solution
               * we suppose that vertex u is painted with color j */
              fnext = fcur;
              degree = ECOgetdegree(ei, u);
              for (d = 1; d <= degree; d++) {
                neighbor = ECOgetneighbor(ei, u, d);
                if (ECOvget(curcoloring, neighbor) == j) fnext++;
                if (ECOvget(curcoloring, neighbor) == i) fnext--;
              }

              /* check if the solution is tabu */
              tabu = 0;
              if (fnext >= fbest) {
                if (ECOmget(tenure, u, j) != 0) tabu = 1;
              }

              /* update the solution */
              if (fnext < fmin && tabu == 0) {
                schema = 1;
                fmin = fnext;
                umin = u;
                imin = i;
                jmin = j;
              }
            }
          }
        }
      }

      /* schema 2: exchange of vertices from different classes */
      for (v = 1; v <= vertices; v++) {
        j = ECOvget(curcoloring, v);

	/* the following condition ensures that a pair (u, v) is not visited twice */
        if (j > i || (j < i && ECOvget(conflictvertexset, v) == 0)) {
          /* computes f of next solution
           * we suppose that vertex u is painted with color j, and v with color i */
          fnext = fcur;
          degree = ECOgetdegree(ei, u);
          for (d = 1; d <= degree; d++) {
            neighbor = ECOgetneighbor(ei, u, d);
            if (neighbor != v) {
              if (ECOvget(curcoloring, neighbor) == j) fnext++;
              if (ECOvget(curcoloring, neighbor) == i) fnext--;
            }
          }
          degree = ECOgetdegree(ei, v);
          for (d = 1; d <= degree; d++) {
            neighbor = ECOgetneighbor(ei, v, d);
            if (neighbor != u) {
              if (ECOvget(curcoloring, neighbor) == i) fnext++;
              if (ECOvget(curcoloring, neighbor) == j) fnext--;
            }
          }

          /* check if the solution is tabu */
          tabu = 0;
          if (fnext >= fbest) {
            if (ECOmget(tenure, u, j) != 0) tabu = 1;
            if (ECOmget(tenure, v, i) != 0) tabu = 1;
          }

          /* update the solution */
          if (fnext < fmin && tabu == 0) {
            schema = 2;
            fmin = fnext;
            umin = u;
            vmin = v;
            imin = i;
            jmin = j;
          }
        }
      }
    }

    /* generate new solution */
    switch (schema) {
      case 0:
	return ECOTABU_TENURETOOHIGH;
      case 1:
        ECOvset(curcoloring, umin, jmin);
	break;
      case 2:
        ECOvset(curcoloring, umin, jmin);
        ECOvset(curcoloring, vmin, imin);
	break;
    }
    fcur = fmin;

    /* update tabu tenure */
    for (v = 1; v <= vertices; v++) {
      for (j = 1; j <= current_k; j++) {
        tabu = ECOmget(tenure, v, j);
        if (tabu > 0 && tabu <= *iter) ECOmset(tenure, v, j, 0);
      }
    }
    tabu = (*iter) + (alpha * conflictcount + beta) / 100;
    ECOmset(tenure, umin, imin, tabu);

    if (fcur < fbest) {
      /* update best solution */
      ECOvcopy(curcoloring, bestcoloring);
      fbest = fcur;
    }
    if (schema == 1) {
      /* update class sizes */
      ECOvset(bigclasses, imin, 0);
      ECOvset(bigclasses, jmin, 1);
    }

    /* computes the conflictive edges again */
    ECOvfill(conflictvertexset, 0);
    ECOvredim(conflictvertexlist, vertices);
    conflictcount = 0;
    for (i = 1; i <= edges; i++) {
      u = ECOvget(ei->grp->edge1, i);
      v = ECOvget(ei->grp->edge2, i);
      if (ECOvget(curcoloring, u) == ECOvget(curcoloring, v)) {
        if (ECOvget(conflictvertexset, u) == 0) {
          ECOvset(conflictvertexset, u, 1);
          conflictcount++;
          ECOvset(conflictvertexlist, conflictcount, u);
        }
        if (ECOvget(conflictvertexset, v) == 0) {
          ECOvset(conflictvertexset, v, 1);
          conflictcount++;
          ECOvset(conflictvertexlist, conflictcount, v);
        }
      }
    }
    ECOvredim(conflictvertexlist, conflictcount);

#ifdef ECODEBUGTABU
    printf("Solution in iter %d has %d conflictive edges:", *iter, fcur);
    ECOshowvector(curcoloring);
    printf("and the conflictive vertices are:");
    ECOshowvector(conflictvertexlist);
    printf("Vertex %d with color %d is marked tabu, with t = %d.\n", umin, imin, tabu - (*iter));
#endif

    clk = ECOclock();
    if (clk - start_t >= maxtim_t) return ECOTABU_TIMEEXHAUSTED;

    /* log */
    if (clk - pass_t >= 10.0) {
      CPXmsg(ei->env->cpxlog, " iter: %d, fbest: %d, time: %.2f seg.\n", *iter, fbest, clk - start_t);
      pass_t = clk;
    }
  }

  return ECOTABU_OK;
}
