
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * threads.c - thread management (only with Linux pthreads)
 * ECOPARALLEL must be defined !!!
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <omp.h>

struct ecointernal *eis; /* other ecointernal structures */

void ECOthreadinit(struct ecointernal *ei)
{
  int maxthr, threads, par;

  /* check CPU core availability */
  maxthr = omp_get_max_threads();
  CPXmsg(ei->env->cpxlog, "Maximum number of CPUs: %d\n", maxthr);
  par = ei->env->par[ECO_PARAM_THREADS];
  if (par == 0 || par >= maxthr) threads = maxthr;
  else threads = par;
  CPXmsg(ei->env->cpxlog, "We are going to use %d threads.\n", threads);
  ei->threads = threads;
  ei->self = pthread_self();
  ei->tstate = 1;

  eis = NULL;
}

int ECOcreateeis(struct ecointernal *ei)
{
  struct ecointernal *curei;
  int i, status;

  /* alloc mem for other structures */
  eis = malloc(sizeof(struct ecointernal) * (ei->threads - 1));
  if (eis == NULL) {
    return ECOERR_MEMORY;
  }

  /* copy master structure and nullify pointers */
  for (i = 2; i <= ei->threads; i++) {
    curei = &eis[i - 2];
    memcpy(curei, ei, sizeof(struct ecointernal));
    ECOsetinternal_onlycallback(curei);
  }

  /* ask for memory for each structure */
  for (i = 2; i <= ei->threads; i++) {
    curei = &eis[i - 2];

    status = ECOtabuinit(curei);
    if (status) {
      ECOfreememeis(ei);
      return status;
    }
 
    curei->x = malloc(sizeof(double)*ei->variables);
    if (curei->x == NULL) {
      ECOfreememeis(ei);
      return ECOERR_MEMORY;
    }

    if (ei->env->par[ECO_PARAM_CUTPLANE] & ECO_CUTPLANE_BIT_ON) {
	  status = ECOinitcutcallback(curei);
      if (status) {
        ECOfreememeis(ei);
        return ECOERR_MEMORY;
      }
	}

    if (ei->env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_ON) {
      status = ECOinitheurcallback(curei);
      if (status) {
        ECOfreememeis(ei);
        return ECOERR_MEMORY;
      }
	}

    if (ei->env->par[ECO_PARAM_BRANCH] == ECO_BRANCH_CALLBACK) {
      status = ECOinitbranchcallback(curei);
      if (status) {
        ECOfreememeis(ei);
        return ECOERR_MEMORY;
      }
    }
    
    /* fill other fields */
    curei->tstate = 0;
    curei->isroot = 0;
  }

  return ECOERR_OK;
}

struct ecointernal *ECOgetei(struct ecointernal *ei)
{
  pthread_t self;
  struct ecointernal *curei;
  int i;

  self = pthread_self();
  if (pthread_equal(self, ei->self)) return ei;

  for (i = 2; i <= ei->threads; i++) {
    curei = &eis[i - 2];
    if (curei->tstate == 0) {
      curei->tstate = 1;
      curei->self = self;
      return curei;
    }
    if (pthread_equal(self, curei->self)) return curei;
  }

  /* if we reach this, it means that an extra thread was created by CPLEX */
  abort();
}

void ECOfreememeis(struct ecointernal *ei)
{
  struct ecointernal *curei;
  int i;

  if (eis != NULL) {
    for (i = 2; i <= ei->threads; i++) {
      curei = &eis[i - 2];

      /* update counters */
      ei->callbacks_made += curei->callbacks_made;
      ei->branchcallback_time += curei->branchcallback_time;
      ei->cutcallback_time += curei->cutcallback_time;
      ei->heurcallback_time += curei->heurcallback_time;
      ei->cuts_applied += curei->cuts_applied;
      ei->clique_cuts_applied += curei->clique_cuts_applied;
      ei->alpha2_cuts_applied += curei->alpha2_cuts_applied;
      ei->block_cuts_applied += curei->block_cuts_applied;
      ei->extra_cuts_applied += curei->extra_cuts_applied;
      ei->neighborhood_cuts_applied += curei->neighborhood_cuts_applied;
      ei->outsideneighbor_cuts_applied += curei->outsideneighbor_cuts_applied;
      ei->ujks_cuts_applied += curei->ujks_cuts_applied;
      ei->cs_cuts_applied += curei->cs_cuts_applied;
      ei->cut_executed += curei->cut_executed;
      ei->primal_executed += curei->primal_executed;
      ei->totaliter += curei->totaliter;

      /* free memory */
      ECOfreemem_onlycallback(curei);
    }
    free(eis);
  }
}
