
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * upperbound.c - initial equitable coloring algorithms
 ***/

#include <ecopt.h>
#include <stdlib.h>
#include <stdio.h>

int ECOinitcol(struct ecointernal *ei)
{
  int max_color, vertices, delta, userub;

  vertices = ei->grp->vertices;

  userub = ei->env->par[ECO_PARAM_USERUB];
  if (userub > 0 && userub < vertices) {
    /* UB is defined by user */
    CPXmsg(ei->env->cpxlog, "Using upper bound supplied by user: %d.\n", userub);
    ei->max_color = userub;
	return ECOERR_OK;
  }

  /* at this point, we only know the trivial coloring
   * takes min(delta+1, n-1) as the max color */
  delta = ECOgetdegree(ei, ECOvget(ei->descending_by_degree, 1));
  CPXmsg(ei->env->cpxlog, "Value of Delta is %d.\n", delta);
  max_color = delta + 1;
  if (max_color == vertices) max_color--;

  CPXmsg(ei->env->cpxlog, "Initial equitable coloring: ");
  if (ei->env->par[ECO_PARAM_UPPERB] == ECO_UPPERB_DELTA) CPXmsg(ei->env->cpxlog, "Trivial.\n");
  else {
    CPXmsg(ei->env->cpxlog, "Naive coloring.\n");
    /* use a naive coloring */
    ei->grp->colors_used = ECOnaivecoloring(ei, ei->grp->coloring);
    if (ei->grp->colors_used < max_color) max_color = ei->grp->colors_used;
  }

  ei->max_color = max_color;
  CPXmsg(ei->env->cpxlog, "Initial coloring has %d colors.\nUpper bound: %d.\n",
         ei->grp->colors_used, max_color);
  return ECOERR_OK;
}

int ECOnaivecoloring(struct ecointernal *ei, ECOVECptr coloring)
{
  int j, max_color, colors_used;

  ECOvfill(coloring, 0); /* color zero is considered "not painted yet" */
  colors_used = ECOvsize(ei->clique_set);
  for (j = 1; j <= colors_used; j++)
    ECOvset(coloring, ECOvget(ei->clique_set, j), j);
#ifdef ECODEBUGINIT
  printf("Partial coloring has %d colors:", colors_used);
  ECOshowvector(coloring);
#endif
  max_color = ECOfurmanczyk(ei, ei->sl_ordering, coloring, colors_used);
#ifdef ECODEBUGINIT
  printf("Naive coloring using SL ordering has %d colors:", max_color);
  ECOshowvector(coloring);
#endif

  return max_color;
}

/* This functions receives a partial coloring and generates a complete equitable coloring.
 * The partial coloring must use all the colors between 1 and colors_used. */
int ECOfurmanczyk(struct ecointernal *ei, ECOVECptr ordering, ECOVECptr coloring, int colors_used)
{
  int v, i, c, n, vertices, degree, mindegree;
  int colmin, colmax, min, max, flag;

  /* greedy coloring */
  vertices = ei->grp->vertices;
  for (i = 1; i <= vertices; i++) {
    v = ECOvget(ordering, i);
    if (ECOvget(coloring, v) == 0) {
      /* v is not painted yet */
      c = 0;
      do {
        c++;
        degree = ECOgetdegree(ei, v);
        flag = 0;
        for (n = 1; n <= degree; n++)
          if (ECOvget(coloring, ECOgetneighbor(ei, v, n)) == c) flag = 1;
      } while (flag);
      if (c > colors_used) colors_used++;
      ECOvset(coloring, v, c);
    }
  }

  flag = 1;
  do {
    /* compute least and most frequently used color */
    colmin = -1;
    colmax = -1;
    min = 1000000;
    max = -1;
    for (c = 1; c <= colors_used; c++) {
      n = 0;
      for (v = 1; v <= vertices; v++) if (ECOvget(coloring, v) == c) n++;
      if (n < min) {
        min = n;
        colmin = c;
      }
      if (n > max) {
        max = n;
        colmax = c;
      }
    }

    if (min + 1 < max) {
      /* the coloring is not equitable */
      v = 0;
      do {
        i = 1;
        v++;
        if (ECOvget(coloring, v) == colmax) {
          degree = ECOgetdegree(ei, v);
          i = 0;
          for (n = 1; n <= degree; n++)
            if (ECOvget(coloring, ECOgetneighbor(ei, v, n)) == colmin) i = 1;
        }
      } while (i == 1 && v < vertices);

      if (i) {
        /* add new color to a colmax vertex with lowest degree */
        mindegree = 1000000;
        n = -1;
        for (v = 1; v <= vertices; v++) {
          if (ECOvget(coloring, v) == colmax) {
            degree = ECOgetdegree(ei, v);
            if (degree < mindegree) {
              mindegree = degree;
              n = v;
            }
          }
        }
        colors_used++;
        ECOvset(coloring, n, colors_used);
      }
      else {
        /* assign colmin to vertex v */
        ECOvset(coloring, v, colmin);
      }
    }
    else flag = 0;
  } while (flag);

  return colors_used;
}
